/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Vector4f.h"

#include "ClipCoords.h"   /// own header is included last


using namespace hxa7241::graphics;




/// standard object services -----------------------------------------------------------------------
ClipCoords::ClipCoords()
{
}


ClipCoords::ClipCoords
(
	const Vector4f& point
)
{
	setForPoint( point );
}


ClipCoords::~ClipCoords()
{
}


ClipCoords::ClipCoords
(
	const ClipCoords& other
)
 :	flags_m( other.flags_m )
{
	copyArray( 6, other.boundaryCoords_m, boundaryCoords_m );
}


ClipCoords& ClipCoords::operator=
(
	const ClipCoords& other
)
{
	if( &other != this )
	{
		copyArray( 6, other.boundaryCoords_m, boundaryCoords_m );
		flags_m = other.flags_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void ClipCoords::setForPoint
(
	const Vector4f& point
)
{
	boundaryCoords_m[0] = point.getX();
	boundaryCoords_m[1] = point.getW() - point.getX();
	boundaryCoords_m[2] = point.getY();
	boundaryCoords_m[3] = point.getW() - point.getY();
	boundaryCoords_m[4] = point.getZ();
	boundaryCoords_m[5] = point.getW() - point.getZ();

	flags_m.setFromBoundaryCoords( boundaryCoords_m );
}




/// queries ----------------------------------------------------------------------------------------
bool ClipCoords::getInterpolator
(
	const ClipCoords& nextPoint,
	float&            interpolator
) const
{
	bool isLineVisible;// = true;

	const ClipFlags clipFlagsOr ( flags_m | nextPoint.flags_m );
	const ClipFlags clipFlagsAnd( flags_m & nextPoint.flags_m );

	if( clipFlagsOr.isUnclipped() )
	{
		/// line inside
		interpolator  = 1.0f;
		isLineVisible = true;
	}
	else if( clipFlagsAnd.isClipped() )
	{
		/// line outside
		interpolator  = 0.0f;
		isLineVisible = false;
	}
	else
	{
		/// line crossing or outside
		float alpha0 = 0.0f;
		float alpha1 = 1.0f;
		//const ClipFlags::EPlane planes[] = { ClipFlags::ZUpper, ClipFlags::ZLower, ClipFlags::YUpper, ClipFlags::YLower, ClipFlags::XUpper, ClipFlags::XLower };
		const ClipFlags::EPlane planes[] = { ClipFlags::XLower, ClipFlags::XUpper, ClipFlags::YLower, ClipFlags::YUpper, ClipFlags::ZLower, ClipFlags::ZUpper };
		for( int i = 6;  (i-- > 0) & (alpha0 <= alpha1); )
		{
			if( clipFlagsOr.isClipped( planes[i] ) )
			{
				const float alpha = boundaryCoords_m[i] / (boundaryCoords_m[i] - nextPoint.boundaryCoords_m[i]);
				if( flags_m.isClipped( planes[i] ) )
				{
					if( alpha0 < alpha )
					{
						alpha0 = alpha;
					}
				}
				else
				{
					if( alpha1 > alpha )
					{
						alpha1 = alpha;
					}
				}
			}
		}

		if( alpha0 <= alpha1 )
		{
			/// line crossing
			interpolator  = flags_m.isClipped() ? alpha0 : alpha1;
			isLineVisible = true;
		}
		else
		{
			/// line outside
			interpolator  = 0.0f;
			isLineVisible = false;
		}
	}

	return isLineVisible;
}




/// implementation ---------------------------------------------------------------------------------
void ClipCoords::copyArray
(
	const int    length,
	const float* pSource,
	float*       pTarget
)
{
	const float* pEnd = pSource + length;

	while( pSource < pEnd )
	{
		*(pTarget++) = *(pSource++);
	}
}
