/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/




#ifndef Quaternion_h
#define Quaternion_h


#include "Primitives.h"
#include "Vector3f.h"
#include <iosfwd>

//#include "hxa7241_utility.h"




#include "hxa7241_graphics.h"
namespace hxa7241 { namespace graphics {
	using namespace hxa7241::primitives;
	//using hxa7241::utility::SerializerOut;
	//using hxa7241::utility::SerializerIn;


/**
 * a unit quaternion.<br/><br/>
 *
 * all angles in radians.<br/><br/>
 *
 * this has a dependency on the matrix class, in that rotations produced by each
 * must match - and the matrix can have either of two handednesss.<br/>
 * for a different matrix, fiddle with the code marked 'alternate form' and run
 * the tests.<br/><br/>
 *
 * references:<br/>
 * <cite>'The Matrix and Quaternions FAQ',
 * http://www.j3d.org/matrix_faq/matrfaq_latest.html</cite><br/>
 * <cite>'CS184: Using Quaternions to Represent Rotation',
 * http://www-inst.eecs.berkeley.edu/~cs184/</cite>
 *
 * @invariants
 * s_m and v_m, considered as a 4 element vector, is always unitized
 */
class Quaternion
{
/// standard object services -----------------------------------------------------------------------
public:
	         Quaternion();
	         Quaternion( float angle, const UnitVector3f& axis );
	explicit Quaternion( const Matrix3f& );
	         Quaternion( float rotationX, float rotationY, float rotationZ );
protected:
	         Quaternion( float s, const Vector3f& v, bool dummy );
	         Quaternion( float s, float vx, float vy, float vz );
public:
	        ~Quaternion();
	         Quaternion( const Quaternion& );
	Quaternion& operator=( const Quaternion& );

	friend  std::ostream& operator<<( std::ostream&, const Quaternion& );
	friend  std::istream& operator>>( std::istream&,       Quaternion& );


/// commands ---------------------------------------------------------------------------------------
	        void  setAngleAndAxis( float angle, const UnitVector3f& axis );
	        void  setMatrix( const Matrix3f& );
	        void  setRotations( float rotationX, float rotationY, float rotationZ );

	        Quaternion& invert();
	        Quaternion& operator*=( const Quaternion& );
	        Quaternion& operator*=( float );


/// queries ----------------------------------------------------------------------------------------
	        void  getAngleAndAxis( float& angle, UnitVector3f& axis )                         const;
	        void  getMatrix( Matrix3f& )                                                      const;

	        Quaternion getInverse()                                                           const;
	        Quaternion operator*( const Quaternion& )                                         const;
	        Quaternion operator*( float )                                                     const;

	        void  rotate( Vector3f& )                                                         const;

	        bool  operator==( const Quaternion& )                                             const;
	        bool  operator!=( const Quaternion& )                                             const;


/// constants --------------------------------------------------------------------------------------
	static const Quaternion& QUARTER_X();
	static const Quaternion& QUARTER_Y();
	static const Quaternion& QUARTER_Z();


/// serialization ----------------------------------------------------------------------------------
	//friend  SerializerOut& operator<<( SerializerOut&, const Quaternion& );
	//friend  SerializerIn&  operator>>( SerializerIn&,        Quaternion& );
	//static const char* CLASS_NAME();
	//static const char* CLASS_DESCRIPTION();


/// test -------------------------------------------------------------------------------------------
	static  bool  test( bool isLogging );


/// implementation ---------------------------------------------------------------------------------
protected:
	        void  setElements( float s, const Vector3f& v );
	        void  setElements( float s, float vx, float vy, float vz );
	        void  unitize();


/// fields -----------------------------------------------------------------------------------------
private:
	float    s_m;
	Vector3f v_m;
};




/// serialization ----------------------------------------------------------------------------------
//SerializerOut& operator<<( SerializerOut&, const Quaternion& );
//SerializerIn&  operator>>( SerializerIn&,        Quaternion& );

std::ostream& operator<<( std::ostream&, const Quaternion& );
std::istream& operator>>( std::istream&,       Quaternion& );


}}




#endif//Quaternion_h
