/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Serialization.h"
#include "hxa7241math.h"
#include <iostream>

#include "UnitVector3f.h"


using namespace hxa7241::graphics;




/// standard object services -----------------------------------------------------------------------
UnitVector3f::UnitVector3f()
 :	vector_m( 0.0f, 0.0f, 0.0f )
{
}


UnitVector3f::UnitVector3f( const float x, const float y, const float z )
 :	vector_m( x, y, z )
{
	unitize();
}


UnitVector3f::UnitVector3f( const Vector3f& v )
 :	vector_m( v )
{
	unitize();
}


UnitVector3f::UnitVector3f( const UnitVector3f& other )
 :	vector_m( other.vector_m )
{
}


UnitVector3f& UnitVector3f::operator=( const UnitVector3f& other )
{
	vector_m = other.vector_m;

	return *this;
}


UnitVector3f& UnitVector3f::operator=( const Vector3f& v )
{
	vector_m = v;
	unitize();

	return *this;
}




/// streaming --------------------------------------------------------------------------------------
std::ostream& hxa7241::graphics::operator<<( std::ostream& out, const UnitVector3f& obj )
{
	return out << "(|" << obj.vector_m.x_m << ", " << obj.vector_m.y_m << ", " << obj.vector_m.z_m << "|)";
}


std::istream& hxa7241::graphics::operator>>( std::istream& in, UnitVector3f& obj )
{
	char c;

	in >> c >> c >> obj.vector_m.x_m >> c >> obj.vector_m.y_m >> c >> obj.vector_m.z_m >> c >> c;
	obj.unitize();

	return in;
}




/// access -----------------------------------------------------------------------------------------
UnitVector3f& UnitVector3f::setXYZ( const float x, const float y, const float z )
{
	vector_m.setXYZ( x, y, z );
	unitize();

	return *this;
}


UnitVector3f& UnitVector3f::setXYZ( const Vector3f& v )
{
	vector_m = v;
	unitize();

	return *this;
}




/// arithmetic -------------------------------------------------------------------------------------
UnitVector3f UnitVector3f::cross( const UnitVector3f& v ) const
{
	return UnitVector3f( vector_m.cross( v.vector_m ) );
}


UnitVector3f& UnitVector3f::crossEq( const UnitVector3f& v )
{
	vector_m.crossEq( v.vector_m );
	unitize();

	return *this;
}


UnitVector3f UnitVector3f::orthogonalize( const UnitVector3f& v ) const
{
	return UnitVector3f( *this ).orthogonalizeEq( v );
}


UnitVector3f& UnitVector3f::orthogonalizeEq( const UnitVector3f& v )
{
	float d = dot( v );
	if( (d == 1.0f) | (d == -1.0f) )
	{
		if( (v != Y()) & (v != -Y()) )
		{
			*this = Y();
		}
		else
		{
			*this = X();
		}
		d = dot( v );
	}

	if( d != 0.0f )
	{
		*this = this->getVector() - (v.getVector() * d);
	}

	return *this;
}


UnitVector3f UnitVector3f::reflect( const UnitVector3f& normal ) const
{
	return UnitVector3f( *this ).reflectEq( normal );
}


UnitVector3f& UnitVector3f::reflectEq( const UnitVector3f& normal )
{
	const float NdotV = vector_m.dot(normal.vector_m);
	vector_m.negate();
	vector_m += (normal.vector_m * (NdotV * 2.0f));

	return *this;
}


UnitVector3f& UnitVector3f::reflectAlongEq( const UnitVector3f& normal )
{
	const float NdotV = vector_m.dot(normal.vector_m);
	vector_m -= (normal.vector_m * (NdotV * 2.0f));

	return *this;
}


UnitVector3f UnitVector3f::refract
(
	const UnitVector3f& normal,
	const float         relativeRefractiveIndex
) const
{
	return UnitVector3f( *this ).refractEq( normal, relativeRefractiveIndex );
}


UnitVector3f& UnitVector3f::refractEq
(
	const UnitVector3f& normal,
	      float         relativeRefractiveIndex
)
{
	const float a = vector_m.dot( normal.vector_m );
	if( a < 0.0f )
	{
		relativeRefractiveIndex = 1.0f / relativeRefractiveIndex;
	}

	const float b = 1.0f - ((relativeRefractiveIndex * relativeRefractiveIndex) * (1.0f - (a * a)));

	const bool totalInternalReflection = (b < 0.0f);
	if( !totalInternalReflection &&
	    !(this->isZero() || normal.isZero()) )
	{
		float c = float(::sqrt( b ));
		if( a < 0.0f )
		{
			c = -c;
		}

		vector_m *= -relativeRefractiveIndex;
		vector_m += (normal.vector_m * ((relativeRefractiveIndex * a) - c));
	}
	else
	{
		vector_m.negate();
		vector_m += (normal.vector_m * (a * 2.0f));
	}

	return *this;
}


bool UnitVector3f::isInTirZone
(
	const UnitVector3f& normal,
	const float         relativeRefractiveIndex
) const
{
	const float a = vector_m.dot( normal.vector_m );
	if( a < 0.0f )
	{
		relativeRefractiveIndex = 1.0f / relativeRefractiveIndex;
	}

	const float b = 1.0f - ((relativeRefractiveIndex * relativeRefractiveIndex) * (1.0f - (a * a)));

	return b < 0.0f;
}


UnitVector3f UnitVector3f::operator-() const
{
	return UnitVector3f( vector_m, false ).negate();
//	return UnitVector3f( vector_m.operator-(), false );
}




/// logical ----------------------------------------------------------------------------------------
bool UnitVector3f::isZero() const
{
	return bool( (vector_m.x_m == 0.0f) & (vector_m.y_m == 0.0f) & (vector_m.z_m == 0.0f) );
}




/// constants --------------------------------------------------------------------------------------
const UnitVector3f& UnitVector3f::ZERO()
{
	static const UnitVector3f k;
	return k;
}


const UnitVector3f& UnitVector3f::X()
{
	static const UnitVector3f k( Vector3f::X(), false );
	return k;
}


const UnitVector3f& UnitVector3f::Y()
{
	static const UnitVector3f k( Vector3f::Y(), false );
	return k;
}


const UnitVector3f& UnitVector3f::Z()
{
	static const UnitVector3f k( Vector3f::Z(), false );
	return k;
}




/// serialization ----------------------------------------------------------------------------------
const char* UnitVector3f::CLASS_NAME()
{
	static const char CLASS_NAME[] = "hxa7241::graphics::UnitVector3f";
	return CLASS_NAME;
}


const char* UnitVector3f::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a unitized vector of 3 floats";
	return CLASS_DESCRIPTION;
}




/// implementation ---------------------------------------------------------------------------------
void UnitVector3f::unitize()
{
	const float dotSelf	= (vector_m.x_m * vector_m.x_m) +
	                      (vector_m.y_m * vector_m.y_m) +
	                      (vector_m.z_m * vector_m.z_m);

	if( dotSelf != 1.0f )
	{
		const float length        = float(::sqrt( dotSelf ));
		const float oneOverLength = length != 0.0f ? 1.0f / length : 0.0f;

		vector_m.x_m *= oneOverLength;
		vector_m.y_m *= oneOverLength;
		vector_m.z_m *= oneOverLength;
	}
}
