/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "hxa7241math.h"
#include "Serialization.h"
#include <iostream>

#include "Vector4f.h"


using namespace hxa7241::graphics;




/// standard object services -----------------------------------------------------------------------
Vector4f::Vector4f()
{
	elements_m[0] = elements_m[1] = elements_m[2] = elements_m[3] = 0.0f;

//	elements_m[0] = 0.0f;
//	elements_m[1] = 0.0f;
//	elements_m[2] = 0.0f;
//	elements_m[3] = 0.0f;
}


Vector4f::Vector4f( const float x, const float y, const float z, const float w )
{
	setXYZW( x, y, z, w );
}


Vector4f::Vector4f( const float xyzw[4] )
{
	setXYZW( xyzw );
}


Vector4f::Vector4f( const Vector3f& v3, const float w )
{
	setXYZW( v3, w );
}


Vector4f::Vector4f( const Vector3f& v3 )
{
	setXYZ( v3 );
}


Vector4f& Vector4f::operator=( const Vector4f& other )
{
	elements_m[0] = other.elements_m[0];
	elements_m[1] = other.elements_m[1];
	elements_m[2] = other.elements_m[2];
	elements_m[3] = other.elements_m[3];

	return *this;
}




/// streaming --------------------------------------------------------------------------------------
std::ostream& hxa7241::graphics::operator<<( std::ostream& out, const Vector4f& obj )
{
	out << "(";

	int i = 0;
	do
	{
		out << obj.elements_m[i];
		if( i < 3 )
		{
			out << ", ";
		}
	}
	while( ++i < 4 );

	out << ")";

	return out;


	//return out << "(" << obj.elements_m[0] << ", " << obj.elements_m[1] << ", " << obj.elements_m[2] << ", " << obj.elements_m[3] << ")";
}


std::istream& hxa7241::graphics::operator>>( std::istream& in, Vector4f& obj )
{
	char c;

	int i = 0;
	do
	{
		in >> c >> obj.elements_m[i];
	}
	while( ++i < 4 );

	in >> c;

	return in;


	//char c;
	//return in >> c >> obj.elements_m[0] >> c >> obj.elements_m[1] >> c >> obj.elements_m[2] >> c >> obj.elements_m[3] >> c;
}




/// access -----------------------------------------------------------------------------------------
Vector4f& Vector4f::setXYZW( const float x,  const float y,  const float z,  const float w )
{
	elements_m[0] = x;
	elements_m[1] = y;
	elements_m[2] = z;
	elements_m[3] = w;

	return *this;
}


Vector4f& Vector4f::setXYZW( const float xyzw[4] )
{
	elements_m[0] = xyzw[0];
	elements_m[1] = xyzw[1];
	elements_m[2] = xyzw[2];
	elements_m[3] = xyzw[3];

	return *this;
}


Vector4f& Vector4f::setXYZW( const Vector3f& v3, const float w )
{
	setXYZ( v3 );
	elements_m[3] = w;

	return *this;
}


Vector4f& Vector4f::setXYZ( const Vector3f& v3 )
{
	elements_m[0] = v3.getX();
	elements_m[1] = v3.getY();
	elements_m[2] = v3.getZ();
	elements_m[3] = 1.0f;

	return *this;
}




void Vector4f::getXYZW( float& x,  float& y,  float& z,  float& w ) const
{
	x = elements_m[0];
	y = elements_m[1];
	z = elements_m[2];
	w = elements_m[3];
}


void Vector4f::getXYZW( float xyzw[4] ) const
{
	xyzw[0] = elements_m[0];
	xyzw[1] = elements_m[1];
	xyzw[2] = elements_m[2];
	xyzw[3] = elements_m[3];
}


void Vector4f::getXYZdivW( Vector3f& v3 ) const
{
	const float oneOverW = (elements_m[3] != 0.0f) ? 1.0f / elements_m[3] : 0.0f;

	v3.setXYZ( elements_m );
	v3 *= oneOverW;
}


Vector3f Vector4f::getXYZdivW() const
{
	const float oneOverW = (elements_m[3] != 0.0f) ? 1.0f / elements_m[3] : 0.0f;

	return Vector3f( elements_m[0] * oneOverW,
	                 elements_m[1] * oneOverW,
	                 elements_m[2] * oneOverW );
}




/// arithmetic -------------------------------------------------------------------------------------
float Vector4f::dot( const Vector4f& v ) const
{
	return (elements_m[0] * v.elements_m[0]) +
	       (elements_m[1] * v.elements_m[1]) +
	       (elements_m[2] * v.elements_m[2]) +
	       (elements_m[3] * v.elements_m[3]);
}


float Vector4f::distance( const Vector4f& v ) const
{
	const float xDif = elements_m[0] - v.elements_m[0];
	const float yDif = elements_m[1] - v.elements_m[1];
	const float zDif = elements_m[2] - v.elements_m[2];
	const float wDif = elements_m[3] - v.elements_m[3];

	return float(::sqrt( (xDif * xDif) +
	                     (yDif * yDif) +
	                     (zDif * zDif) +
	                     (wDif * wDif) ));

//	//const Vector4f dif( *this - v );
//	Vector4f dif( *this );
//	dif -= v;
//
//	return dif.length();
}


float Vector4f::length() const
{
	return float(::sqrt( (elements_m[0] * elements_m[0]) +
	                     (elements_m[1] * elements_m[1]) +
	                     (elements_m[2] * elements_m[2]) +
	                     (elements_m[3] * elements_m[3]) ));
}


float Vector4f::total() const
{
	return elements_m[0] + elements_m[1] + elements_m[2] + elements_m[3];
}


float Vector4f::average() const
{
	return (elements_m[0] + elements_m[1] + elements_m[2] + elements_m[3]) * 0.25f;
}


float Vector4f::smallest() const
{
	float smallest;
	smallest = elements_m[0] <= elements_m[1] ? elements_m[0] : elements_m[1];
	smallest = smallest      <= elements_m[2] ? smallest      : elements_m[2];
	smallest = smallest      <= elements_m[3] ? smallest      : elements_m[3];

	return smallest;
}


float Vector4f::largest() const
{
	float largest;
	largest = elements_m[0] >= elements_m[1] ? elements_m[0] : elements_m[1];
	largest = largest       >= elements_m[2] ? largest       : elements_m[2];
	largest = largest       >= elements_m[3] ? largest       : elements_m[3];

	return largest;
}


Vector4f Vector4f::unitized() const
{
	const float length = float(::sqrt( (elements_m[0] * elements_m[0]) +
	                                   (elements_m[1] * elements_m[1]) +
	                                   (elements_m[2] * elements_m[2]) +
	                                   (elements_m[3] * elements_m[3]) ));
	const float oneOverLength = (length > 0.0f) ? 1.0f / length : 0.0f;

	return Vector4f( elements_m[0] * oneOverLength,
					 elements_m[1] * oneOverLength,
					 elements_m[2] * oneOverLength,
					 elements_m[3] * oneOverLength );
}


Vector4f& Vector4f::unitizeEq()
{
	const float length = float(::sqrt( (elements_m[0] * elements_m[0]) +
	                                   (elements_m[1] * elements_m[1]) +
	                                   (elements_m[2] * elements_m[2]) +
	                                   (elements_m[3] * elements_m[3]) ));
	const float oneOverLength = (length > 0.0f) ? 1.0f / length : 0.0f;

	elements_m[0] *= oneOverLength;
	elements_m[1] *= oneOverLength;
	elements_m[2] *= oneOverLength;
	elements_m[3] *= oneOverLength;

	return *this;
}


//Vector4f Vector4f::cross( const Vector4f& v ) const
//{
//	return Vector4f( ,
//	                 ,
//	                 ,
//	                  );
//}
//
//
//Vector4f& Vector4f::crossEq( const Vector4f& v )
//{
//	const float x = ;
//	const float y = ;
//	const float z = ;
//	const float w = ;
//
//	elements_m[0] = x;
//	elements_m[1] = y;
//	elements_m[2] = z;
//	elements_m[3] = w;
//
//	return *this;
//}


Vector4f Vector4f::abs() const
{
	return Vector4f( *this ).absEq();
}


Vector4f& Vector4f::absEq()
{
	int i = 3;
	do
	{
		if( elements_m[i] < 0.0f )   /// doesnt catch -0.0  but does that matter ?
		{
			elements_m[i] = -elements_m[i];
		}
	} while( i-- > 0 );

	return *this;
}




/// plus minus -------------------------------------------------------------------------------------
Vector4f Vector4f::operator-() const
{
	return Vector4f( -elements_m[0],
	                 -elements_m[1],
	                 -elements_m[2],
	                 -elements_m[3] );
}


Vector4f& Vector4f::negate()
{
	elements_m[0] = -elements_m[0];
	elements_m[1] = -elements_m[1];
	elements_m[2] = -elements_m[2];
	elements_m[3] = -elements_m[3];

	return *this;
}


Vector4f Vector4f::operator+( const Vector4f& v ) const
{
	return Vector4f( elements_m[0] + v.elements_m[0],
	                 elements_m[1] + v.elements_m[1],
	                 elements_m[2] + v.elements_m[2],
	                 elements_m[3] + v.elements_m[3] );
}


Vector4f& Vector4f::operator+=( const Vector4f& v )
{
	elements_m[0] += v.elements_m[0];
	elements_m[1] += v.elements_m[1];
	elements_m[2] += v.elements_m[2];
	elements_m[3] += v.elements_m[3];

	return *this;
}


Vector4f Vector4f::operator-( const Vector4f& v ) const
{
	return Vector4f( elements_m[0] - v.elements_m[0],
	                 elements_m[1] - v.elements_m[1],
	                 elements_m[2] - v.elements_m[2],
	                 elements_m[3] - v.elements_m[3] );
}


Vector4f& Vector4f::operator-=( const Vector4f& v )
{
	elements_m[0] -= v.elements_m[0];
	elements_m[1] -= v.elements_m[1];
	elements_m[2] -= v.elements_m[2];
	elements_m[3] -= v.elements_m[3];

	return *this;
}




/// mult div ---------------------------------------------------------------------------------------
Vector4f Vector4f::operator*( const Vector4f& v ) const
{
	return Vector4f( elements_m[0] * v.elements_m[0],
	                 elements_m[1] * v.elements_m[1],
	                 elements_m[2] * v.elements_m[2],
	                 elements_m[3] * v.elements_m[3] );
}


Vector4f& Vector4f::operator*=( const Vector4f& v )
{
	elements_m[0] *= v.elements_m[0];
	elements_m[1] *= v.elements_m[1];
	elements_m[2] *= v.elements_m[2];
	elements_m[3] *= v.elements_m[3];

	return *this;
}


Vector4f Vector4f::operator/( const Vector4f& v ) const
{
	return Vector4f( elements_m[0] / v.elements_m[0],
	                 elements_m[1] / v.elements_m[1],
	                 elements_m[2] / v.elements_m[2],
	                 elements_m[3] / v.elements_m[3] );
}


Vector4f& Vector4f::operator/=( const Vector4f& v )
{
	elements_m[0] /= v.elements_m[0];
	elements_m[1] /= v.elements_m[1];
	elements_m[2] /= v.elements_m[2];
	elements_m[3] /= v.elements_m[3];

	return *this;
}




Vector4f Vector4f::operator*( const float f ) const
{
	return Vector4f( elements_m[0] * f,
	                 elements_m[1] * f,
	                 elements_m[2] * f,
	                 elements_m[3] * f );
}


Vector4f& Vector4f::operator*=( const float f )
{
	elements_m[0] *= f;
	elements_m[1] *= f;
	elements_m[2] *= f;
	elements_m[3] *= f;

	return *this;
}


Vector4f Vector4f::operator/( const float f ) const
{
	const float oneOverF = 1.0f / f;

	return Vector4f( elements_m[0] * oneOverF,
	                 elements_m[1] * oneOverF,
	                 elements_m[2] * oneOverF,
	                 elements_m[3] * oneOverF );
}


Vector4f& Vector4f::operator/=( const float f )
{
	const float oneOverF = 1.0f / f;

	elements_m[0] *= oneOverF;
	elements_m[1] *= oneOverF;
	elements_m[2] *= oneOverF;
	elements_m[3] *= oneOverF;

	return *this;
}




/// pow --------------------------------------------------------------------------------------------
Vector4f Vector4f::operator^( const float f ) const
{
	return Vector4f( *this ) ^= f;
}


Vector4f& Vector4f::operator^=( const float f )
{
	int i = 3;
	do
	{
		elements_m[i] = float(::pow( elements_m[i], f ));
	} while( i-- > 0 );

	return *this;
}


Vector4f Vector4f::operator^( const Vector4f& v ) const
{
	return Vector4f( *this ) ^= v;
}


Vector4f& Vector4f::operator^=( const Vector4f& v )
{
	int i = 3;
	do
	{
		elements_m[i] = float(::pow( elements_m[i], v.elements_m[i] ));
	} while( i-- > 0 );

	return *this;
}




/// friends ----------------------------------------------------------------------------------------
Vector4f hxa7241::graphics::operator*( const float f, const Vector4f& v )
{
	return Vector4f( f * v.elements_m[0],
	                 f * v.elements_m[1],
	                 f * v.elements_m[2],
	                 f * v.elements_m[3] );
}




Vector4f hxa7241::graphics::operator/( const float f, const Vector4f& v )
{
	return Vector4f( f / v.elements_m[0],
	                 f / v.elements_m[1],
	                 f / v.elements_m[2],
	                 f / v.elements_m[3] );
}




/// logical ---------------------------------------------------------------------------------------
bool Vector4f::operator==( const Vector4f& v ) const
{
	return (elements_m[0] == v.elements_m[0]) &
	       (elements_m[1] == v.elements_m[1]) &
	       (elements_m[2] == v.elements_m[2]) &
	       (elements_m[3] == v.elements_m[3]);
}


bool Vector4f::operator!=( const Vector4f& v ) const
{
	return (elements_m[0] != v.elements_m[0]) |
	       (elements_m[1] != v.elements_m[1]) |
	       (elements_m[2] != v.elements_m[2]) |
	       (elements_m[3] != v.elements_m[3]);
}


bool Vector4f::isZero() const
{
	return bool( (elements_m[0] == 0.0f) &
	             (elements_m[1] == 0.0f) &
	             (elements_m[2] == 0.0f) &
	             (elements_m[3] == 0.0f) );
}




/// clamps -----------------------------------------------------------------------------------------
Vector4f& Vector4f::clampToMinOf( const Vector4f& min )
{
	int i = 3;
	do
	{
		if( elements_m[i] < min.elements_m[i] )
		{
			elements_m[i] = min.elements_m[i];
		}
	} while( i-- > 0 );

	return *this;
}


Vector4f& Vector4f::clampToMaxOf( const Vector4f& max )
{
	int i = 3;
	do
	{
		if( elements_m[i] > max.elements_m[i] )
		{
			elements_m[i] = max.elements_m[i];
		}
	} while( i-- > 0 );

	return *this;
}


Vector4f& Vector4f::clampBetween( const Vector4f& min, const Vector4f& max )
{
	int i = 3;
	do
	{
		if( elements_m[i] > max.elements_m[i] )
		{
			elements_m[i] = max.elements_m[i];
		}
		else
		if( elements_m[i] < min.elements_m[i] )
		{
			elements_m[i] = min.elements_m[i];
		}
	} while( i-- > 0 );

	return *this;
}


Vector4f& Vector4f::clamp01()   /// [0,1)  0 to almost 1
{
	int i = 3;
	do
	{
		if( elements_m[i] >= 1.0f )
		{
			elements_m[i]  = FLOAT_ALMOST_ONE;
		}
		else
		if( elements_m[i] < 0.0f )
		{
			elements_m[i] = 0.0f;
		}
	} while( i-- > 0 );

	return *this;
}


void Vector4f::clamp01( Vector4f& result ) const   /// [0,1)  0 to almost 1
{
	result = *this;
	result.clamp01();
}




/// constants --------------------------------------------------------------------------------------
const Vector4f& Vector4f::ZERO()
{
	static const Vector4f k( 0.0f, 0.0f, 0.0f, 0.0f );
	return k;
}


const Vector4f& Vector4f::HALF()
{
	static const Vector4f k( 0.5f, 0.5f, 0.5f, 0.5f );
	return k;
}


const Vector4f& Vector4f::ONE()
{
	static const Vector4f k( 1.0f, 1.0f, 1.0f, 1.0f );
	return k;
}


const Vector4f& Vector4f::EPSILON()
{
	static const Vector4f k( FLOAT_EPSILON, FLOAT_EPSILON, FLOAT_EPSILON, FLOAT_EPSILON );
	return k;
}


const Vector4f& Vector4f::ALMOST_ONE()
{
	static const Vector4f k( FLOAT_ALMOST_ONE, FLOAT_ALMOST_ONE, FLOAT_ALMOST_ONE, FLOAT_ALMOST_ONE );
	return k;
}


const Vector4f& Vector4f::MIN()
{
	static const Vector4f k( FLOAT_MIN_NEG, FLOAT_MIN_NEG, FLOAT_MIN_NEG, FLOAT_MIN_NEG );
	return k;
}


const Vector4f& Vector4f::MAX()
{
	static const Vector4f k( FLOAT_MAX, FLOAT_MAX, FLOAT_MAX, FLOAT_MAX );
	return k;
}


const Vector4f& Vector4f::X()
{
	static const Vector4f k( 1.0f, 0.0f, 0.0f, 0.0f );
	return k;
}


const Vector4f& Vector4f::Y()
{
	static const Vector4f k( 0.0f, 1.0f, 0.0f, 0.0f );
	return k;
}


const Vector4f& Vector4f::Z()
{
	static const Vector4f k( 0.0f, 0.0f, 1.0f, 0.0f );
	return k;
}


const Vector4f& Vector4f::W()
{
	static const Vector4f k( 0.0f, 0.0f, 0.0f, 1.0f );
	return k;
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& hxa7241::graphics::operator<<( SerializerOut& out, const Vector4f& obj )
{
	static const char* names[] = { "x", "y", "z", "w" };

	int i = 0;
	do
	{
		out.write( names[i], obj.elements_m[i] );
	} while( ++i < 4 );

	return out;
}


SerializerIn& hxa7241::graphics::operator>>( SerializerIn& in, Vector4f& obj )
{
	static const char* names[] = { "x", "y", "z", "w" };

	int i = 0;
	do
	{
		in.read( names[i], obj.elements_m[i] );
	} while( ++i < 4 );

	return in;
}


const char* Vector4f::CLASS_NAME()
{
	static const char CLASS_NAME[] = "hxa7241::graphics::Vector4f";
	return CLASS_NAME;
}


const char* Vector4f::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a vector of 4 floats";
	return CLASS_DESCRIPTION;
}
