/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Exception.h"
#include "Logger.h"
#include "XmlSerializerOut.h"
#include "XmlSerializerIn.h"
#include "ByteSerializerOut.h"
#include "ByteSerializerIn.h"

#include "RenderDataFiling.h"


using namespace perceptuum2::control;




/// constants --------------------------------------------------------------------------------------
const char RenderDataFiling::SCENE_FILE_EXTENSION[]   = ".scene.p2.xml";
const char RenderDataFiling::OPTIONS_FILE_EXTENSION[] = ".options.p2.xml";
const char RenderDataFiling::STATE_FILE_EXTENSION[]   = ".state.p2";
const char RenderDataFiling::IMAGE_FILE_EXTENSION[]   = ".image.p2";
const char RenderDataFiling::STATUS_FILE_EXTENSION[]  = ".status.p2.xml";

const char RenderDataFiling::CLASS_NAME[]             = "perceptuum2::control::RenderDataFiling";
const char RenderDataFiling::EXCEPTION_MESSAGE1[]     = "the options file (named something.options.p2.xml) doesnt exist";
const char RenderDataFiling::EXCEPTION_MESSAGE2[]     = "the scene file (named something.scene.p2.xml) doesnt exist";
const char RenderDataFiling::EXCEPTION_MESSAGE3[]     = "the state.p2 or image.p2 file is missing. find the missing one and put it with other, or delete the remaining one to start again";
const char RenderDataFiling::EXCEPTION_MESSAGE4[]     = "image.p2 and state.p2 files dont match, delete both to start again";




/// standard object services -----------------------------------------------------------------------
RenderDataFiling::RenderDataFiling
(
	const string&        filesPathname,
	const ThreadFactory& threadFactory
)
 :	RenderData(),
	filesPathname_m( filesPathname ),
	stateFilePathname_m(),
	imageFilePathname_m(),
	statusFilePathname_m(),
	stateFile_m(),
	imageFile_m(),
	options_m(),
	scene_m(),
	state_m(),
	image_m(),
	status_m(),
	pThread_m( threadFactory.createThread( this ) ),
	pStateCopy_m( 0 ),
	pImageCopy_m( 0 ),
	pStatusCopy_m( 0 )
{
	syncFilesAndObjects( filesPathname );

	setOptions( &options_m );
	setScene( &scene_m );
	setState( &state_m );
	setImage( &image_m );
	setStatus( &status_m );
}


RenderDataFiling::~RenderDataFiling()
{
	using hxa7241::utility::Logger;

	try
	{
		pThread_m->join();
	}
	catch( const std::exception& e )
	{
		Logger() << Logger::EXC2 << e.what();
	}
	catch( ... )
	{
		hxa7241::utility::Exception e;
		Logger() << Logger::EXC2 << e.what();
	}

	delete pThread_m;
}




/// queries --------------------------------------------------------------------
void RenderDataFiling::getFilesPathname( string& name ) const
{
	name = filesPathname_m;
}


void RenderDataFiling::getStartFileNames( vector<string>& names )
{
	names.resize( 2 );
	names[0] = SCENE_FILE_EXTENSION + 1;
	names[1] = OPTIONS_FILE_EXTENSION + 1;
}




/// commands -------------------------------------------------------------------
void RenderDataFiling::save()
{
	pThread_m->join();

	/// needs to copy objects, then spawn a thread to save the copy, in order to be synchronized
	/// make sync copies
	pStateCopy_m  = new RenderingState( state_m );
	pImageCopy_m  = new ImageCieXyz( image_m );
	pStatusCopy_m = new RenderingStatus( status_m );

	pThread_m->start();


//	run();   /// or do it synchronously (and comment out all thread stuff)
}


void RenderDataFiling::run()
{
	saveState();
	saveImage();
	saveStatus();

	/// delete sync copies
	delete pStateCopy_m;    pStateCopy_m  = 0;
	delete pImageCopy_m;    pImageCopy_m  = 0;
	delete pStatusCopy_m;   pStatusCopy_m = 0;
}


void RenderDataFiling::saveState()
{
	stateFile_m.seekp( 0 );
	hxa7241::utility::ByteSerializerOut::serializeOut( state_m, stateFile_m );
	stateFile_m.flush();
	hxa7241::utility::Logger() << "saved state render data";
}


void RenderDataFiling::saveImage()
{
	imageFile_m.seekp( 0 );
	hxa7241::utility::ByteSerializerOut::serializeOut( image_m, imageFile_m );
	imageFile_m.flush();
	hxa7241::utility::Logger() << "saved image render data";
}


void RenderDataFiling::saveStatus()
{
	hxa7241::utility::XmlSerializerOut::serializeOut( status_m, statusFilePathname_m.c_str() );
	hxa7241::utility::Logger() << "saved status render data";
}




/// implementation ---------------------------------------------------------------------------------
void RenderDataFiling::syncFilesAndObjects( const string& filesPathname )
{
	syncReadFilesAndObjects( filesPathname, options_m, scene_m );

	syncWriteFilesAndObjects( filesPathname );
	syncStatusFileAndObject( filesPathname );

	save();
}


void RenderDataFiling::syncReadFilesAndObjects
(
	const string&     filesPathname,
	RenderingOptions& options,
	Scene&            scene
)
{
	static const char METHOD_NAME[] = "syncReadFilesAndObjects";

	hxa7241::utility::Logger() << "synchronizing options and scene render data";

	using hxa7241::utility::XmlSerializerIn;

	{
		const string optionsFilePathname( filesPathname + OPTIONS_FILE_EXTENSION );
		ifstream optionsFile( optionsFilePathname.c_str() );
		if( !optionsFile )
		{
			throw hxa7241::utility::Exception( EXCEPTION_MESSAGE1, CLASS_NAME, METHOD_NAME );
		}
		XmlSerializerIn::serializeIn( options, optionsFile, optionsFilePathname.c_str() );
	}

	{
		const string sceneFilePathname( filesPathname + SCENE_FILE_EXTENSION );
		ifstream sceneFile( sceneFilePathname.c_str() );
		if( !sceneFile )
		{
			throw hxa7241::utility::Exception( EXCEPTION_MESSAGE2, CLASS_NAME, METHOD_NAME );
		}
		scene.setLightsourceEmittanceThreshold( options.getLightsourceEmittanceThreshold() );
		XmlSerializerIn::serializeIn( scene, sceneFile, sceneFilePathname.c_str() );
	}
}


void RenderDataFiling::syncWriteFilesAndObjects
(
	const string& filesPathname
)
{
	static const char METHOD_NAME[] = "syncWriteFilesAndObjects";

	/// open-read state
	stateFilePathname_m = filesPathname + STATE_FILE_EXTENSION;
	ifstream stateFile( stateFilePathname_m.c_str(), ifstream::binary );

	/// open-read image
	imageFilePathname_m = filesPathname + IMAGE_FILE_EXTENSION;
	ifstream imageFile( imageFilePathname_m.c_str(), ifstream::binary );

	/// check existence
	if( !( !!stateFile | !!imageFile ) )
	{
		/// both absent
		syncForStartingRender( stateFile, imageFile );
	}
	else if( !!stateFile & !!imageFile )
	{
		/// both present
		syncForContinuingRender( stateFile, imageFile );
	}
	else
	{
		/// invalid (one or other missing)
		throw hxa7241::utility::Exception( EXCEPTION_MESSAGE3, CLASS_NAME, METHOD_NAME );
	}

	/// open-write state
	stateFile_m.open( stateFilePathname_m.c_str(), ifstream::binary );

	/// open-write image
	imageFile_m.open( imageFilePathname_m.c_str(), ifstream::binary );
}


void RenderDataFiling::syncForStartingRender
(
	ifstream& stateFile,
	ifstream& imageFile
)
{
	hxa7241::utility::Logger() << "synchronizing state and image render data for starting render";

	/// close both
	stateFile.close();
	imageFile.close();

	/// set objects using options
	{
		state_m.setOptions( options_m.getImmutableOptions() );

		const dword width  = options_m.getImageWidth();
		const dword height = options_m.getImageHeight();
		image_m.setSize( width, height );
		state_m.setImage( &image_m );
	}
}


void RenderDataFiling::syncForContinuingRender
(
	ifstream& stateFile,
	ifstream& imageFile
)
{
	static const char METHOD_NAME[] = "syncForContinuingRender";

	hxa7241::utility::Logger() << "synchronizing state and image render data for continuing render";

	/// read objects from files
	using hxa7241::utility::ByteSerializerIn;
	ByteSerializerIn::serializeIn( state_m, stateFile, stateFilePathname_m.c_str() );
	ByteSerializerIn::serializeIn( image_m, imageFile, imageFilePathname_m.c_str() );

	/// close both
	stateFile.close();
	imageFile.close();

	/// check image size and state size match
	if( (image_m.getWidth() != state_m.getImageWidth()) || (image_m.getHeight() != state_m.getImageHeight()) )
	{
		throw hxa7241::utility::Exception( EXCEPTION_MESSAGE4, CLASS_NAME, METHOD_NAME );
	}

	/// re-link image to state
	{
		state_m.setImage( &image_m );
	}

	/// set immutable choices from state back to options
	{
		const RenderingOptionsImmutable& roi = state_m.getImmutableOptions();
		options_m.setImmutableOptions( roi );
	}
}


void RenderDataFiling::syncStatusFileAndObject( const string& filesPathname )
{
	/// open-read status
	statusFilePathname_m = filesPathname + STATUS_FILE_EXTENSION;
	ifstream statusFile( statusFilePathname_m.c_str() );

	if( !!statusFile )
	{
		/// read object from file
		using hxa7241::utility::XmlSerializerIn;
		XmlSerializerIn::serializeIn( status_m, statusFile, statusFilePathname_m.c_str() );
	}
}
