/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Serialization.h"

#include "DisplayMeasures.h"   /// own header is included last


using namespace perceptuum2::imaging;




/// statics
/// average crt
const CieXyz   DisplayMeasures::BLACKPOINT_DEFAULT(           CieXyz( 0.951f, 1.00f, 1.089f ) *=  2.5f );
const CieXyz   DisplayMeasures::WHITEPOINT_DEFAULT(           CieXyz( 0.951f, 1.00f, 1.089f ) *= 85.0f );
const Vector3f DisplayMeasures::CHROMATICITIES_DEFAULT[ 3 ] =
	                                                        { Vector3f( 0.61f, 0.29f, 0.150f ),     /// x
	                                                          Vector3f( 0.35f, 0.59f, 0.063f ),     /// y
	                                                          Vector3f( 0.04f, 0.12f, 0.787f ) };   /// z
const Vector3f DisplayMeasures::GAMMA_DEFAULT( 2.222f, 2.222f, 2.222f );   /// CCIR rec.709 (HDTV) standardises coding with a power function of 0.45




/// standard object services -----------------------------------------------------------------------
DisplayMeasures::DisplayMeasures()
 :	blackPoint_m(),
	whitePoint_m(),
	chromaticitiesXyzX_m(),
	chromaticitiesXyzY_m(),
	chromaticitiesXyzZ_m(),
	gammaXyz_m()
{
	DisplayMeasures::setToDefault();
}


DisplayMeasures::~DisplayMeasures()
{
}


DisplayMeasures::DisplayMeasures( const DisplayMeasures& other )
{
	DisplayMeasures::operator=( other );
}


DisplayMeasures& DisplayMeasures::operator=( const DisplayMeasures& other )
{
	if( &other != this )
	{
		blackPoint_m         = other.blackPoint_m;
		whitePoint_m         = other.whitePoint_m;
		chromaticitiesXyzX_m = other.chromaticitiesXyzX_m;
		chromaticitiesXyzY_m = other.chromaticitiesXyzY_m;
		chromaticitiesXyzZ_m = other.chromaticitiesXyzZ_m;
		gammaXyz_m           = other.gammaXyz_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void DisplayMeasures::setToDefault()
{
	blackPoint_m         = BLACKPOINT_DEFAULT;
	whitePoint_m         = WHITEPOINT_DEFAULT;
	chromaticitiesXyzX_m = CHROMATICITIES_DEFAULT[0];
	chromaticitiesXyzY_m = CHROMATICITIES_DEFAULT[1];
	chromaticitiesXyzZ_m = CHROMATICITIES_DEFAULT[2];
	gammaXyz_m           = GAMMA_DEFAULT;
}




/// queries ----------------------------------------------------------------------------------------
void DisplayMeasures::getBlackPoint( CieXyz& c ) const
{
	c = blackPoint_m;
}


void DisplayMeasures::getWhitePoint( CieXyz& c ) const
{
	c = whitePoint_m;
}


void DisplayMeasures::getChromaticitiesXyzX( Vector3f& c ) const
{
	c = chromaticitiesXyzX_m;
}


void DisplayMeasures::getChromaticitiesXyzY( Vector3f& c ) const
{
	c = chromaticitiesXyzY_m;
}


void DisplayMeasures::getChromaticitiesXyzZ( Vector3f& c ) const
{
	c = chromaticitiesXyzZ_m;
}


void DisplayMeasures::getChromaticitiesXyz( Vector3f c[3] ) const
{
	c[0] = chromaticitiesXyzX_m;
	c[1] = chromaticitiesXyzY_m;
	c[2] = chromaticitiesXyzZ_m;
}


void DisplayMeasures::getGammaXyz( Vector3f& c ) const
{
	c = gammaXyz_m;
}




/// implementation ---------------------------------------------------------------------------------
void DisplayMeasures::enforceConstraints()
{
	blackPoint_m.clampToMinOf( CieXyz::ZERO() );
	whitePoint_m.clampToMinOf( CieXyz::ZERO() );

	chromaticitiesXyzX_m.clampToMinOf( CieXyz::ZERO() );
	chromaticitiesXyzY_m.clampToMinOf( CieXyz::ZERO() );
	chromaticitiesXyzZ_m.clampToMinOf( CieXyz::ZERO() );

	gammaXyz_m.clampToMinOf( CieXyz::ZERO() );
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::imaging::operator<<( SerializerOut& out, const DisplayMeasures& obj )
{
	out.write( "blackPoint",         obj.blackPoint_m,         "CIE XYZ" );
	out.write( "whitePoint",         obj.whitePoint_m,         "CIE XYZ" );
	out.write( "chromaticitiesXyzX", obj.chromaticitiesXyzX_m, "CIE xyz" );
	out.write( "chromaticitiesXyzY", obj.chromaticitiesXyzY_m, "CIE xyz" );
	out.write( "chromaticitiesXyzZ", obj.chromaticitiesXyzZ_m, "CIE xyz" );
	out.write( "gammaXyz",           obj.gammaXyz_m,           "CIE xyz" );

	return out;
}


SerializerIn& perceptuum2::imaging::operator>>( SerializerIn& in, DisplayMeasures& obj )
{
	in.read( "blackPoint",         obj.blackPoint_m         );
	in.read( "whitePoint",         obj.whitePoint_m         );
	in.read( "chromaticitiesXyzX", obj.chromaticitiesXyzX_m );
	in.read( "chromaticitiesXyzY", obj.chromaticitiesXyzY_m );
	in.read( "chromaticitiesXyzZ", obj.chromaticitiesXyzZ_m );
	in.read( "gammaXyz",           obj.gammaXyz_m           );

	obj.enforceConstraints();

	return in;
}


const char* DisplayMeasures::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::imaging::DisplayMeasures";
	return CLASS_NAME;
}


const char* DisplayMeasures::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a precise description of the display that the application will use to show rendering progress";
	return CLASS_DESCRIPTION;
}
