/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "ImageCieXyz.h"
#include "CieXyz.h"
#include "Float01.h"
#include "FpToInt.h"
#include "Exception.h"
#include "Serialization.h"
#include "hxa7241math.h"
#include "hxa7241vector.h"

#include "ImageAccumulation.h"   /// own header is included last


using namespace perceptuum2::imaging;




/// standard object services -----------------------------------------------------------------------
ImageAccumulation::ImageAccumulation()
 :	pImage_m              ( 0 ),
	pixelStats_m          (),
	requiredSamples_m     (),
	isAnySamplesRequired_m( false ),
	samplesCount_m        ( 0 ),
	samplesRequiredCount_m( 0 ),
	completedRatio_m      ( 0.0f )
{
}


ImageAccumulation::~ImageAccumulation()
{
}


ImageAccumulation::ImageAccumulation( const ImageAccumulation& other )
 :	pImage_m              ( other.pImage_m ),
	pixelStats_m          ( other.pixelStats_m ),
	requiredSamples_m     ( other.requiredSamples_m ),
	isAnySamplesRequired_m( other.isAnySamplesRequired_m ),
	samplesCount_m        ( other.samplesCount_m ),
	samplesRequiredCount_m( other.samplesRequiredCount_m ),
	completedRatio_m      ( other.completedRatio_m )
{
}


ImageAccumulation& ImageAccumulation::operator=( const ImageAccumulation& other )
{
	if( &other != this )
	{
		pImage_m               = other.pImage_m;
		pixelStats_m           = other.pixelStats_m;
		hxa7241::utility::assignUnreserve( other.requiredSamples_m, requiredSamples_m );
		isAnySamplesRequired_m = other.isAnySamplesRequired_m;

		samplesCount_m         = other.samplesCount_m;
		samplesRequiredCount_m = other.samplesRequiredCount_m;
		completedRatio_m       = other.completedRatio_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void ImageAccumulation::setImage
(
	ImageCieXyz* pImage
)
{
	if( pImage == 0 )
	{
		throw hxa7241::utility::Exception( "null pointer to image given", CLASS_NAME(), "setImage" );
	}
	pImage_m = pImage;

	pixelStats_m.setSize( pImage_m->getWidth(), pImage_m->getHeight() );

	samplesCount_m = 0;
	hxa7241::utility::resizeReserve( requiredSamples_m, pImage_m->getLength() );
	analyseSamplesRequired( *pImage_m, pixelStats_m, requiredSamples_m, isAnySamplesRequired_m, samplesRequiredCount_m, completedRatio_m );
}


bool ImageAccumulation::shouldSample
(
	const dword x,
	const dword y
)
{
	static const char METHOD_NAME[] = "shouldSample";

	bool should = false;

	if( checkSheets( METHOD_NAME ) )
	{
		const dword width = pImage_m->getWidth();
		should = requiredSamples_m[ x + (y * width) ];
	}

	return should;
}


void ImageAccumulation::addSample
(
	const dword   x,
	const dword   y,
	const float   ,//xOffset01,
	const float   ,//yOffset01,
	const CieXyz& xyz,
	const float   alpha
)
{
	static const char METHOD_NAME[] = "addSample";

	if( checkSheets( METHOD_NAME ) )
	{
		/// get old mean (= pixel color) and stats
		CieXyz&     mean  = pImage_m->getElement( x, y );
		PixelStats& stats = pixelStats_m.getElement( x, y );

		/// set scalings for new and old
		const float count    = stats.getCount();
		const float newScale = 1.0f / (count + 1.0f);
		const float oldScale = newScale * count;//count / (count + 1.0f);

		/// calc and set new variance or maxdeviation
		{
			/// (first sample has nothing to compare to, so make difference zero)
			const float difference = (xyz.getY() - mean.getY()) * float(count > 0.0f);
			const float deviation  = difference * difference;

			/// variance
			//float variance = stats.getVariance();
			//variance *= oldScale;
			//variance += (deviation * newScale);
			//stats.setVariance( variance );

			/// maximum deviation
			float maxDeviation = stats.getVariance();
			maxDeviation = maxDeviation >= deviation ? maxDeviation : deviation;
			stats.setVariance( maxDeviation );
		}

		/// accumulate new mean
		{
			mean *= oldScale;
			mean += (xyz * newScale);
		}

		/// accumulate new alpha mean
		{
			byte& alphaRef = pImage_m->getAlphaElement( x, y );

			float alphaMean = hxa7241::utility::ubyteToFp( ubyte(alphaRef) );
			alphaMean *= oldScale;
			alphaMean += (hxa7241::utility::Float01::clamp01(alpha) * newScale);

			dword alphaMeanByte = hxa7241::utility::fpToIntTrunc( alphaMean * 256.0f );
			if( alphaMeanByte > 255 )
			{
				alphaMeanByte = 255;
			}
			alphaRef = byte(alphaMeanByte);
		}

		/// increment count
		stats.incCount();
		++samplesCount_m;
	}


	/*/// possible unassociated alpha version...
	if( checkSheets( METHOD_NAME ) )
	{
		/// get old mean (= pixel color)
		CieXyz& mean = pImage_m->getElement( x, y );

		/// get old alpha mean
		static const float ONE_OVER_256 = 1.0f / 256.0f;
		byte& alphaMeanRef    = pImage_m->getAlphaElement( x, y );
		const float alphaMean = float(ubyte( alphaMeanRef )) * ONE_OVER_256;

		/// get old stats
		PixelStats& stats = pixelStats_m.getElement( x, y );
		const float count = stats.getCount();

		/// calc and set new variance or maxdeviation
		{
			/// (first sample has nothing to compare to, so make difference zero)
			const float difference = ((xyz.getY() * alpha) - (mean.getY() * alphaMean) * float(count > 0.0f);
			const float deviation  = difference * difference;

			/// maximise deviation
			if( deviation > stats.getVariance() )
			{
				stats.setVariance( deviation );
			}
		}

		/// calc and set new mean
		{
			/// new colmean = (colmean * (alphatotal / new alphatotal)) + (new col * (new alpha / new alphatotal))
			/// alpha total = alpha mean * pixel count
			/// new alpha total = alpha total + new alpha
			const float alphaCount           = alphaMean * count;   /// count must be of all samples, not just alpha == 1
			const float oneOverNewAlphaCount = 1.0f / (alphaCount + alpha);
			const float oldScale             = alphaCount * oneOverNewAlphaCount;
			const float newScale             = alpha * oneOverNewAlphaCount;

			mean = (mean * oldScale) += (xyz * newScale);
		}

		/// calc and set new alpha mean
		{
			const float newScale = 1.0f / (count + 1.0f);
			const float oldScale = count * newScale;//count / (count + 1.0f);

			const Float01 newAlphaMean( (alphaMean * oldScale) + (hxa7241::utility::Float01::clamp01(alpha) * newScale) );

			alphaMeanRef = byte(hxa7241::utility::fpToIntTrunc( newAlphaMean * 256.0f ));
		}

		/// increment count
		stats.incCount();
		++samplesCount_m;
	}*/
}


void ImageAccumulation::analyse()
{
	static const char METHOD_NAME[] = "analyse";

	if( checkSheets( METHOD_NAME ) )
	{
		analyseSpatialDeviation( *pImage_m, pixelStats_m );
		analyseSamplesRequired( *pImage_m, pixelStats_m, requiredSamples_m, isAnySamplesRequired_m, samplesRequiredCount_m, completedRatio_m );
	}
}




/// queries ----------------------------------------------------------------------------------------
dword ImageAccumulation::getPixelCount() const
{
	return pixelStats_m.getLength();
}


dword ImageAccumulation::getWidth() const
{
	return pixelStats_m.getWidth();
}


dword ImageAccumulation::getHeight() const
{
	return pixelStats_m.getHeight();
}


dword ImageAccumulation::getSamplesCount() const
{
	return samplesCount_m;
}


dword ImageAccumulation::getSamplesRequiredCount() const
{
	return samplesRequiredCount_m;
}


float ImageAccumulation::getCompletedRatio() const
{
	return completedRatio_m;

//	float ratio = 0.0f;
//	if( samplesRequiredCount_m > 0 )
//	{
//		ratio = float(samplesCount_m) / float(samplesRequiredCount_m);
//	}
//
//	return ratio;
}


bool ImageAccumulation::isCompleted() const
{
	return !isAnySamplesRequired_m;
}


float ImageAccumulation::getMaxDeviation() const
{
	float maxVariance = 0.0f;
	{
		const PixelStats* pStat = pixelStats_m.getMemory();
		const PixelStats* pEnd  = pStat + pixelStats_m.getLength();
		for( ;  pStat < pEnd;  ++pStat )
		{
			const float variance = pStat->getVariance();
			if( maxVariance < variance )
			{
				maxVariance = variance;
			}
		}
	}

	return float(::sqrt( maxVariance ));
}




/// implementation ---------------------------------------------------------------------------------
void ImageAccumulation::analyseSpatialDeviation
(
	const ImageCieXyz& image,
	Sheet<PixelStats>& pixelStats
) const
{
	/// maximize each pixels deviation with adjacent interpixel differences (maybe do at various spatial scales...)

	const dword width  = pixelStats.getWidth();
	const dword height = pixelStats.getHeight();

	/// do y
	for( dword y = 0;  y < height;  ++y )
	{
		/// is bottom row
		const bool isBottomRow = (y+1 == height);

		PixelStats*   pStat    = pixelStats.getRow( y );
		PixelStats*   pStatEnd = pStat + width;
		const CieXyz* pPixel   = image.getRow( y );

		/// do x
		float rightDeviation = 0.0f;
		for( ;  pStat < pStatEnd;  ++pStat, ++pPixel )
		{
			float pixelDeviation = pStat->getVariance();

			/// maximize pixel with rightdev
			pixelDeviation = pixelDeviation >= rightDeviation ? pixelDeviation : rightDeviation;

			/// if not bottom row
			if( !isBottomRow )
			{
				/// calc belowdev
				const float dif = (pPixel+width)->getY() - pPixel->getY();
				const float belowDeviation = dif * dif;

				/// maximize pixel with belowdev
				pixelDeviation = pixelDeviation >= belowDeviation ? pixelDeviation : belowDeviation;

				/// maximize below pixel with belowdev
				{
					float belowPixelDeviation = (pStat + width)->getVariance();
					belowPixelDeviation = belowPixelDeviation >= belowDeviation ? belowPixelDeviation : belowDeviation;
					(pStat + width)->setVariance( belowPixelDeviation );
				}
			}

			/// if not rightmost column
			if( (pStat+1) < pStatEnd )
			{
				/// calc next rightdev
				{
					const float dif = (pPixel+1)->getY() - pPixel->getY();
					rightDeviation = dif * dif;
				}

				/// maximize pixel with rightdev
				pixelDeviation = pixelDeviation >= rightDeviation ? pixelDeviation : rightDeviation;
			}

			pStat->setVariance( pixelDeviation );
		}
	}
}


void ImageAccumulation::analyseSamplesRequired
(
	const ImageCieXyz&       image,
	const Sheet<PixelStats>& pixelStats,
	vector<bool>&            required,
	bool&                    isAnySamplesRequired,
	dword&                   samplesRequiredCount,
	float&                   completedRatio
) const
{
	isAnySamplesRequired = false;
	samplesRequiredCount = 0;
	float completedRatioSum = 0.0f;

	const CieXyz*     pCieXyz = image.getMemory() + image.getLength();
	const PixelStats* pStats  = pixelStats.getMemory() + pixelStats.getLength();

	for( dword i = required.size();  pCieXyz--, pStats--, i-- > 0; )
	{
		const float mean         = pCieXyz->getY();
		const float maxDeviation = float(::sqrt( pStats->getVariance() ));
		const float samplesDone  = pStats->getCount();

		/// set minimum samples required
		float requiredSamples = 200.0f;//16.0f;

		if( (mean + maxDeviation) > 0.0f )
		{
			/// 1% difference in brightness is the perceptual threshold
			/// and, the max visible contrast is 100:1
			const float PERCEPTUAL_THRESHOLD = 0.00975f;

			/// if n samples are taken, each can only alter the mean by 1/n
			/// so each possible alteration = (sample-mean)/n
			/// if (worst-case difference) / n < perceptual threshold, its not worth taking more samples
			/// so maximum needed samples = (worst-case difference) / perceptual threshold
			/// (worst-case difference) has to be formulated in the same way as the perceptual threshold: proportional difference
			///
			/// ...ahh, but this hasnt taken into account the tone mapping
			///
			const float contrastRatio = 1.0f - (mean / (mean + maxDeviation));
			const float rs = (contrastRatio / PERCEPTUAL_THRESHOLD);

			if( rs > requiredSamples )
			{
				requiredSamples = rs;
			}
		}

		/// note for each pixel, and whether any
		{
			const bool isSampleRequired = samplesDone < requiredSamples;
			required[i] = isSampleRequired;
			isAnySamplesRequired |= isSampleRequired;
		}

		/// accumulate total
		samplesRequiredCount += hxa7241::utility::fpToInt( float(std::ceil( requiredSamples )) );

		/// accumulate completed ratio sum
		const float ratio = samplesDone / requiredSamples;
		completedRatioSum += (ratio <= 1.0f ? ratio : 1.0f);
	}

	/// calc completed ratio as a mean
	if( required.size() > 0 )
	{
		completedRatio = completedRatioSum / float(required.size());
	}
	else
	{
		completedRatio = 0.0f;
	}
}


bool ImageAccumulation::checkSheets( const char methodName[] ) const
{
	if( pImage_m == 0 )
	{
		throw hxa7241::utility::Exception( "pointer to image isnt set", CLASS_NAME(), methodName );
	}
	else if( !( isSameSize( pixelStats_m, *pImage_m ) &&
	          ( pixelStats_m.getLength() == dword(requiredSamples_m.size()) ) ) )
	{
		throw hxa7241::utility::Exception( "image size doesnt match pixel stats", CLASS_NAME(), methodName );
	}

	return true;
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::imaging::operator<<( SerializerOut& out, const ImageAccumulation& obj )
{
	/// write size
	out.write( "width" , obj.pixelStats_m.getWidth()  );
	out.write( "height", obj.pixelStats_m.getHeight() );

	/// write stats array
	{
		const PixelStats* pElements = obj.pixelStats_m.getMemory();
		const PixelStats* pEnd      = pElements + obj.pixelStats_m.getLength();
		for( ;  pElements < pEnd;  ++pElements )
		{
			out.write( "stats", *pElements );
		}
	}

	out.write( "samplesCount",         obj.samplesCount_m );
	out.write( "samplesRequiredCount", obj.samplesRequiredCount_m );
	out.write( "completedRatio",       obj.completedRatio_m );

	return out;
}


SerializerIn& perceptuum2::imaging::operator>>( SerializerIn& in, ImageAccumulation& obj )
{
	/// read size
	dword width;
	dword height;
	in.read( "width" , width  );
	in.read( "height", height );

	/// read stats array
	{
		/// allocate storage
		obj.pixelStats_m.setSize( width, height );

			  PixelStats* pElements = obj.pixelStats_m.getMemory();
		const PixelStats* pEnd      = pElements + obj.pixelStats_m.getLength();
		for( ;  pElements < pEnd;  ++pElements )
		{
			in.read( "stats", *pElements );
		}
	}

	/// read sample counts
	{
		in.read( "samplesCount",         obj.samplesCount_m );
		in.read( "samplesRequiredCount", obj.samplesRequiredCount_m );
		in.read( "completedRatio",       obj.completedRatio_m );

		obj.samplesCount_m         = obj.samplesCount_m         >= 0 ? obj.samplesCount_m         : 0;
		obj.samplesRequiredCount_m = obj.samplesRequiredCount_m >= 0 ? obj.samplesRequiredCount_m : 0;
		obj.completedRatio_m       = obj.completedRatio_m       >= 0.0f ? obj.completedRatio_m : 0.0f;
		obj.completedRatio_m       = obj.completedRatio_m       <= 1.0f ? obj.completedRatio_m : 1.0f;
	}

	return in;
}


const char* ImageAccumulation::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::imaging::ImageAccumulation";
	return CLASS_NAME;
}


const char* ImageAccumulation::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "image made by iterative refinement";
	return CLASS_DESCRIPTION;
}
