/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "FpToInt.h"
#include "Exception.h"

#include "ImageRgba.h"


using namespace perceptuum2::imaging;




/// statics
const int ImageRgba::BYTE_SHIFTS_RGBA[4][4] = { { 16,  8,  0, 24 },
                                                {  0,  8, 16, 24 },
                                                { 24, 16,  8,  0 },
                                                {  8, 16, 24,  0 } };

const char ImageRgba::CLASS_NAME[]         = "perceptuum2::imaging::ImageRgba";
const char ImageRgba::EXCEPTION_MESSAGE1[] = "width and/or height are not within allowable range";




/// standard object services -----------------------------------------------------------------------
ImageRgba::ImageRgba()
 :	storage_m(),
	pPixels_m( 0 ),
	width_m( 0 ),
	height_m( 0 ),
	byteShifts_m()
{
}


ImageRgba::ImageRgba
(
	dword*const          pPixels,
	const dword          width,
	const dword          height,
	const ERgbByteFormat byteFormat
)
{
	ImageRgba::set( pPixels, width, height, byteFormat );
}


ImageRgba::ImageRgba
(
	const dword          width,
	const dword          height,
	const ERgbByteFormat byteFormat
)
{
	ImageRgba::set( width, height, byteFormat );
}


ImageRgba::~ImageRgba()
{
}


ImageRgba::ImageRgba( const ImageRgba& other )
{
	ImageRgba::operator=( other );
}


ImageRgba& ImageRgba::operator=( const ImageRgba& other )
{
	if( &other != this )
	{
		storage_m.setSize( 0, 0 );

		pPixels_m    = other.pPixels_m;
		width_m      = other.width_m;
		height_m     = other.height_m;
		byteShifts_m = other.byteShifts_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void ImageRgba::set
(
	dword*const          pPixels,
	const dword          width,
	const dword          height,
	const ERgbByteFormat byteFormat
)
{
	static const char METHOD_NAME[] = "set";

	if( !isSizeWithinRange( width, height ) )
	{
		throw hxa7241::utility::Exception( EXCEPTION_MESSAGE1, CLASS_NAME, METHOD_NAME );
	}

	storage_m.setSize( 0, 0 );

	pPixels_m    = pPixels;
	width_m      = width;
	height_m     = height;
	byteShifts_m = BYTE_SHIFTS_RGBA[ byteFormat ];
}


void ImageRgba::set
(
	const dword          width,
	const dword          height,
	const ERgbByteFormat byteFormat
)
{
	storage_m.setSize( width, height );

	pPixels_m    = storage_m.getMemory();
	width_m      = width;
	height_m     = height;
	byteShifts_m = BYTE_SHIFTS_RGBA[ byteFormat ];
}


void ImageRgba::zeroMemory() const
{
	dword* pPixel = pPixels_m + (width_m * height_m);
	while( pPixel-- > pPixels_m )
	{
		*pPixel = 0;
	}
}


void ImageRgba::associateAlpha() const
{
	/// step thru pixels
	dword* pPixel = pPixels_m + (width_m * height_m);
	while( pPixel-- > pPixels_m )
	{
		/// multiply rgb by a
		*pPixel = associateAlphaP( *pPixel );
	}
}


void ImageRgba::associateAlpha
(
	const dword row
) const
{
	/// step thru pixels
	dword*       pPixel = getRow( row );
	dword* const pEnd   = pPixel + width_m;
	for( ;  pPixel < pEnd;  ++pPixel )
	{
		/// multiply rgb by a
		*pPixel = associateAlphaP( *pPixel );
	}
}


void ImageRgba::unassociateAlpha() const
{
	/// step thru pixels
	dword* pPixel = pPixels_m + (width_m * height_m);
	while( pPixel-- > pPixels_m )
	{
		/// divide rgb by a, or set to zero if a is zero
		*pPixel = unassociateAlphaP( *pPixel );
	}
}


void ImageRgba::unassociateAlpha
(
	const dword row
) const
{
	/// step thru pixels
	dword*       pPixel = getRow( row );
	dword* const pEnd   = pPixel + width_m;
	for( ;  pPixel < pEnd;  ++pPixel )
	{
		/// divide rgb by a, or set to zero if a is zero
		*pPixel = unassociateAlphaP( *pPixel );
	}
}




/// queries ----------------------------------------------------------------------------------------
dword ImageRgba::getWidth() const
{
	return width_m;
}


dword ImageRgba::getHeight() const
{
	return height_m;
}


dword ImageRgba::getLength() const
{
	return width_m * height_m;
}


dword& ImageRgba::operator[]( int i ) const
{
	return pPixels_m[ i ];
}


dword* ImageRgba::getMemory() const
{
	return pPixels_m;
}


dword* ImageRgba::getRow( dword y ) const
{
	y = (y >= 0) ? y : -y;
	return pPixels_m + (y % height_m) * width_m;
}


dword ImageRgba::toDword
(
	const dword rgba[4]
) const
{
	return ((rgba[0] & 0xFF) << byteShifts_m[0]) |
	       ((rgba[1] & 0xFF) << byteShifts_m[1]) |
	       ((rgba[2] & 0xFF) << byteShifts_m[2]) |
	       ((rgba[3] & 0xFF) << byteShifts_m[3]);
}


dword ImageRgba::toDword
(
	const float rgba[4]
) const
{
	using hxa7241::utility::fpToIntTrunc;
	return ((fpToIntTrunc(rgba[0] * 256.0f) & 0xFF) << byteShifts_m[0]) |
	       ((fpToIntTrunc(rgba[1] * 256.0f) & 0xFF) << byteShifts_m[1]) |
	       ((fpToIntTrunc(rgba[2] * 256.0f) & 0xFF) << byteShifts_m[2]) |
	       ((fpToIntTrunc(rgba[3] * 256.0f) & 0xFF) << byteShifts_m[3]);
}


dword* ImageRgba::fromDword
(
	const dword d,
	      dword rgba[4]
) const
{
	rgba[0] = (d >> byteShifts_m[0]) & 0xFF;
	rgba[1] = (d >> byteShifts_m[1]) & 0xFF;
	rgba[2] = (d >> byteShifts_m[2]) & 0xFF;
	rgba[3] = (d >> byteShifts_m[3]) & 0xFF;

	return rgba;
}


float* ImageRgba::fromDword
(
	const dword d,
	      float rgba[4]
) const
{
	rgba[0] = hxa7241::utility::ubyteToFp( (d >> byteShifts_m[0]) & 0xFF );
	rgba[1] = hxa7241::utility::ubyteToFp( (d >> byteShifts_m[1]) & 0xFF );
	rgba[2] = hxa7241::utility::ubyteToFp( (d >> byteShifts_m[2]) & 0xFF );
	rgba[3] = hxa7241::utility::ubyteToFp( (d >> byteShifts_m[3]) & 0xFF );

	return rgba;
}


dword ImageRgba::removeAlpha( const dword d ) const
{
	dword d2;

	switch( static_cast<ERgbByteFormat>(byteShifts_m - BYTE_SHIFTS_RGBA[0]) )
	{
		case ARGB :
		case ABGR :
		{
			d2 = d & 0x00FFFFFF;
			break;
		}
		case RGBA :
		case BGRA :
		{
			d2 = (d >> 8) & 0x00FFFFFF;
			break;
		}
	}

	return d2;
}


dword ImageRgba::associateAlphaP
(
	const dword pixel
) const
{
	/// multiply rgb by alpha
	/// postcondition: each channel <= alpha

	dword pixelAssociated;

	const dword alpha = (pixel >> byteShifts_m[3]) & 0xFF;
	if( alpha != 0 )
	{
		const dword rgba = pixel | (0xFF << byteShifts_m[3]);

		dword b02 = (alpha * (rgba & 0x00FF00FF)) + 0x00800080;
		b02 = ((b02 + ((b02 >> 8) & 0x00FF00FF)) >> 8) & 0x00FF00FF;

		dword b13 = (alpha * ((rgba >> 8) & 0x00FF00FF)) + 0x00800080;
		b13 = (b13 + ((b13 >> 8) & 0x00FF00FF)) & 0xFF00FF00;

		pixelAssociated = b02 | b13;
	}
	else
	{
		pixelAssociated = 0;
	}

	return pixelAssociated;
}


dword ImageRgba::unassociateAlphaP
(
	const dword pixel
) const
{
	/// divide rgb by alpha, or set to zero if alpha is zero
	/// precondition: that pixels are associated, ie: each channel <= alpha

	dword pixelUnassociated;

	dword rgba[4];
	fromDword( pixel, rgba );

	const dword alpha = rgba[3];
	if( alpha != 0 )
	{
		rgba[0] = (rgba[0] * 255) / alpha;
		rgba[1] = (rgba[1] * 255) / alpha;
		rgba[2] = (rgba[2] * 255) / alpha;

		pixelUnassociated = toDword( rgba );
	}
	else
	{
		pixelUnassociated = 0;
	}

	return pixelUnassociated;
}


bool ImageRgba::isSizeWithinRange( const dword width, const dword height )   /// static
{
	bool isWithin = true;

	/// is negative ?
	if( (width < 0) || (height < 0) )
	{
		isWithin = false;
	}
	else if( height != 0 )   /// prevent divide by zero
	{
		/// is the total number of elements greater than max length ?
		if( width > (DWORD_MAX / height) )
		{
			isWithin = false;
		}
	}

	return isWithin;
}
