/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Light.h"
#include "Vector3f.h"
#include "Matrix3f.h"
#include "Polar.h"
#include "RandomGood.h"
#include "HaltonSequence.h"
#include "HistogramSmall.h"
#include "Logger.h"
#include <vector>
#include <memory>

#include "Brdf.h"   /// own header is included last


using namespace perceptuum2::interaction;




//#define ENABLE_TEST_CODE


/// statics
const float Brdf::PI               = float( 3.1415926535897932384626433832795 );
const float Brdf::TWO_PI           = float( 2.0 * 3.1415926535897932384626433832795 );
const float Brdf::FOUR_PI          = float( 4.0 * 3.1415926535897932384626433832795 );
const float Brdf::ONE_OVER_PI      = float( 1.0 / 3.1415926535897932384626433832795 );
const float Brdf::ONE_OVER_TWO_PI  = float( 1.0 / (2.0 * 3.1415926535897932384626433832795) );
const float Brdf::ONE_OVER_FOUR_PI = float( 1.0 / (4.0 * 3.1415926535897932384626433832795) );




/// standard object services -----------------------------------------------------------------------
Brdf::Brdf()
{
	Brdf::setBrdf( UnitVector3f::Y(), UnitVector3f::X() );
}


Brdf::Brdf
(
	const UnitVector3f& normal,
	const UnitVector3f& tangent
)
{
	Brdf::setBrdf( normal, tangent );
}


Brdf::~Brdf()
{
}


Brdf::Brdf( const Brdf& other )
 :	normal_m  ( other.normal_m   ),
	tangentX_m( other.tangentX_m ),
	tangentY_m( other.tangentY_m )
{
}


Brdf& Brdf::operator=( const Brdf& other )
{
	if( &other != this )
	{
		normal_m   = other.normal_m;
		tangentX_m = other.tangentX_m;
		tangentY_m = other.tangentY_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void Brdf::setBrdf
(
	const UnitVector3f& normal,
	const UnitVector3f& tangent
)
{
	if( !normal.isZero() )
	{
		normal_m = normal;
	}
	else
	{
		/// default normal to -Y
		normal_m = -UnitVector3f::Y();
	}

	if( !tangent.isZero() )
	{
		tangentX_m = tangent;
	}
	else
	{
		/// default tangent to X
		tangentX_m = UnitVector3f::X();
	}


	tangentX_m.orthogonalizeEq( normal_m );

	tangentY_m = tangentX_m;
	tangentY_m.crossEq( normal_m );
}




/// queries ----------------------------------------------------------------------------------------
Float01 Brdf::getReflectivity() const
{
	return getReflectivity_();
}


void Brdf::getPartitions
(
	float   proportions[],
	Float01 peakness[],
	dword&  length
) const
{
	getPartitions_( proportions, peakness, length );
}


Float01 Brdf::getCos
(
	const dword         part,
	const UnitVector3f& outDirection
) const
{
	return getCos_( part, outDirection );
}


void Brdf::getScaling
(
	const dword         part,
	const UnitVector3f& inDirection,
	const UnitVector3f& outDirection,
	Light&              scaling
) const
{
	UnitVector3f reversedIn( inDirection );
	reversedIn.negate();

	const float IdotN = getNormal().dot( reversedIn );
	const float OdotN = getNormal().dot( outDirection );

	const bool isNotTransmission = (IdotN >= 0.0f) & (OdotN >= 0.0f);
	if( isNotTransmission )
	{
		getScaling_( part, IdotN, OdotN, reversedIn, outDirection, scaling );
	}
	else
	{
		scaling = Light::ZERO();
	}
}


bool Brdf::getSample
(
	const dword         part,
	const Float01       ,//roulette,
	const Vector2f01&   stochastic,
	const UnitVector3f& inDirection,
	UnitVector3f&       outDirection,
	Light&              scaling
) const
{
	UnitVector3f reversedIn( inDirection );
	reversedIn.negate();

	const bool isNotTransmission = getNormal().dot( reversedIn ) >= 0.0f;
	if( isNotTransmission )
	{
		getSample_( part, stochastic, reversedIn, outDirection, scaling );
	}
	else
	{
		outDirection = UnitVector3f::ZERO();
		scaling      = Light::ZERO();
	}

	return false;
}




void Brdf::getDirectionFromFrameCoords
(
	const Vector3f& frameCoords,
	UnitVector3f&   direction
) const
{
	hxa7241::graphics::Vector3f d;
	d  = (tangentX_m.getVector() * frameCoords.getX());
	d += (tangentY_m.getVector() * frameCoords.getY());
	d += (normal_m.getVector()   * frameCoords.getZ());

	direction = d;


//	const float dx = frameCoords.dot( Vector3f(tangentX_m.getX(), tangentY_m.getX(), normal_m.getX()) );
//	const float dy = frameCoords.dot( Vector3f(tangentX_m.getY(), tangentY_m.getY(), normal_m.getY()) );
//	const float dz = frameCoords.dot( Vector3f(tangentX_m.getZ(), tangentY_m.getZ(), normal_m.getZ()) );
//
//	direction.setXYZ( x, y, z );
}


void Brdf::getFrameCoordsFromDirection
(
	const UnitVector3f& direction,
	Vector3f&           frameCoords
) const
{
	frameCoords.setXYZ( direction.dot( tangentX_m ),
	                    direction.dot( tangentY_m ),
	                    direction.dot( normal_m   ) );
}


void Brdf::getDirectionFrom2DCoords
(
	const Vector2f01& coords,
	UnitVector3f&     direction
) const
{
	float x, y, z;
	hxa7241::graphics::squareToHemisphere( coords.getX(), coords.getY(), x, y, z );

	getDirectionFromFrameCoords( Vector3f(x,y,z), direction );
}


void Brdf::get2DCoordsFromDirection
(
	const UnitVector3f& direction,
	Vector2f01&         coords
) const
{
	Vector3f frameCoords;
	getFrameCoordsFromDirection( direction, frameCoords );
	if( frameCoords.getZ() < 0.0f )
	{
		frameCoords.z_m = -frameCoords.z_m;
	}

	float x, y;
	hxa7241::graphics::hemisphereToSquare( frameCoords.getX(), frameCoords.getY(), frameCoords.getZ(), x, y );
	coords.setXY( x, y );
}




bool Brdf::test
(
	const bool isLogging
) const
{
	return testAll( *this, isLogging );
}


void Brdf::testView
(
	const bool           isBrdfNotSdf,
	const dword          partitionsBits,
	const bool           isCos,
	const bool           isTransmission,
	const Sheet<CieXyz>& image
) const
{
	Brdf::viewAll( *this, isBrdfNotSdf, partitionsBits, isCos, isTransmission, image );
}








/// testing ----------------------------------------------------------------------------------------
#ifndef ENABLE_TEST_CODE
bool Brdf::testAll
(
	const Brdf&,
	const bool  isLogging
)
{
	if(isLogging) hxa7241::utility::Logger() << "Brdf::test  [disabled]";

	return false;
}


bool Brdf::testBrdf
(
	const Brdf&,
	const bool
)
{
	return false;
}


bool Brdf::testSdf
(
	const Brdf&,
	const bool
)
{
	return false;
}


bool Brdf::testOrientation
(
	const Brdf&,
	const bool
)
{
	return false;
}


void Brdf::viewAll
(
	const Brdf&,
	const bool,
	const dword,
	const bool,
	const bool,
	const Sheet<CieXyz>&
)
{
	hxa7241::utility::Logger() << "Brdf::view  [disabled]";
}


void Brdf::viewBrdf
(
	const Brdf&,
	const dword,
	const bool,
	const bool,
	const Sheet<CieXyz>&
)
{
}


void Brdf::viewSdf
(
	const Brdf&,
	const dword,
	const bool,
	const bool,
	const Sheet<CieXyz>&
)
{
}


#else
bool Brdf::testAll
(
	const Brdf& brdf,
	const bool  isLogging
)
{
	bool isSucceeded = true;

	if(isLogging) hxa7241::utility::Logger() << "Brdf::test  === brdf, sdf and orientation test ===";

	isSucceeded &= testBrdf( brdf, isLogging );
	isSucceeded &= testSdf( brdf, isLogging );
	isSucceeded &= testOrientation( brdf, isLogging );

	return isSucceeded;
}


void ::getScalingComplete
(
	const Brdf&         brdf,
	const dword         partitionsBits,
	const bool          isReflectivity,
	const bool          isCos,
	const UnitVector3f& inDirection,
	const UnitVector3f& outDirection,
	Light&              scaling
)
{
	scaling = Light::ZERO();

	float   proportions[Brdf::PARTITIONS_MAX];
	Float01 peakness[Brdf::PARTITIONS_MAX];
	dword   partsCount;
	brdf.getPartitions( proportions, peakness, partsCount );

	for( dword p = 0;  p < partsCount;  ++p )
	{
		if( (1 << p) & partitionsBits )
		{
			Light s;
			brdf.getScaling( p, inDirection, outDirection, s );

			const float cos = isCos ? brdf.getCos(p, outDirection) : 1.0f;
			scaling += (s *= (proportions[p] * cos));
		}
	}

	if( isReflectivity )
	{
		scaling *= brdf.getReflectivity();
	}
}


void ::getSampleComplete
(
	const Brdf&         brdf,
	const bool          isCos,
	const Float01       roulettes[3],
	const Vector2f01&   stochastic,
	const UnitVector3f& inDirection,
	UnitVector3f&       outDirection,
	Light&              scaling
)
{
	const float r[] = { brdf.getReflectivity(), 0.0f };
	const hxa7241::utility::HistogramSmall hs( 2, r );

	if( hs.getChoice( roulettes[0] ) == 0 )
	{
		hxa7241::utility::HistogramSmall hs( 0 );
		{
			float   proportions[Brdf::PARTITIONS_MAX];
			Float01 peakness[Brdf::PARTITIONS_MAX];
			dword   partsCount;
			brdf.getPartitions( proportions, peakness, partsCount );

			hs.setElements( partsCount, proportions );
		}

		const dword part = hs.getChoice( roulettes[1] );

		brdf.getSample( part, roulettes[2], stochastic, inDirection, outDirection, scaling );
		if( isCos )
		{
			scaling *= brdf.getCos( part, outDirection );
		}
		//if( isCos && (rouletteRescaled >= brdf.getCos( part, outDirection )) )
		//{
		//	outDirection = UnitVector3f::ZERO();
		//	scaling      = Light::ZERO();
		//}
	}
	else
	{
		outDirection = UnitVector3f::ZERO();
		scaling      = Light::ZERO();
	}
}


bool Brdf::testBrdf
(
	const Brdf& brdf,
	const bool  isLogging
)
{
	/// normalisation
	///    sum( brdf( dirIn, dirOut[i] ) * (pi/n) ) == reflectance <= 1
	///    (for one in direction, sum of scalings for all out directions == spec reflectance + diff reflectance <= 1)

	/// reciprocity
	///    brdf( in, out ) == brdf( out, in )


	/// set test resolution
	const dword WIDTH         = 12;//32;   /// 32^4 ~ 1 million   /// must be < 256
	const float oneOverWidth  = 1.0f / float(WIDTH);
	const float oneOverLength = 1.0f / float(WIDTH * WIDTH);

	/// init totals
	Light totalTotal;
	Light totalTotalCos;
	float maxScaling          = 0.0f;
	float maxScalingCos       = 0.0f;
	float reciprocityErrorMax = 0.0f;
	bool  isScalingNeg      = false;
	bool  isEnergyViolation = false;

	/// do in direction loops
	for( dword yIn = WIDTH;  yIn-- > 0; )
	{
		for( dword xIn = WIDTH;  xIn-- > 0; )
		{
			Light total;
			Light totalCos;

			/// make in direction
			UnitVector3f inDirection;
			brdf.getUniformSampleDirection( Vector2f01((xIn + 0.5f) * oneOverWidth, (yIn + 0.5f) * oneOverWidth), inDirection );
			inDirection.negate();

			/// do out direction loops
			for( dword yOut = WIDTH;  yOut-- > 0; )
			{
				for( dword xOut = WIDTH;  xOut-- > 0; )
				{
					/// make out direction
					UnitVector3f outDirection;
					brdf.getUniformSampleDirection( Vector2f01((xOut + 0.5f) * oneOverWidth, (yOut + 0.5f) * oneOverWidth), outDirection );

					/// get brdf scaling
					Light scaling;
					Light scalingCos;
					::getScalingComplete( brdf, 0xFF, true, false, inDirection, outDirection, scaling );
					::getScalingComplete( brdf, 0xFF, true, true,  inDirection, outDirection, scalingCos );

					/// accumulate sub totals
					total    += scaling;
					totalCos += scalingCos;
					isScalingNeg |= (scaling.smallest() < 0.0f) | (scalingCos.smallest() < 0.0f);

					if( maxScaling < scaling.largest() )
					{
						maxScaling = scaling.largest();
					}
					if( maxScalingCos < scalingCos.largest() )
					{
						maxScalingCos = scalingCos.largest();
					}

					/// test reciprocity
					{
						Light scalingInv;

						::getScalingComplete( brdf, 0xFF, true, false, -outDirection, -inDirection, scalingInv );

						const float difference = (scaling - scalingInv).average();
						float errorPercent = 0.0f;
						if( difference != 0.0f )
						{
							const float scalingMean  = scaling.average();
							errorPercent = ::fabs( difference / ((scalingMean != 0) ? scalingMean : scalingInv.average()) ) * 100.0f;
						}
						if( reciprocityErrorMax < errorPercent )
						{
							reciprocityErrorMax = errorPercent;
						}
					}
				}
			}

			/// accumulate totals
			total         *= (Brdf::PI * oneOverLength);
			totalCos      *= (Brdf::PI * oneOverLength);
			totalTotal    += total;
			totalTotalCos += totalCos;
			isEnergyViolation |= total.average() > (brdf.getReflectivity() + 0.000005f);

			if( isLogging & (WIDTH < 7) ) hxa7241::utility::Logger() << "Brdf::test  " << "total: " << total.average() << "  totalCos: " << totalCos.average();
		}
	}

	totalTotal    *= oneOverLength;
	totalTotalCos *= oneOverLength;


	const float reflDif             = totalTotal.average() - brdf.getReflectivity();
	const float reflDifErrorPercent = (brdf.getReflectivity() != 0.0f) ? ((reflDif / brdf.getReflectivity()) * 100.0f) : FLOAT_LARGE;

	/// log statistics
	if( isLogging )
	{
		hxa7241::utility::Logger() << "Brdf::test  " << "sample count: " << (WIDTH * WIDTH * WIDTH * WIDTH);
		hxa7241::utility::Logger() << "Brdf::test  " << "intended reflectivity mean:      " << brdf.getReflectivity();
		hxa7241::utility::Logger() << "Brdf::test  " << "brdf reflectivity mean:          " << totalTotal.average();
		hxa7241::utility::Logger() << "Brdf::test  " << "brdf reflectivity with cos mean: " << totalTotalCos.average();
		hxa7241::utility::Logger() << "Brdf::test  " << "difference:                      " << reflDif << "  error: " << reflDifErrorPercent << "%";
		hxa7241::utility::Logger() << "Brdf::test  " << "max scaling: " << maxScaling << "  max scaling with cos: " << maxScalingCos;
		hxa7241::utility::Logger() << "Brdf::test  " << "reciprocity error abs max: " << reciprocityErrorMax << "%";
	}


	/// test quantitys
	bool isSucceeded = true;

	/// this is too demanding for common implementations - normalisations adjustments never account for angle...
//	const bool isRN = (float(::fabs(reflDifErrorPercent)) < 2.0f);
//	if( isLogging & !isRN ) hxa7241::utility::Logger() << "Brdf::test  " << "*** reflectivity normalisation test failed";
//	isSucceeded &= isRN;

	const bool isRC = (totalTotal.average() <= (brdf.getReflectivity() + 0.001f));
	if( isLogging & !isRC ) hxa7241::utility::Logger() << "Brdf::test  " << "*** reflectivity energy conservation test failed";
	isSucceeded &= isRC;

	const bool isRCC = (totalTotalCos.average() <= 1.001f);
	if( isLogging & !isRCC ) hxa7241::utility::Logger() << "Brdf::test  " << "*** reflectivity with cos energy conservation test failed";
	isSucceeded &= isRCC;

	const bool isR = (reciprocityErrorMax < 2.0f);
	if( isLogging & !isR ) hxa7241::utility::Logger() << "Brdf::test  " << "*** reciprocity test failed";
	isSucceeded &= isR;

	if( isLogging & isScalingNeg ) hxa7241::utility::Logger() << "Brdf::test  " << "*** negative scaling test failed";
	isSucceeded &= !isScalingNeg;

	if( isLogging & isEnergyViolation ) hxa7241::utility::Logger() << "Brdf::test  " << "*** partial energy conservation test failed";
	isSucceeded &= !isEnergyViolation;


	if( isLogging ) hxa7241::utility::Logger() << "Brdf::test  " << (isSucceeded ? "--- brdf test succeeded" : "*** brdf test failed");
	return isSucceeded;
}


bool Brdf::testSdf
(
	const Brdf& brdf,
	const bool  isLogging
)
{
	/// sdfs matching brdfs


	/// set test resolution
	const dword WIDTH         = 12;//12;   /// 12^4 ~ 20000
	const dword SAMPLE_RATE   = 256;
	const dword SAMPLES       = (WIDTH * WIDTH) * SAMPLE_RATE;
	const float oneOverWidth  = 1.0f / float(WIDTH);
	const float oneOverLength = 1.0f / float(WIDTH * WIDTH);

	const dword seed = dword(std::time(0));
	hxa7241::utility::RandomGood rand( seed );

	std::vector<float> sampleCounts( WIDTH * WIDTH );
	std::vector<float> sampleCountsCos( WIDTH * WIDTH );

	/// init totals
	float totalTotal    = 0.0f;
	float totalTotalCos = 0.0f;
	float totalDifSum    = 0.0f;
	float totalDifSumCos = 0.0f;
	//float totalDifMax    = 0.0f;
	//float totalDifMaxCos = 0.0f;
	bool  isScalingNeg      = false;
	bool  isEnergyViolation = false;

	/// do in direction loops
	for( dword yIn = WIDTH;  yIn-- > 0; )
	{
		for( dword xIn = WIDTH;  xIn-- > 0; )
		{
			/// zero counter
			sampleCounts.resize( 0 );
			sampleCounts.resize( WIDTH * WIDTH );
			sampleCountsCos.resize( 0 );
			sampleCountsCos.resize( WIDTH * WIDTH );

			/// make in direction
			UnitVector3f inDirection;
			brdf.getUniformSampleDirection( Vector2f01((xIn + 0.5f) * oneOverWidth, (yIn + 0.5f) * oneOverWidth), inDirection );
			inDirection.negate();

			/// make histogram of sdf
			for( dword i = SAMPLES;  i-- > 0; )
			{
				/// make out direction
				UnitVector3f outDirection[2];
				Light        scaling[2];
				{
					const float sx = rand.getFloat();
					rand.next();
					const float sy = rand.getFloat();
					rand.next();
					const Vector2f01 stochastic( sx, sy );

					Float01 roulettes[3];
					roulettes[0] = rand.getFloat();
					rand.next();
					roulettes[1] = rand.getFloat();
					rand.next();
					roulettes[2] = rand.getFloat();
					rand.next();

					/// get sdf samples
					::getSampleComplete( brdf, false, roulettes, stochastic, inDirection, outDirection[0], scaling[0] );
					::getSampleComplete( brdf, true,  roulettes, stochastic, inDirection, outDirection[1], scaling[1] );
				}

				vector<float>* counts[2] = { &sampleCounts, &sampleCountsCos };
				for( dword j = 0;  j < 2;  ++j )
				{
					if( !outDirection[j].isZero() & !scaling[j].isZero() )
					{
						/// map direction to counter grid
						dword gridX, gridY;
						{
							float sqrX, sqrY;
							{
								const float hx = outDirection[j].dot( brdf.getTangentX() );
								const float hy = outDirection[j].dot( brdf.getTangentY() );
								const float hz = outDirection[j].dot( brdf.getNormal() );
								hxa7241::graphics::hemisphereToSquare( hx, hy, hz, sqrX, sqrY );
							}

							gridX = dword(sqrX * float(WIDTH));
							gridX = gridX < WIDTH ? gridX : WIDTH-1;

							gridY = dword(sqrY * float(WIDTH));
							gridY = gridY < WIDTH ? gridY : WIDTH-1;
						}

						/// add to counter
						(*(counts[j]))[gridX + (gridY * WIDTH)] += (scaling[j].average() * (1.0f/(SAMPLE_RATE * Brdf::PI)));
					}
				}
			}

			/// compare with brdf
			float total    = 0.0f;
			float totalCos = 0.0f;
			float difSum    = 0.0f;
			float difSumCos = 0.0f;
			for( dword yOut = WIDTH;  yOut-- > 0; )
			{
				for( dword xOut = WIDTH;  xOut-- > 0; )
				{
					UnitVector3f outDirection;
					brdf.getUniformSampleDirection( Vector2f01((xOut + 0.5f) * oneOverWidth, (yOut + 0.5f) * oneOverWidth), outDirection );

					Light scaling;
					Light scalingCos;

					::getScalingComplete( brdf, 0xFF, true, false, inDirection, outDirection, scaling );
					::getScalingComplete( brdf, 0xFF, true, true,  inDirection, outDirection, scalingCos );

					const float scalingSdf    = sampleCounts[xOut + (yOut * WIDTH)];
					const float scalingSdfCos = sampleCountsCos[xOut + (yOut * WIDTH)];

					total    +=	scalingSdf;
					totalCos +=	scalingSdfCos;
					const float difference    = scaling.average()    - scalingSdf;
					const float differenceCos = scalingCos.average() - scalingSdfCos;
					difSum    += difference;
					difSumCos += differenceCos;
					isScalingNeg |= (scalingSdf < 0.0f) | (scalingSdfCos < 0.0f);
				}
			}

			total         *= (Brdf::PI * oneOverLength);
			totalCos      *= (Brdf::PI * oneOverLength);
			totalTotal    += total;
			totalTotalCos += totalCos;
			isEnergyViolation |= total > (brdf.getReflectivity() + 0.02f);
			const float difMean    = difSum    / float(WIDTH * WIDTH);
			const float difMeanCos = difSumCos / float(WIDTH * WIDTH);

			totalDifSum    += difMean;
			totalDifSumCos += difMeanCos;
			//totalDifMax    = totalDifMax >= difMax ? totalDifMax : difMax;
			//totalDifMaxCos = totalDifMaxCos >= difMaxCos ? totalDifMaxCos : difMaxCos;

			if( isLogging & (WIDTH < 7) ) hxa7241::utility::Logger() << "Brdf::test  " << "total: " << total << "  totalCos: " << totalCos << "  difMean= " << difMean << "  difMeanCos= " << difMeanCos;
		}
	}


	totalTotal    *= oneOverLength;
	totalTotalCos *= oneOverLength;
	const float totalDifMean    = totalDifSum    / float(WIDTH * WIDTH);
	const float totalDifMeanCos = totalDifSumCos / float(WIDTH * WIDTH);


	const float reflDif             = totalTotal - brdf.getReflectivity();
	const float reflDifErrorPercent = (brdf.getReflectivity() != 0.0f) ? ((reflDif / brdf.getReflectivity()) * 100.0f) : FLOAT_LARGE;
	const float brdfMatchErrorPercent    = (brdf.getReflectivity() != 0.0f) ? ((totalDifMean    / brdf.getReflectivity()) * 100.0f) : FLOAT_LARGE;
	const float brdfCosMatchErrorPercent = (brdf.getReflectivity() != 0.0f) ? ((totalDifMeanCos / brdf.getReflectivity()) * 100.0f) : FLOAT_LARGE;

	/// log statistics
	if( isLogging )
	{
		hxa7241::utility::Logger() << "Brdf::test  " << "sample count: " << (WIDTH * WIDTH * WIDTH * WIDTH);
		hxa7241::utility::Logger() << "Brdf::test  " << "intended reflectivity mean:     " << brdf.getReflectivity();
		hxa7241::utility::Logger() << "Brdf::test  " << "sdf reflectivity mean:          " << totalTotal;
		hxa7241::utility::Logger() << "Brdf::test  " << "sdf reflectivity with cos mean: " << totalTotalCos;
		hxa7241::utility::Logger() << "Brdf::test  " << "difference:                     " << reflDif         << "  error: " << reflDifErrorPercent << "%";
		hxa7241::utility::Logger() << "Brdf::test  " << "brdf dif mean:                  " << totalDifMean    << "  error: " << brdfMatchErrorPercent << "%";
		hxa7241::utility::Logger() << "Brdf::test  " << "brdf with cos dif mean:         " << totalDifMeanCos << "  error: " << brdfCosMatchErrorPercent << "%";
	}


	/// check quantitys
	bool isSucceeded = true;

	const bool isRN = float(::fabs(reflDifErrorPercent)) < 2.0f;
	if( isLogging & !isRN ) hxa7241::utility::Logger() << "Brdf::test  " << "*** sdf reflectivity normalisation test failed";
	isSucceeded &= isRN;

	const bool isRC = totalTotal < (brdf.getReflectivity() + 0.02f);
	if( isLogging & !isRC ) hxa7241::utility::Logger() << "Brdf::test  " << "*** sdf reflectivity energy conservation test failed";
	isSucceeded &= isRC;

	const bool isRCC = totalTotalCos < 1.02f;
	if( isLogging & !isRCC ) hxa7241::utility::Logger() << "Brdf::test  " << "*** sdf reflectivity with cos energy conservation test failed";
	isSucceeded &= isRCC;

	const bool isMB = (float(::fabs(brdfMatchErrorPercent)) < 5.0f) & (float(::fabs(brdfCosMatchErrorPercent)) < 5.0f);
	if( isLogging & !isMB ) hxa7241::utility::Logger() << "Brdf::test  " << "*** sdf matching brdf test failed";
	isSucceeded &= isMB;

	if( isLogging & isScalingNeg ) hxa7241::utility::Logger() << "Brdf::test  " << "*** sdf negative scaling test failed";
	isSucceeded &= !isScalingNeg;

	if( isLogging & isEnergyViolation ) hxa7241::utility::Logger() << "Brdf::test  " << "*** sdf partial energy conservation test failed";
	isSucceeded &= !isEnergyViolation;


	if( isLogging ) hxa7241::utility::Logger() << "Brdf::test  " << (isSucceeded ? "--- sdf test succeeded" : "*** sdf test failed");
	return isSucceeded;
}


bool Brdf::testOrientation
(
	const Brdf& brdf,
	const bool  isLogging
)
{
	/// orientability
	//		brdf at one orientation == brdf at other orientation
	//		sdf at one orientation == sdf at other orientation


	static const dword ITERATIONS     = 32;
	static const dword WIDTH          = 12;
	static const float ONE_OVER_WIDTH = 1.0f / float(WIDTH);
	static const dword SAMPLES        = WIDTH * WIDTH;

	bool isBrdfSame = true;
	bool isSdfSame  = true;

	/// init rand
	const dword seed = dword(std::time(0));
	hxa7241::utility::RandomGood rand( seed );

	/// make rotatable brdf
	std::auto_ptr<Brdf> pBrdfRotated( brdf.clone() );

	/// do loop
	for( dword i = 0;  i < ITERATIONS;  ++i )
	{
		/// make rotation
		hxa7241::graphics::Matrix3f rotation( hxa7241::graphics::Matrix3f::IDENTITY() );
		{
			hxa7241::graphics::Matrix3f m0;
			const float rx = rand.getFloat() * 2.0 * Brdf::PI;
			rand.next();
			m0.setToRotateX( rx );
			rotation.multiplyEq( m0 );

			const float ry = rand.getFloat() * 2.0 * Brdf::PI;
			rand.next();
			m0.setToRotateY( ry );
			rotation.multiplyEq( m0 );

			const float rz = rand.getFloat() * 2.0 * Brdf::PI;
			rand.next();
			m0.setToRotateZ( rz );
			rotation.multiplyEq( m0 );
//hxa7241::utility::Logger() << "Brdf::test  " << "  rotation: " << rotation.getRow0() << "  " << rotation.getRow1() << "  " << rotation.getRow2() << "  " << rotation.getCol3();
		}

		/// set rotated brdf
		{
			UnitVector3f normalRotated;
			UnitVector3f tangentXRotated;
			rotation.multiply( brdf.getNormal(), normalRotated );
			rotation.multiply( brdf.getTangentX(), tangentXRotated );

			pBrdfRotated->setBrdf( normalRotated, tangentXRotated );
		}


		/// do in y loop
		for( dword yIn = 0;  yIn < WIDTH;  ++yIn )
		{
			/// do in x loop
			for( dword xIn = 0;  xIn < WIDTH;  ++xIn )
			{
//hxa7241::utility::Logger() << "Brdf::test  ";
				/// make in direction
				UnitVector3f inDirection;
				brdf.getUniformSampleDirection( Vector2f01((xIn + 0.5f) * ONE_OVER_WIDTH, (yIn + 0.5f) * ONE_OVER_WIDTH), inDirection );
				inDirection.negate();

				/// do out y loop
				for( dword yOut = 0;  yOut < WIDTH;  ++yOut )
				{
					/// do out x loop
					for( dword xOut = 0;  xOut < WIDTH;  ++xOut )
					{
						/// make out direction
						UnitVector3f outDirection;
						brdf.getUniformSampleDirection( Vector2f01((xOut + 0.5f) * ONE_OVER_WIDTH, (yOut + 0.5f) * ONE_OVER_WIDTH), outDirection );

						/// get unrotated scaling
						Light scaling;
						{
							/// call brdf for scaling
							::getScalingComplete( brdf, 0xFF, true, true,  inDirection, outDirection, scaling );
						}

						/// get rotated scaling
						Light scalingRotated;
						{
							/// rotate in and out direction
							UnitVector3f inDirectionRotated;
							UnitVector3f outDirectionRotated;
							rotation.multiply( inDirection, inDirectionRotated );
							rotation.multiply( outDirection, outDirectionRotated );
//hxa7241::utility::Logger() << "Brdf::test  " << "  in: " << inDirection << "  out: " << outDirection << "  inrot: " << inDirectionRotated << "  outrot: " << outDirectionRotated;

							/// call rotated brdf for scaling
							::getScalingComplete( *pBrdfRotated, 0xFF, true, true,  inDirectionRotated, outDirectionRotated, scalingRotated );
						}

						/// compare the two scalings
						{
							const Light dif( scaling - scalingRotated );
							const float larger = ::max( ::fabs(dif.largest()), ::fabs(dif.smallest()) );
							isBrdfSame &= larger < 0.0001f;
//if( larger >= 0.000001f ) hxa7241::utility::Logger() << "Brdf::test  " << xIn << " " << yIn << ", " << xOut << " " << yOut << "  scaling: " << scaling << "  scalingRotated: " << scalingRotated << "    dif: " << dif << "  larger: " << larger;
//hxa7241::utility::Logger(true) << "Brdf::test  " << "brdf  scalingdif= " << dif << "  scalingdiflarger= " << larger;
						}
					}
				}

				///
				for( dword s = 0;  s < SAMPLES;  ++s )
				{
					/// make stochastics
					Vector2f01 stochastic;
					Float01    roulettes[3];
					{
						const float sx = rand.getFloat();
						rand.next();
						const float sy = rand.getFloat();
						rand.next();
						stochastic.setXY( sx, sy );

						roulettes[0] = rand.getFloat();
						rand.next();
						roulettes[1] = rand.getFloat();
						rand.next();
						roulettes[2] = rand.getFloat();
						rand.next();
					}

					/// get unrotated sample
					UnitVector3f sampleDirection;
					Light        sampleScaling;
					{
						/// call brdf for sample
						::getSampleComplete( brdf, true, roulettes, stochastic, inDirection, sampleDirection, sampleScaling );

						/// rotate out direction
						rotation.multiply( sampleDirection );
					}

					/// get rotated sample
					UnitVector3f sampleDirectionRotated;
					Light        sampleScalingRotated;
					{
						/// rotate in direction
						UnitVector3f inDirectionRotated;
						rotation.multiply( inDirection, inDirectionRotated );

						/// call rotated brdf for sample
						::getSampleComplete( *pBrdfRotated, true, roulettes, stochastic, inDirectionRotated, sampleDirectionRotated, sampleScalingRotated );
					}

					/// compare the two out directions and scalings
					{
						const Light dif( sampleScaling - sampleScalingRotated );
						const float larger = ::max( ::fabs(dif.largest()), ::fabs(dif.smallest()) );
						isSdfSame &= (larger < 0.0001f);
						isSdfSame &= (sampleDirection == sampleDirectionRotated) | (sampleDirection.dot( sampleDirectionRotated ) > 0.999f);
//if( (larger >= 0.00001f) | !((sampleDirection == sampleDirectionRotated) | (sampleDirection.dot( sampleDirectionRotated ) > 0.999f)) )
//	hxa7241::utility::Logger() << "Brdf::test  " << "sdf  " << xIn << " " << yIn << "  scaling: " << sampleScaling << "  scalingRot: " << sampleScalingRotated << "  direction: " << sampleDirection << "  directionRot: " << sampleDirectionRotated << "    dif= " << dif << "  larger= " << larger << "  dot= " << sampleDirection.dot( sampleDirectionRotated );
//hxa7241::utility::Logger(true) << "Brdf::test  " << "sdf  scalingdif= " << dif << "  scalingdiflarger= " << larger << "  dot= " << sampleDirection.dot( sampleDirectionRotated );
					}
				}
			}
		}
	}


	bool isSucceeded = true;

	if( isLogging & !isBrdfSame ) hxa7241::utility::Logger() << "Brdf::test  " << "*** brdf orientation test failed (" << seed << ")";
	isSucceeded &= isBrdfSame;

	if( isLogging & !isSdfSame ) hxa7241::utility::Logger() << "Brdf::test  " << "*** sdf orientation test failed (" << seed << ")";
	isSucceeded &= isSdfSame;

	if( isLogging ) hxa7241::utility::Logger() << "Brdf::test  " << (isSucceeded ? "--- orientation test succeeded" : "*** orientation test failed");
	return isSucceeded;
}


void Brdf::viewAll
(
	const Brdf&          brdf,
	const bool           isBrdfNotSdf,
	const dword          partitionsBits,
	const bool           isCos,
	const bool           isTransmission,
	const Sheet<CieXyz>& image
)
{
	if( isBrdfNotSdf )
	{
		viewBrdf( brdf, partitionsBits, isCos, isTransmission, image );
	}
	else
	{
		viewSdf( brdf, partitionsBits, isCos, isTransmission, image );
	}
}


void Brdf::viewBrdf
(
	const Brdf&          brdf,
	const dword          partitionsBits,
	const bool           isCos,
	const bool           isTransmission,
	const Sheet<CieXyz>& image
)
{
	/// draw grid of input angle responses

	/// find size of largest square within image size
	const dword imageSize = image.getWidth() > image.getHeight() ? image.getWidth() : image.getHeight();

	/// make block size
	const dword blocksCount        = 11;
	const dword blockSizeOuter     = (imageSize / blocksCount) & 0xFFFFFFFE;
	const dword blockSize          = blockSizeOuter - 1;
	const float oneOverBlocksCount = 1.0f / float(blocksCount);
	const float oneOverBlockSize   = 1.0f / float(blockSize);

	/// clear image
	image.zeroMemory();

	/// do in y loop
	for( dword yIn = 0;  yIn < blocksCount;  ++yIn )
	{
		/// do in x loop
		for( dword xIn = 0;  xIn < blocksCount;  ++xIn )
		{
			/// make in direction
			UnitVector3f inDirection;
			brdf.getUniformSampleDirection( Vector2f01((xIn + 0.5f) * oneOverBlocksCount, (yIn + 0.5f) * oneOverBlocksCount), inDirection );
			inDirection.negate();

			/// do out y loop
			for( dword yOut = 0;  yOut < blockSize;  ++yOut )
			{
				/// do out x loop
				for( dword xOut = 0;  xOut < blockSize;  ++xOut )
				{
					/// make out direction
					UnitVector3f outDirection;
					{
						/// to fill block
						//brdf.getUniformSampleDirection( Vector2f01((xOut + 0.5f) * oneOverBlockSize, (yOut + 0.5f) * oneOverBlockSize), outDirection );

						/// to fill disk inside block
						hxa7241::graphics::Vector2f xy( xOut + 0.5f, yOut + 0.5f );
						((xy *= oneOverBlockSize) *= 2.0f) -= hxa7241::graphics::Vector2f::ONE();
						const float r2 = ((xy.getX() * xy.getX()) + (xy.getY() * xy.getY()));
						if( r2 <= 1.0f )
						{
							brdf.getDirectionFromFrameCoords( Vector3f( xy.getX(), xy.getY(), ::sqrt(1.0f - r2) ), outDirection );
						}
						else
						{
							outDirection = UnitVector3f::ZERO();
						}
					}

					if( !outDirection.isZero() )
					{
						if( isTransmission )
						{
							outDirection.negate();
						}

						/// call brdf for value
						Light scaling;
						::getScalingComplete( brdf, partitionsBits, false, isCos, inDirection, outDirection, scaling );

						/// write image pixel as value in range 0-100
						const dword x = (xIn * blockSizeOuter) + xOut;
						const dword y = (yIn * blockSizeOuter) + yOut;
						image.getElement( x, y ) = (scaling.getXYZ() *= 100.0f);
					}
				}
			}
		}
	}
}


void Brdf::viewSdf
(
	const Brdf&          brdf,
	const dword          partitionsBits,
	const bool           isCos,
	const bool           isTransmission,
	const Sheet<CieXyz>& image
)
{
	/// draw grid of input angle responses

	/// find size of largest square within image size
	const dword imageSize = image.getWidth() > image.getHeight() ? image.getWidth() : image.getHeight();

	/// make block size
	const dword blocksCount        = 11;
	const dword blockSizeOuter     = (imageSize / blocksCount) & 0xFFFFFFFE;
	const dword blockSize          = blockSizeOuter - 1;
	const float oneOverBlocksCount = 1.0f / float(blocksCount);
	const float hemisphereArea     = 0.5f * (4.0f * Brdf::PI * (float(blockSize) * 0.5) * (float(blockSize) * 0.5));
	const dword SAMPLE_COUNT       = 16;

	/// call sdf to get partitions
	hxa7241::utility::HistogramSmall parts( 0 );
	{
		float   proportions[Brdf::PARTITIONS_MAX];
		Float01 peakness[Brdf::PARTITIONS_MAX];
		dword   partsCount;
		brdf.getPartitions( proportions, peakness, partsCount );

		parts.setElements( partsCount, proportions );
	}

	/// init rand
	//const dword seed = dword(std::time(0));
	//hxa7241::utility::RandomGood rand( seed );
	/// (halton noticeably better than rand)
	hxa7241::utility::HaltonSequence halt0( 0 );
	hxa7241::utility::HaltonSequence halt1( 1 );
	hxa7241::utility::HaltonSequence halt2( 2 );
	hxa7241::utility::HaltonSequence halt3( 3 );
	hxa7241::utility::HaltonSequence halt4( 4 );

	/// clear image
	image.zeroMemory();

	/// do in y loop
	for( dword yIn = 0;  yIn < blocksCount;  ++yIn )
	{
		/// do in x loop
		for( dword xIn = 0;  xIn < blocksCount;  ++xIn )
		{
			/// make in direction
			UnitVector3f inDirection;
			brdf.getUniformSampleDirection( Vector2f01((xIn + 0.5f) * oneOverBlocksCount, (yIn + 0.5f) * oneOverBlocksCount), inDirection );
			inDirection.negate();

			/// do sampling loop
			for( dword s = 0;  s < (hemisphereArea * SAMPLE_COUNT);  ++s )
			{
				/// select partition with roulette
				//Float01 roulette( rand.getFloat() );
				//rand.next();
				const Float01 r0( halt0.getItem() );
				halt0.next();
				const dword part = parts.getChoice( r0 );
				if( (1 << part) & partitionsBits )
				{
					/// make stochastic coords
					//const float sx = rand.getFloat();
					//rand.next();
					//const float sy = rand.getFloat();
					//rand.next();
					const float sx = halt1.getItem();
					halt1.next();
					const float sy = halt2.getItem();
					halt2.next();
					const Vector2f01 stochastic( sx, sy );

					const Float01 r1( halt3.getItem() );
					halt3.next();
					const Float01 r2( halt4.getItem() );
					halt4.next();

					/// call sdf for out direction and scaling
					UnitVector3f outDirection;
					Light scaling;
					brdf.getSample( part, r1, stochastic, inDirection, outDirection, scaling );

					/// select if cos eliminates sample, with roulette
					const float cos = brdf.getCos( part, outDirection );
					if( !isCos | (r2 < cos) )
					{
						if( isTransmission )
						{
							outDirection.negate();
						}

						Vector3f xyz;
						brdf.getFrameCoordsFromDirection( outDirection, xyz );

						if( xyz.getZ() >= 0.0f )
						{
							/// make out xy from out direction
							//float squrX01;
							//float squrY01;
							//hxa7241::graphics::hemisphereToSquare( xyz.getX(), xyz.getY(), xyz.getZ(), squrX01, squrY01 );
							const float squrX01 = (xyz.getX() + 1.0f) * 0.5f;
							const float squrY01 = (xyz.getY() + 1.0f) * 0.5f;
							scaling *= outDirection.dot(brdf.getNormal());

							/// add to image pixel at out xy
							const dword x = (xIn * blockSizeOuter) + (squrX01 * blockSize);
							const dword y = (yIn * blockSizeOuter) + (squrY01 * blockSize);
							image.getElement( x, y ) += (scaling.getXYZ() *= (100.0f * Brdf::ONE_OVER_PI / SAMPLE_COUNT));
						}
					}
				}
			}
		}
	}
}
#endif
