/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef BssdfPhong_h
#define BssdfPhong_h


#include "Primitives.h"
#include "Bssdf.h"
#include "Light.h"
#include "Float01.h"




#include "perceptuum2_interaction.h"
namespace perceptuum2 { namespace interaction {
	using namespace hxa7241::primitives;
	using hxa7241::utility::Float01;


/**
 * phong brdf implementation, augmented with constant and spike.<br/><br/>
 *
 * a weighted sum of perfect-diffuse, normalised-phong and perfect-specular.
 * <br/><br/>
 *
 * weights are determined by the roughness. roughness of 1 is all perfect-diffuse,
 * roughness of 0 is all perfect-specular, between is a combination of diffuse and
 * phong, or phong and specular.
 *
 * @invariants
 * colors each have an average of 1.0<br/>
 * diffuseness, glossyness and specularness sum to 1.0<br/>
 */
class BssdfPhong
	: public Bssdf
{
/// standard object services -----------------------------------------------------------------------
public:
	         BssdfPhong();
	         BssdfPhong( const UnitVector3f& normal,
	                     const UnitVector3f& tangent,
	                     float               refractiveIndexOutside,
	                     float               refractiveIndexInside,
	                     Float01             transparency,
	                     Float01             reflectivity,
	                     const Light&        colorDiffuse,
	                     const Light&        colorSpecular,
	                     Float01             roughness );

	virtual ~BssdfPhong();
	         BssdfPhong( const BssdfPhong& );
	BssdfPhong& operator=( const BssdfPhong& );


/// commands ---------------------------------------------------------------------------------------
	virtual void  setBssdfPhong( Float01      reflectivity,
	                             const Light& colorDiffuse,
	                             const Light& colorSpecular,
	                             Float01      roughness );


/// queries ----------------------------------------------------------------------------------------
	virtual Brdf* clone()                                                                     const;


/// implementation ---------------------------------------------------------------------------------
protected:
	/// overrides
	virtual Float01 getReflectivity_()                                                        const;
	virtual void    getPartitions_( float   proportions[PARTITIONS_MAX],
	                                Float01 peakness[PARTITIONS_MAX],
	                                dword&  length )                                          const;
	virtual Float01 getCos_( dword               part,
	                         const UnitVector3f& outDirection )                               const;
	virtual void    getScaling_( dword               part,
	                             float               IdotN,
	                             float               OdotN,
	                             const UnitVector3f& inDirection,
	                             const UnitVector3f& outDirection,
	                             Light&              scaling )                                const;
	virtual void    getSample_( dword               part,
	                            const Vector2f01&   stochastic,
	                            const UnitVector3f& inDirection,
	                            UnitVector3f&       outDirection,
	                            Light&              scaling )                                 const;

private:
	/// implementation
	enum EParts { DIFFUSE_PART, GLOSSY_PART, SPECULAR_PART };

	        void  brdfDiffuse( const UnitVector3f& inDirection,
	                           const UnitVector3f& outDirection,
	                           Light&              scaling )                                  const;
	        void  brdfGlossy( const UnitVector3f& inDirection,
	                          const UnitVector3f& outDirection,
	                          Light&              scaling )                                   const;
	        void  brdfSpecular( const UnitVector3f& inDirection,
	                            const UnitVector3f& outDirection,
	                            Light&              scaling )                                 const;

	        void  sdfDiffuse( const Vector2f01&   stochastic,
	                          const UnitVector3f& inDirection,
	                          UnitVector3f&       outDirection,
	                          Light&              scaling )                                   const;
	        void  sdfGlossy( const Vector2f01&   stochastic,
	                         const UnitVector3f& inDirection,
	                         UnitVector3f&       outDirection,
	                         Light&              scaling )                                    const;
	        void  sdfSpecular( const Vector2f01&   stochastic,
	                           const UnitVector3f& inDirection,
	                           UnitVector3f&       outDirection,
	                           Light&              scaling )                                  const;


/// fields -----------------------------------------------------------------------------------------
private:
	/// parameters
	Float01 reflectivity_m;
	Light   colorDiffuse_m;
	Light   colorSpecular_m;
	Float01 roughness_m;

	/// auxilliarys
	Light colorGloss_m;

	float diffuseness_m;
	float glossyness_m;
	float specularness_m;

	float exponent_m;
	float normalisation_m;
};


}}




#endif//BssdfPhong_h
