/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "hxa7241math.h"

#include "BssdfSchlick.h"   /// own header is included last


using namespace perceptuum2::interaction;




/// standard object services -----------------------------------------------------------------------
BssdfSchlick::BssdfSchlick
(
	const UnitVector3f& normal,
	const UnitVector3f& tangent,
	const float         refractiveIndexOutside,
	const float         refractiveIndexInside,
	const Float01       transparency,
	const Light&        reflectivity,
	const Float01       roughness,
	const Float01       anisotropy
)
 :	Bssdf( normal, tangent, refractiveIndexOutside, refractiveIndexInside, transparency )
{
	BssdfSchlick::setBssdfSchlick( reflectivity, roughness, anisotropy );
}


BssdfSchlick::~BssdfSchlick()
{
}


BssdfSchlick::BssdfSchlick( const BssdfSchlick& other )
{
	BssdfSchlick::operator=( other );
}


BssdfSchlick& BssdfSchlick::operator=( const BssdfSchlick& other )
{
	if( &other != this )
	{
		Bssdf::operator=( other );

		reflectivitySpectral_m = other.reflectivitySpectral_m;
		roughness_m            = other.roughness_m;
		isotropy_m             = other.isotropy_m;

		reflectivity_m = other.reflectivity_m;
		color_m        = other.color_m;

		g_m = other.g_m;
		d_m = other.d_m;
		s_m = other.s_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void BssdfSchlick::setBssdfSchlick
(
	const Light&  reflectivity,
	const Float01 roughness,
	const Float01 anisotropy
)
{
	/// precondition reflectivity
	reflectivitySpectral_m = reflectivity;
	reflectivitySpectral_m.clampTo01Positive();

	/// precondition roughness
	roughness_m = Float01::moveUp( roughness );
	//static const float ROUGHNESS_MIN = 1.0f / 128.0f;   /// a hundredth of the max ...?
	//roughness_m = roughness;
	//if( roughness_m < ROUGHNESS_MIN )
	//{
	//	roughness_m = ROUGHNESS_MIN;
	//}

	/// precondition isotropy
	isotropy_m  = 1.0f - anisotropy;
	//static const float ISOTROPY_MIN = 1.0f / 128.0f;   /// a hundredth of the max ...?
	//isotropy_m  = 1.0f - anisotropy;
	//if( isotropy_m < ISOTROPY_MIN )
	//{
	//	isotropy_m = ISOTROPY_MIN;
	//}


	/// set auxilliarys
	{
		/// make reflectivity and color
		{
			float mean;
			reflectivitySpectral_m.getMeanAndColor( mean, color_m );
			reflectivity_m = mean;
		}

		/// make glossy, diffuse, specular weightings
		g_m = 4.0f * roughness_m * (1.0f - roughness_m);
		d_m = roughness_m < 0.5f ? 0.0f : 1.0f - g_m;
		s_m = roughness_m < 0.5f ? 1.0f - g_m : 0.0f;
	}
}




/// queries ----------------------------------------------------------------------------------------
Brdf* BssdfSchlick::clone() const
{
	Brdf* pClone = new BssdfSchlick( *this );
	if( pClone == 0 )
	{
		throw "BssdfSchlick::clone()\ncouldnt get memory";
	}

	return  pClone;
}




/// implementation ---------------------------------------------------------------------------------
Float01 BssdfSchlick::getReflectivity_() const
{
	return reflectivity_m;
}


void BssdfSchlick::getPartitions_
(
	float   proportions[],
	Float01 peakness[],
	dword&  length
) const
{
	proportions[DIFFUSE_PART]  = d_m;
	peakness[DIFFUSE_PART]     = Float01::MIN();

	proportions[GLOSSY_PART]   = g_m;
	peakness[GLOSSY_PART]      = roughness_m.MAX() - roughness_m;

	proportions[SPECULAR_PART] = s_m;
	peakness[SPECULAR_PART]    = Float01::MAX();

	length = 3;
}


Float01 BssdfSchlick::getCos_
(
	const dword         part,
	const UnitVector3f& outDirection
) const
{
	float cos;

	if( part != SPECULAR_PART )
	{
		cos = getNormal().dot( outDirection );
		if( cos < 0.0f )
		{
			cos = -cos;
		}

		if( part == GLOSSY_PART )
		{
			cos = 1.0f + (roughness_m * (cos - 1.0f));
		}
	}
	else
	{
		cos = 1.0f;
	}

	return cos;
}


void BssdfSchlick::getScaling_
(
	const dword         part,
	const float         IdotN,
	const float         OdotN,
	const UnitVector3f& inDirection,
	const UnitVector3f& outDirection,
	Light&              scaling
) const
{
	/// half vector
	const UnitVector3f h( inDirection.getVector() + outDirection.getVector() );

	/// spectral factor
	Light spectral;
	{
		const float u                = h.dot( outDirection );
		const float oneMinusU        = 1.0f - u;
		const float oneMinusUSquared = oneMinusU * oneMinusU;
		const float oneMinusUFifth   = oneMinusUSquared * oneMinusUSquared * oneMinusU;

		spectral  = Light::ONE();
		spectral -= color_m;
		spectral *= oneMinusUFifth;
		spectral += color_m;
	}

	/// part coefficient
	float coefficient;
	switch( part )
	{
		case DIFFUSE_PART  :
		{
			brdfDiffuse( inDirection, outDirection, coefficient );
			break;
		}
		case GLOSSY_PART   :
		{
			brdfGlossy( h, IdotN, OdotN, coefficient );
			break;
		}
		case SPECULAR_PART :
		{
			brdfSpecular( inDirection, outDirection, coefficient );
			break;
		}
		default :
		{
			coefficient = 0.0f;
			break;
		}
	}

	scaling = (spectral *= coefficient);
}


void BssdfSchlick::getSample_
(
	const dword         ,//part,
	const Vector2f01&   ,//stochastic,
	const UnitVector3f& ,//inDirection,
	UnitVector3f&       outDirection,
	Light&              scaling
) const
{
//	switch( part )
//	{
//		case DIFFUSE_PART  :
//		{
//			sdfDiffuse( stochastic, inDirection, outDirection );
//			break;
//		}
//		case GLOSSY_PART   :
//		{
//			sdfGlossy( stochastic, inDirection, outDirection );
//			break;
//		}
//		case SPECULAR_PART :
//		{
//			sdfSpecular( stochastic, inDirection, outDirection );
//			break;
//		}
//		default :
//		{
//			outDirection = UnitVector3f::ZERO();
//			scaling      = Light::ZERO();
//			break;
//		}
//	}
//
//	scaling = color_m;


	outDirection = UnitVector3f::ZERO();
	scaling      = Light::ZERO();
}




inline
void BssdfSchlick::brdfDiffuse
(
	const UnitVector3f& ,//inDirection,
	const UnitVector3f& ,//outDirection,
	float&              scaling
) const
{
	scaling = ONE_OVER_PI;
}


void BssdfSchlick::brdfGlossy
(
	const UnitVector3f& h,
	const float         vp,
	const float         v,
	float&              scaling
) const
{
	/// precondition: (getNormal().dot( inDirection ) >= 0.0f) & (getNormal().dot( outDirection ) >= 0.0f) )

	if( (vp == 0.0f) | (v == 0.0f) )
	{
		/// grazing vectors
		if( h.isZero() )
		{
			/// straight across
			scaling = 1.0f;
		}
	}
	else
	{
		const float t = h.dot( getNormal() );
		const float w = getTangentX().dot( UnitVector3f(h.getVector() - (getNormal().getVector() * t)) );

		float Z;
		{
			const float t2 = t * t;
			const float a  = 1.0f + (roughness_m * t2) - t2;
			Z = roughness_m / (a * a);
		}

		float A;
		{
			const float i2 = isotropy_m * isotropy_m;
			const float w2 = w * w;
			A = float(::sqrt( isotropy_m / (i2 - (i2 * w2) + w2) ));
		}

		const float G  = v  / (roughness_m - (roughness_m * v)  + v);
		const float Gp = vp / (roughness_m - (roughness_m * vp) + vp);

		const float GGp = G * Gp;

		scaling = ((GGp * Z * A) + 1.0f - GGp) / (FOUR_PI * v * vp);
	}
}


inline
void BssdfSchlick::brdfSpecular
(
	const UnitVector3f& ,//inDirection,
	const UnitVector3f& ,//outDirection,
	float&              scaling
) const
{
	/// not really do-able numerically...

	/*UnitVector3f mirrorDirection( inDirection );
	mirrorDirection.reflectEq( getNormal() );

	const float outDotMirror = mirrorDirection.dot( outDirection );
	if( outDotMirror >= 1.0f )
	{
		scaling = FLOAT_LARGE;
	}
	else
	{
		scaling = 0.0f;
	}*/


	scaling = 0.0f;
}


inline
void BssdfSchlick::sdfDiffuse
(
	const Vector2f01&   stochastic,
	const UnitVector3f& ,//inDirection,
	UnitVector3f&       outDirection
) const
{
	getUniformSampleDirection( stochastic, outDirection );
}


void BssdfSchlick::sdfGlossy
(
	const Vector2f01&   ,//stochastic,
	const UnitVector3f& ,//inDirection,
	UnitVector3f&       ,//outDirection,
	float&              //scaling
) const
{
	/// precondition: getNormal().dot( inDirection ) >= 0.0f


	// const float t   = ::sqrt( stochastic.getX() / (roughness_m - (roughness_m * stochastic.getX()) + stochastic.getX()) );

	// const float i2  = isotropy_m * isotropy_m;
	// const float sy2 = stochastic.getY() * stochastic.getY();
	// const float w   = ::cos( PI_OVER_TWO * float(::sqrt( (i2 * sy2) / (1.0f - sy2 + (i2 * sy2)) )) );

	// const float v  = getNormal().dot( outDirection );
	// const float G  = v  / (roughness_m - (roughness_m * v)  + v);
	// scaling =


//	/// clamp direction to hemisphere
//	{
//		const float nDotOut = getNormal().dot( outDirection );
//		if( nDotOut < 0.0f )
//		{
//			outDirection = outDirection.getVector() - (getNormal().getVector() * nDotOut);
//			if( outDirection.isZero() )
//			{
//				outDirection = getTangentX();
//			}
//		}
//	}
}


inline
void BssdfSchlick::sdfSpecular
(
	const Vector2f01&   stochastic,
	const UnitVector3f& inDirection,
	UnitVector3f&       outDirection
) const
{
	/// precondition: getNormal().dot( inDirection ) >= 0.0f

	outDirection = inDirection;
	outDirection.reflectEq( getNormal() );
}
