/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "LightStream.h"
#include "Ray.h"
#include "Intersection.h"
#include "Qualities.h"
#include "ObjectRenderable.h"
#include "UnitVector3f.h"
#include "Polar.h"
#include "hxa7241math.h"
#include "FpToInt.h"
#include "HistogramSmall.h"
#include "Histogram.h"
#include <iostream>

#include "SurfacePoint.h"   /// own header is included last


using namespace perceptuum2::interaction;




//#define ENABLE_TEST_CODE




/// standard object services -----------------------------------------------------------------------
SurfacePoint::SurfacePoint()
{
}


SurfacePoint::SurfacePoint
(
	const Intersection&       intersection,
	const ObjectRenderableId& outsideObject
)
{
	SurfacePoint::setSurfacePoint( intersection, outsideObject );
}


SurfacePoint::~SurfacePoint()
{
}


SurfacePoint::SurfacePoint( const SurfacePoint& other )
 :	object_m             ( other.object_m ),
	outsideObject_m      ( other.outsideObject_m ),
	position_m           ( other.position_m ),
	normalUnmodified_m   ( other.normalUnmodified_m ),
	emitted_m            ( other.emitted_m ),
	brdf_m               ( other.brdf_m )
	//emittedDistribution_m( other.emittedDistribution_m )
{
}


SurfacePoint& SurfacePoint::operator=( const SurfacePoint& other )
{
	assign( other );

	return *this;
}




/// streaming --------------------------------------------------------------------------------------
std::ostream& perceptuum2::interaction::operator<<( std::ostream& out, const SurfacePoint& obj )
{
	return out << "{obj:" << obj.object_m << ", outobj:" << obj.outsideObject_m << ", pos:" << obj.position_m <<
	              ", nrmun:" << obj.normalUnmodified_m << ", emi:" << obj.emitted_m << ", " << "Bssdf }";
}




/// commands ---------------------------------------------------------------------------------------
void SurfacePoint::setSurfacePoint
(
	const Intersection&       intersection,
	const ObjectRenderableId& outsideObject
)
{
	object_m           = intersection.object_m;
	outsideObject_m    = outsideObject;
	position_m         = intersection.position_m;
	normalUnmodified_m = intersection.normal_m;

	Qualities qualities;
	object_m->getQualities( intersection, qualities );

	setQualities( object_m, outsideObject_m, position_m, qualities );
}


void SurfacePoint::setQualities
(
	const ObjectRenderableId& ,//object,
	const ObjectRenderableId& outsideObject,
	const Vector3f&           position,
	const Qualities&          qualities
)
{
	emitted_m = qualities.emitted_m;

	/// adjust transparency so the total scaling in one side and out the other combines to (roughly) equal to qualities.transparency_m
	const float transparency = qualities.transparency_m;//float(::sqrt( qualities.transparency_m ));

	brdf_m.setBrdf( qualities.normal_m, qualities.tangent_m );
	brdf_m.setBssdf( outsideObject->getRefractiveIndex( outsideObject, position ), qualities.refractiveIndex_m, transparency );
//	brdf_m.setBssdfPhong( qualities.reflectivitySpecular_m.average() + qualities.reflectivityDiffuse_m.average(),
//	                      Light( qualities.reflectivityDiffuse_m ), Light( qualities.reflectivitySpecular_m ),
//	                      (qualities.roughnessX_m + qualities.roughnessY_m) * 0.5f );
	brdf_m.setBssdfWard( Light( qualities.reflectivityDiffuse_m ), Light( qualities.reflectivitySpecular_m ),
	                     qualities.roughnessX_m, qualities.roughnessY_m );
	//emittedDistribution_m.setBrdf( qualities.normal_m, qualities.tangent_m );
	//emittedDistribution_m.setBssdfPhong( 1.0f, Light::ONE(), Light::ONE(), qualities.emittedRoughness_m );
}




/// queries ----------------------------------------------------------------------------------------
const Vector3f& SurfacePoint::getPosition() const
{
	return position_m;
}


const ObjectRenderableId& SurfacePoint::getObject() const
{
	return object_m;
}


const ObjectRenderableId& SurfacePoint::getObject
(
	const Vector3f& direction
) const
{
	return (direction.dot( normalUnmodified_m.getVector() ) >= 0.0f) ? outsideObject_m : object_m;
}


void SurfacePoint::getRay
(
	const Vector3f& toPosition,
	Ray&            ray
) const
{
	ray.set( position_m, UnitVector3f(toPosition - position_m), normalUnmodified_m, object_m );
}


void SurfacePoint::getRay
(
	const UnitVector3f& direction,
	const bool          isTransmitted,
	Ray&                ray
) const
{
	ray.set( position_m, direction, normalUnmodified_m, object_m, isTransmitted );
}


bool SurfacePoint::getEmittedRay
(
	const Qualities&    qualities,
	const Vector2f01&   stochastic,
	const UnitVector3f& normalUnmodified,
	const bool          isCosStochastic,
	UnitVector3f&       direction,
	Light&              light,
	bool&               isDiffuse
)
{
	UnitVector3f normal;
	UnitVector3f tangentX;
	UnitVector3f tangentY;
	makeCoordFrame( qualities.normal_m, qualities.tangent_m, normal, tangentX, tangentY );

	/// make out direction
	{
		float x, y, z;
		if( isCosStochastic )
		{
			hxa7241::graphics::squareToDisk( stochastic.getX(), stochastic.getY(), x, y );
			z = ::sqrt( 1.0f - ((x * x) + (y * y)) );
		}
		else
		{
			hxa7241::graphics::squareToHemisphere( stochastic.getX(), stochastic.getY(), x, y, z );
		}

		hxa7241::graphics::Vector3f d;
		d  = (tangentX.getVector() * x);
		d += (tangentY.getVector() * y);
		d += (normal.getVector()   * z);

		direction = d;
	}

	/// make out light
	{
		light = qualities.emitted_m;

		if( !isCosStochastic )
		{
			/// (extra 1/1.05 is to roughly compensate for numerical integration, which always seems to be ~5% too large ...(?))
			/// (should now be ~<1% too small)
			static const float NORMALISATION = 1.0 / ((3.0 / Brdf::TWO_PI) * 1.05);

			light *= (float(::fabs(normal.dot( direction ))) * NORMALISATION);
		}
	}

	isDiffuse = true;

	/// clamp to true surface
	clampToSurface( normalUnmodified, normal, tangentX, direction );

	return false;



//	BssdfPhong emittanceDistribution( qualities.normal_m, qualities.tangent_m,
//	                                  1.0f, 1.0f, 0.0f,
//	                                  1.0f, Light::ONE(), Light::ONE(), qualities.emittedRoughness_m );
//
//	Light   scaling;
//	Float01 r;
//	const bool isTransmitted = emittanceDistribution.getSample( 1, 0.0f, stochastic, -qualities.normal_m, direction, scaling, r );
//
//	(light = qualities.emitted_m) *= scaling;
//
//	/// clamp to true surface
//	clampToSurface( normalUnmodified, emittanceDistribution.getNormal(), emittanceDistribution.getTangentX(), direction );
//
//	return isTransmitted;
}


/*
/// unreviewed, untested, unsure about use
/// pending further consideration

void SurfacePoint::makeImportanceHistogram
(
	const vector<LightStream>& illuminations,
	const UnitVector3f&        inDirection,
	vector<float>&             gridBuffer,
	Histogram&                 importance
) const
{
	/// make two square grids of side 2 * sqrt(photons count)
	const dword gridWidth = hxa7241::utility::fpToInt(2.0f * float(::floor(::sqrt( illuminations.size() ))));
	gridBuffer.resize( (gridWidth * gridWidth) << 1 );

	/// accumulate inward light distribution to grid
	for( dword i = illuminations.size();  i-- > 0; )
	{
		const LightStream& illumination = illuminations[i];

		/// map photon direction to grid coords
		dword xGrid;
		dword yGrid;
		bool  isUpper;
		{
			Vector2f01 xy01;
			brdf_m.getCoordsFromDirection( -illumination.direction_m, xy01, isUpper );

			xGrid = hxa7241::utility::fpToIntTrunc(xy01.getX() * float(gridWidth));
			if( xGrid >= gridWidth )
			{
				xGrid = gridWidth - 1;
			}
			yGrid = hxa7241::utility::fpToIntTrunc(xy01.getY() * float(gridWidth));
			if( yGrid >= gridWidth )
			{
				yGrid = gridWidth - 1;
			}
		}

		/// add photon power to grid cell at coords
		const dword side = dword(!isUpper) * (gridBuffer.size() >> 1);
		const dword cell = (yGrid * gridWidth) + xGrid;
		gridBuffer[ side + cell ] += illumination.light_m.average();
	}

	/// scale grid by scatering distribution
	for( dword s = 1;  s-- > 0; )
	{
		const dword side = s * (gridBuffer.size() >> 1);
		for( dword y = gridWidth;  y-- > 0; )
		{
			for( dword x = gridWidth;  x-- > 0; )
			{
				UnitVector3f outDirection;
				{
					const Vector2f01 xy01( (float(x) + 0.5f) / float(gridWidth),
					                       (float(y) + 0.5f) / float(gridWidth) );
					const bool       isUpper = (s == 0);
					brdf_m.getDirectionFromCoords( xy01, isUpper, outDirection );
				}

				Light bssdfScaling;
				getPathStepScaling( inDirection, outDirection, bssdfScaling );

				const dword cell = (y * gridWidth) + x;
				gridBuffer[ side + cell ] *= bssdfScaling.average();
			}
		}
	}

	/// put grid into histogram
	for( dword i = gridBuffer.size();  i-- > 0; )
	{
		/// add cell power to histogram
		importance.addElement( gridBuffer[i] );
	}
}


bool SurfacePoint::getSamplingDirection
(
	const Histogram&    importance,
	const Float01       stochastics[3],
	const UnitVector3f& inDirection,
	UnitVector3f&       direction,
	Light&              scaling
) const
{
	bool isTransmitted;

	const dword gridLength = importance.getSize() >> 1;

	/// choose grid cell using histogram
	dword cellIndex;
	const bool isFound = importance.findElement( stochastics[0], cellIndex );
	if( isFound & (gridLength > 0) )
	{
		const dword gridWidth = hxa7241::utility::fpToIntTrunc(float(::sqrt( gridLength )));

		/// map cell index to direction with uniform mapping and cell dimensions
		const bool   isUpperGrid = cellIndex < gridLength;
		UnitVector3f outDirection;
		{
			/// map cell index to grid coords
			const dword gridOffset       = cellIndex - (dword(!isUpperGrid) * gridLength);
			const dword gridY            = gridOffset / gridWidth;
			const dword gridX            = gridOffset - (gridY * gridWidth);
			const float oneOverGridWidth = 1.0f / float(gridWidth);
			const Vector2f01 xy01( (float(gridX) + stochastics[1]) * oneOverGridWidth,
			                       (float(gridY) + stochastics[2]) * oneOverGridWidth);

			brdf_m.getDirectionFromCoords( xy01, isUpperGrid, outDirection );
		}

		isTransmitted = isUpperGrid ^ (inDirection.dot( brdf_m.getNormal() ) < 0.0f);

		UnitVector3f outDirClamped( outDirection );
		clampToTrueSurface( outDirClamped );

		/// set direction
		direction = outDirClamped;

		/// set light scaling to (1 / probability) * (cell area / whole grid area) * reflectivity
		(scaling = Light::ONE()) *= (brdf_m.getReflectivity() / (importance.getElement(cellIndex) * float(gridLength)));
	}
	else
	{
		direction     = UnitVector3f::ZERO();
		scaling       = Light::ZERO();
		isTransmitted = false;
	}

	return isTransmitted;
}*/




/// implementation ---------------------------------------------------------------------------------
void SurfacePoint::getPathStep
(
	const dword         conditions,
	const Float01       stochastics[7],
	const UnitVector3f& inDirection,
	Ray&                ray,
	Light&              scaling,
	dword&              status
) const
{
	status = 0;
	bool isStepMade = false;

	/// maybe roulette reflectivity
	const Float01 reflectivity = brdf_m.getReflectivity();
	if( !bool(conditions & IS_ROULETTE) | (stochastics[2] < reflectivity) )
	{
		/// choose part
		dword part;
		bool  isPartDiffuse;
		{
			float   proportions[brdf_m.PARTITIONS_MAX];
			Float01 peakness[brdf_m.PARTITIONS_MAX];
			dword   partsCount;
			brdf_m.getPartitions( proportions, peakness, partsCount );

			/// effectively apply a simple gain function to proportions
			/// - to eliminate isolated infrequent stochastic effects
			for( dword i = partsCount;  i-- > 0; )
			{
				if( proportions[i] < 0.1f )
				{
					proportions[i] = 0.0f;
				}
			}

			const hxa7241::utility::HistogramSmall hp( partsCount, proportions );
			part = hp.getChoice( stochastics[3] );

			/// apply a simple gain function to peakness
			/// - to eliminate isolated infrequent stochastic effects
			float peaknessGained = peakness[part];
			if( peaknessGained > 0.9f )
			{
				peaknessGained = 1.0f;
			}
			else if( peaknessGained < 0.1f )
			{
				peaknessGained = 0.0f;
			}

			isPartDiffuse = bool(stochastics[4] > peaknessGained);

			const bool isDirectSeparate = stochastics[4] > (1.0f - ((1.0f - peakness[part]) * 100.0f));
			status |= (~(dword(isDirectSeparate)-1) & IS_DIRECT_SEPARATE);
		}

		if( bool(conditions & IS_DIFFUSE_STEP_ALLOWED) | !isPartDiffuse )
		{
			UnitVector3f inDirClamped( inDirection );
			clampToFakeSurface( inDirClamped );

			/// get out direction
			UnitVector3f outDirection;
			Light        outScaling;
			const bool isTransmitted = brdf_m.getSample( part, stochastics[5], Vector2f01( stochastics[0], stochastics[1] ),
			                                             inDirClamped, outDirection, outScaling );

			if( !outDirection.isZero() )
			{
				/// maybe roulette cos
				const Float01 cos = brdf_m.getCos( part, outDirection );
				if( !bool(conditions & IS_ROULETTE) | (stochastics[6] < cos) )
				{
					UnitVector3f outDirClamped( outDirection );
					clampToTrueSurface( outDirClamped );

					/// make output
					getRay( outDirClamped, isTransmitted, ray );
					scaling = outScaling;
					if( !bool(conditions & IS_ROULETTE) )
					{
						scaling *= (reflectivity * cos);
					}
					status |= (~(dword(isPartDiffuse)-1) & IS_DIFFUSE);

					isStepMade = true;
				}
			}
		}
	}

	/// if step not made, zero output
	if( !isStepMade )
	{
		getRay( UnitVector3f::ZERO(), false, ray );
		scaling  = Light::ZERO();
		status  |= IS_DIFFUSE;
		status &= ~IS_DIRECT_SEPARATE;
	}
}


void SurfacePoint::getPathStepScaling
(
	const UnitVector3f& inDirection,
	const UnitVector3f& outDirection,
	Light&              scaling
) const
{
	scaling = Light::ZERO();

	float   proportions[brdf_m.PARTITIONS_MAX];
	Float01 peakness[brdf_m.PARTITIONS_MAX];
	dword   partsCount;
	brdf_m.getPartitions( proportions, peakness, partsCount );

	for( dword p = partsCount;  p-- > 0 ; )
	{
		Light partScaling;
		brdf_m.getScaling( p, inDirection, outDirection, partScaling );

		const float cos = brdf_m.getCos( p, outDirection );

		scaling += (partScaling *= (proportions[p] * cos));
	}

	scaling *= brdf_m.getReflectivity();
}


void SurfacePoint::getEmittedLight
(
	const UnitVector3f& outDirection,
	Light&              emitted
) const
{
	getEmittedDiffuseLight( emitted_m, outDirection, emitted );
}


void SurfacePoint::getEmittedDiffuseLight
(
	const Light&        emittance,
	const UnitVector3f& outDirection,
	Light&              outLight
) const
{
	/// emitted light = (emittance * edf diffuse parts * cos outs)

	/// (extra 1/1.05 is to roughly compensate for numerical integration, which always seems to be ~5% too large ...(?))
	/// (should now be ~<1% too small)
	static const float NORMALISATION = 1.0 / ((3.0 / Brdf::TWO_PI) * 1.05);

	UnitVector3f outDirClamped( outDirection );
	clampNearToFakeSurface( outDirClamped );

	float cos = brdf_m.getNormal().dot( outDirClamped );
	if( cos < 0.0f )
	{
		cos = -cos;
	}

	(outLight = emittance) *= (Brdf::ONE_OVER_TWO_PI * NORMALISATION * cos);
}




const Brdf& SurfacePoint::getBrdf() const
{
	return brdf_m;
}


float SurfacePoint::getDistance2
(
	const Vector3f& position
) const
{
	Vector3f dif( position_m );
	dif -= position;

	float distance2 = dif.dot( dif );
	distance2 = distance2 >= 0.000001f ? distance2 : 0.000001f;

	return distance2;
}


void SurfacePoint::makeCoordFrame
(
	const UnitVector3f& normal,
	const UnitVector3f& tangent,
	UnitVector3f&       normal_,
	UnitVector3f&       tangentX,
	UnitVector3f&       tangentY
)
{
	if( !normal.isZero() )
	{
		normal_ = normal;
	}
	else
	{
		/// default normal to -Y
		normal_ = -UnitVector3f::Y();
	}

	if( !tangent.isZero() )
	{
		tangentX = tangent;
	}
	else
	{
		/// default tangent to X
		tangentX = UnitVector3f::X();
	}


	tangentX.orthogonalizeEq( normal_ );

	tangentY = tangentX;
	tangentY.crossEq( normal_ );
}


bool SurfacePoint::isFalseCrossing
(
	const UnitVector3f& dir
) const
{
	return (dir.dot( normalUnmodified_m ) >= 0.0f) != (dir.dot( brdf_m.getNormal() ) >= 0.0f);
}


bool SurfacePoint::clampToFakeSurface
(
	UnitVector3f& direction
) const
{
	return clampToSurface( brdf_m.getNormal(), normalUnmodified_m, brdf_m.getTangentX(), direction );
}


bool SurfacePoint::clampToTrueSurface
(
	UnitVector3f& direction
) const
{
	return clampToSurface( normalUnmodified_m, brdf_m.getNormal(), brdf_m.getTangentX(), direction );
}


bool SurfacePoint::clampNearToFakeSurface
(
	UnitVector3f& direction
) const
{
	const UnitVector3f& normalFake = brdf_m.getNormal();

	/// check for false crossing
	const float fnDotDir      = direction.dot( normalFake );
	const float tnDotDir      = direction.dot( normalUnmodified_m );
	const bool  isToBeClamped = (fnDotDir >= 0.0f) != (tnDotDir >= 0.0f);

	if( isToBeClamped )
	{
		/// clamp
		direction = direction.getVector() - (normalFake.getVector() * fnDotDir);
		if( direction.isZero() )
		{
			direction = brdf_m.getTangentX();
		}

		/// nudge to one side of surface
		{
			/// 1/1000 degree
			static const float COSSMALL = 0.99999999984769129011051202417815f;
			static const float SINSMALL = 1.7453292519057199613549105685129e-5f;
			const float sinSmallSigned = (tnDotDir > 0.0f) ? SINSMALL : -SINSMALL;

			direction = ((direction.getVector() * COSSMALL) += (normalFake.getVector() * sinSmallSigned));
		}
	}

	return isToBeClamped;
}


bool SurfacePoint::clampToSurface
(
	const UnitVector3f& normalTrue,
	const UnitVector3f& normalFake,
	const UnitVector3f& tangentTrue,
	UnitVector3f&       direction
)
{
	/// check for false crossing
	const float nDotDir       = direction.dot( normalTrue );
	const bool  isToBeClamped = (nDotDir >= 0.0f) != (direction.dot( normalFake ) >= 0.0f);

	if( isToBeClamped )
	{
		/// clamp
		direction = direction.getVector() - (normalTrue.getVector() * nDotDir);
		if( direction.isZero() )
		{
			direction = tangentTrue;
		}
	}

	return isToBeClamped;
}


void SurfacePoint::assign
(
	const SurfacePoint& other
)
{
	if( &other != this )
	{
		object_m              = other.object_m;
		outsideObject_m       = other.outsideObject_m;
		position_m            = other.position_m;
		normalUnmodified_m    = other.normalUnmodified_m;
		emitted_m             = other.emitted_m;
		brdf_m                = other.brdf_m;
		//emittedDistribution_m = other.emittedDistribution_m;
	}
}




/// testing ----------------------------------------------------------------------------------------
#include "Logger.h"
#include "RandomGood.h"
#include <time>


#ifndef ENABLE_TEST_CODE
bool SurfacePoint::test
(
	const bool isLogging
)
{
	if(isLogging) hxa7241::utility::Logger() << "SurfacePoint::test  [disabled]";

	return false;
}


#else
bool SurfacePoint::test
(
	const bool isLogging
)
{
	static const char PREFIX[] = "SurfacePoint::  ";

	bool isSucceeded = true;

	if(isLogging) hxa7241::utility::Logger() << PREFIX << "=== general test ===";


	SurfacePoint sp;

	/// check emitted ray
	{
		static const dword RESOLUTION          = 128;
		static const float ONE_OVER_RESOLUTION = 1.0f / float(RESOLUTION);

		Qualities qs;
		qs.normal_m  = sp.brdf_m.getNormal();
		qs.tangent_m = sp.brdf_m.getTangentX();
		qs.emitted_m = CieXyz::ONE();

		vector<float> grid( RESOLUTION * RESOLUTION );

		const dword seed = dword(std::time(0));
		hxa7241::utility::RandomGood rand( seed );

		static const dword ITERATIONS = 1048576;
		for( dword i = ITERATIONS;  i-- > 0; )
		{
			const float r0 = rand.getFloat();
			rand.next();
			const float r1 = rand.getFloat();
			rand.next();
			const Vector2f01 stochastic( r0, r1 );

			UnitVector3f direction;
			Light        light;
			bool         isDiffuse;
			getEmittedRay( qs, stochastic, qs.normal_m, true,
						   direction, light, isDiffuse );

			Vector2f01 coords;
			sp.brdf_m.get2DCoordsFromDirection( direction, coords );

			dword gridX = dword(coords.getX() * float(RESOLUTION));
			dword gridY = dword(coords.getY() * float(RESOLUTION));
			gridX = gridX < RESOLUTION ? gridX : RESOLUTION;
			gridY = gridY < RESOLUTION ? gridY : RESOLUTION;

			grid[gridX + (gridY * RESOLUTION)] += light.average() / float(ITERATIONS);
		}


		float outLightSum = 0.0f;
		float gridSum     = 0.0f;
//		float difSum      = 0.0f;
//		float difAbsSum   = 0.0f;
//		float difAbsMax   = 0.0f;
//{hxa7241::utility::Logger l_;
//l_ << "\n";
		for( dword yOut = RESOLUTION;  yOut-- > 0; )
		{
			for( dword xOut = RESOLUTION;  xOut-- > 0; )
			{
				const Vector2f01 coords( (xOut + 0.5f) * ONE_OVER_RESOLUTION,
				                         (yOut + 0.5f) * ONE_OVER_RESOLUTION );
				UnitVector3f outDirection;
				Light        outLight;
				bool         isDiffuse;
				getEmittedRay( qs, coords, qs.normal_m, false,
				               outDirection, outLight, isDiffuse );

				const float outLightFloat = outLight.average() / float(RESOLUTION * RESOLUTION);
				outLightSum += outLightFloat;

				gridSum += grid[xOut + (yOut * RESOLUTION)];
//				const float dif = ((grid[xOut + (yOut * RESOLUTION)] - outLightFloat) / outLightFloat) * 100.0f;
//				difSum    += dif;
//				difAbsSum += float(::fabs(dif));
//				difAbsMax = difAbsMax >= float(::fabs(dif)) ? difAbsMax : float(::fabs(dif));
////l_ << outLightFloat << " ";
//l_ << grid[xOut + (yOut * RESOLUTION)] << " ";
////l_ << dif << " ";
			}
//l_ << "\n";
		}
//}
		//outLightSum *= 1.0f / float(RESOLUTION * RESOLUTION);


		const float error      = (outLightSum - qs.emitted_m.average()) * 100.0f;
//		const float difMean    = difSum / float(RESOLUTION * RESOLUTION);
//		const float difAbsMean = difAbsSum / float(RESOLUTION * RESOLUTION);
		const float gridMean   = gridSum;
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "emitted: " << outLightSum << "  emitted error: " << error << "%";
//		if(isLogging) hxa7241::utility::Logger() << PREFIX << "dif mean: " << difMean << "  dif abs mean: " << difAbsMean << "  dif abs max: " << difAbsMax;
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "grid mean: " << gridMean ;

		if(isLogging) hxa7241::utility::Logger() << PREFIX << "emitted ray test ...";
	}


	/// check emittance
	{
		bool isEmittanceOk;// = true;

		static const dword RESOLUTION          = 128;
		static const float ONE_OVER_RESOLUTION = 1.0f / float(RESOLUTION);

		//const dword seed = dword(std::time(0));
		//hxa7241::utility::RandomGood rand( seed );

		const Light emittance( Light::ONE() );

		float outLightSum = 0.0f;
		float outLightMax = 0.0f;
		float outLightMin = FLOAT_MAX;
		for( dword yOut = RESOLUTION;  yOut-- > 0; )
		{
			for( dword xOut = RESOLUTION;  xOut-- > 0; )
			{
//				const float phi = ((xOut + 0.5f) * ONE_OVER_RESOLUTION) * Brdf::TWO_PI;
//				float x, z;
//				hxa7241::graphics::polarToRectangular( phi, 1.0f, x, z );
//
//				const float phi_ = (((yOut + 0.5f) * ONE_OVER_RESOLUTION) * Brdf::PI) - (Brdf::PI * 0.5f);
//				float y, h;
//				hxa7241::graphics::polarToRectangular( phi_, 1.0f, y, h );
//
//				UnitVector3f outDirection;
//				sp.brdf_m.getDirectionFromFrameCoords( Vector3f(x,z,y), outDirection );


				UnitVector3f outDirection;
				sp.brdf_m.getUniformSampleDirection( Vector2f01((xOut + 0.5f) * ONE_OVER_RESOLUTION, (yOut + 0.5f) * ONE_OVER_RESOLUTION), outDirection );
//				const float r0 = rand.getFloat();
//				rand.next();
//				const float r1 = rand.getFloat();
//				rand.next();
//				sp.brdf_m.getUniformSampleDirection( Vector2f01(r0, r1), outDirection );

				Light outLight;
				sp.getEmittedDiffuseLight( emittance, outDirection, outLight );

				const float outLightFloat = outLight.largest();
				outLightSum += outLightFloat;
				outLightMax = outLightMax < outLightFloat ? outLightFloat : outLightMax;
				outLightMin = outLightMin > outLightFloat ? outLightFloat : outLightMin;
			}
		}

		outLightSum *= Brdf::TWO_PI / float(RESOLUTION * RESOLUTION);

		const float error = (outLightSum - emittance.average()) * 100.0f;
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "emitted: " << outLightSum << "  emitted error: " << error << "%";
		if(isLogging) hxa7241::utility::Logger() << PREFIX << "emitted min: " << outLightMin << "  emitted max: " << outLightMax;

		isEmittanceOk = ::fabs(error) < 1.0f;
		if(isLogging) hxa7241::utility::Logger() << PREFIX << (isEmittanceOk  ? "--- emittance test succeeded" : "*** emittance test failed");

		isSucceeded &= isEmittanceOk;
	}


	if(isLogging) hxa7241::utility::Logger() << PREFIX << (isSucceeded  ? "--- general test succeeded" : "*** general test failed");

	return isSucceeded;
}
#endif
