/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Serialization.h"
#include "ImageReader.h"
#include "Scene.h"
#include "Polar.h"
#include "FpToInt.h"
#include "Vector3f.h"
#include "Exception.h"
#include "Logger.h"
#include "hxa7241math.h"
#include <fstream>

#include "Material.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// statics
const char Material::LOAD_FAIL_MESSAGE[] = "couldnt load texture image";




/// standard object services -----------------------------------------------------------------------
Material::Material()
 :	baseQualities_m()
{
	Material::resetImageFlags();
}


Material::~Material()
{
}


Material::Material( const Material& other )
 :	name_m                 ( other.name_m ),
	baseQualities_m        ( other.baseQualities_m ),
	imagePath_m            ( other.imagePath_m ),
	colorImageName_m       ( other.colorImageName_m ),
	colorSpecImageName_m   ( other.colorSpecImageName_m ),
	specDiffImageName_m    ( other.specDiffImageName_m ),
	roughnessXImageName_m  ( other.roughnessXImageName_m ),
	roughnessYImageName_m  ( other.roughnessYImageName_m ),
	transparencyImageName_m( other.transparencyImageName_m ),
	emittedImageName_m     ( other.emittedImageName_m ),
	tangentImageName_m     ( other.tangentImageName_m ),
	normalImageName_m      ( other.normalImageName_m )
{
	Material::resetImages();
}


Material& Material::operator=( const Material& other )
{
	if( &other != this )
	{
		name_m = other.name_m;

		baseQualities_m = other.baseQualities_m;

		imagePath_m             = other.imagePath_m;
		colorImageName_m        = other.colorImageName_m;
		colorSpecImageName_m    = other.colorSpecImageName_m;
		specDiffImageName_m     = other.specDiffImageName_m;
		roughnessXImageName_m   = other.roughnessXImageName_m;
		roughnessYImageName_m   = other.roughnessYImageName_m;
		transparencyImageName_m = other.transparencyImageName_m;
		emittedImageName_m      = other.emittedImageName_m;
		tangentImageName_m      = other.tangentImageName_m;
		normalImageName_m       = other.normalImageName_m;

		resetImages();
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void Material::setImagePath
(
	const string& path
)
{
	imagePath_m = path;
}




/// queries ----------------------------------------------------------------------------------------
void Material::getQualities
(
	const Vector2f&     barycentrics,
	const Vector2<word> cornerCoords[3],
	Qualities&          qualities
) const
{
	Vector2f coords;
	{
		/// calculate interpolation by using two edges as axes scaled by the barycentrics
		const Vector2f corner0( cornerCoords[0].getX(), cornerCoords[0].getY() );
		const Vector2f corner1( cornerCoords[1].getX(), cornerCoords[1].getY() );
		const Vector2f corner2( cornerCoords[2].getX(), cornerCoords[2].getY() );

		Vector2f vert0( corner0 );
		Vector2f edge0( corner1 - corner0 );
		Vector2f edge1( corner2 - corner0 );

		coords = ( ((edge0 *= barycentrics.getX()) += (edge1 *= barycentrics.getY())) += vert0 );
	}

	/// delegate to other method
	getQualities( coords, qualities );
}


void Material::getQualities
(
	const Vector2f& coords,
	Qualities&      qualities
) const
{
	getReflectivity( coords, qualities.reflectivityDiffuse_m, qualities.reflectivitySpecular_m );
	getRoughness   ( coords, qualities.roughnessX_m, qualities.roughnessY_m );
	getTransparency( coords, qualities.transparency_m );
	getEmitted     ( coords, qualities.emitted_m );
	getTangent     ( coords, qualities.tangent_m );
	getNormal      ( coords, qualities.normal_m );

	qualities.refractiveIndex_m  = baseQualities_m.refractiveIndex_m;
	qualities.emittedRoughness_m = baseQualities_m.emittedRoughness_m;
}


const Qualities& Material::getQualitiesBase() const
{
	return baseQualities_m;
}


Float01 Material::getTransparency
(
	const Vector3f& //position
) const
{
	return baseQualities_m.transparency_m;
}


float Material::getRefractiveIndex
(
	const Vector3f& //position
) const
{
	return baseQualities_m.refractiveIndex_m;
}


void Material::getEmittedMean
(
	const Vector2<word> coords[3],
	CieXyz&             emittedMean
) const
{
	//emittedMean = baseQualities_m.emitted_m;


	/// constant on/off implementation
	if( coords[0].isZero() )
	{
		emittedMean = baseQualities_m.emitted_m;
	}
	else
	{
		emittedMean = CieXyz::ZERO();
	}


	/// scan through texels like a rasterization, averaging values instead of drawing
}


void Material::getSpecularnessMean
(
	const Vector2<word> coords[3],
	Float01&            specularnessMean
)  const
{
	/// scan triangle for texel stats
	/// coverage expanded to whole texels

	float sum   = 0.0f;
	dword count = 0;
	if( !(roughnessXImageName_m.empty() & roughnessYImageName_m.empty() &
	      colorImageName_m.empty() & colorSpecImageName_m.empty()) )
	{
		const Vector2f HALFISH( 0.5f + FLOAT_EPSILON, 0.5f + FLOAT_EPSILON );

		/// handle degenerate triangle point special case
		if( (coords[0] == coords[1]) && (coords[1] == coords[2]) )
		{
			Float01 specularness;
			getSpecularness( Vector2f(float(coords[0].getX()), float(coords[0].getY())) += HALFISH, specularness );

			sum   = specularness;
			count = 1;
		}
		else
		{
			/// sort triangle corners
			const Vector2<word>* pCorner0;
			const Vector2<word>* pCorner1;
			const Vector2<word>* pCorner2;
			{
				const dword compare01y = coords[0].getY() - coords[1].getY();
				const dword compare12y = coords[1].getY() - coords[2].getY();
				const dword compare20y = coords[2].getY() - coords[0].getY();
				const dword _0above1   = (compare01y < 0) | ((compare01y == 0) & (coords[0].getX() <= coords[1].getX()));
				const dword _1above2   = (compare12y < 0) | ((compare12y == 0) & (coords[1].getX() <= coords[2].getX()));
				const dword _2above0   = (compare20y < 0) | ((compare20y == 0) & (coords[2].getX() <= coords[0].getX()));
				const dword tableShift = ((_0above1 << 2) | (_1above2 << 1) | (_2above0)) << 1;

				/// lookup tables
				/// 0a1 1a2 2a0   top mid bot
				/// 0   0   x     2   1   0
				/// 0   0   1     2   1   0
				/// 0   1   0     1   0   2
				/// 0   1   1     1   2   0
				/// 1   0   0     0   2   1
				/// 1   0   1     2   0   1
				/// 1   1   0     0   1   2
				/// 1   1   x     0   1   2
				const dword TOP = 0x085A;
				const dword MID = 0x5285;
				const dword BOT = 0xA520;

				pCorner0 = &(coords[ (TOP >> tableShift) & 0x03 ]);
				pCorner1 = &(coords[ (MID >> tableShift) & 0x03 ]);
				pCorner2 = &(coords[ (BOT >> tableShift) & 0x03 ]);

				///// assert sort postcondition
				//if( ((pCorner0->getY() > pCorner1->getY()) |
				//	 ((pCorner0->getY() == pCorner1->getY()) & (pCorner0->getX() > pCorner1->getX()))) |
				//	((pCorner1->getY() > pCorner2->getY()) |
				//	 ((pCorner1->getY() == pCorner2->getY()) & (pCorner1->getX() > pCorner2->getX()))) |
				//	((pCorner0->getY() > pCorner2->getY()) |
				//	 ((pCorner0->getY() == pCorner2->getY()) & (pCorner0->getX() > pCorner2->getX()))) )
				//{
				//	hxa7241::utility::Logger() << "+++ FAULT:  corner sort assertion failed";
				//}
			}


			/// calc gradients
			float grad01 = FLOAT_MAX;
			{
				const dword dy01 = pCorner1->getY() - pCorner0->getY();
				if( dy01 != 0 )
				{
					grad01 = float(pCorner1->getX() - pCorner0->getX()) / float(dy01);
				}
			}
			float grad02 = FLOAT_MAX;
			{
				const dword dy02 = pCorner2->getY() - pCorner0->getY();
				if( dy02 != 0 )
				{
					grad02 = float(pCorner2->getX() - pCorner0->getX()) / float(dy02);
				}
			}
			float grad12 = FLOAT_MAX;
			{
				const dword dy12 = pCorner2->getY() - pCorner1->getY();
				if( dy12 != 0 )
				{
					grad12 = float(pCorner2->getX() - pCorner1->getX()) / float(dy12);
				}
			}


			/// make two trapezoids, >= 1 wide and high
			const dword trapStarts[2] = { pCorner0->getY(), pCorner1->getY() };
			const dword trapEnds[2]   = { pCorner1->getY(), pCorner2->getY() + (pCorner2->getY() > pCorner0->getY() ? 0 : 1) };

			const float grads0112[2] = { grad01, grad12 };
			const float grads0202[2] = { grad02, grad02 };
			      float starts01[2]  = { pCorner0->getX(), pCorner1->getX() };
			      float starts02[2]  = { pCorner0->getX(), pCorner0->getX() + (grad02 * float(trapStarts[1] - trapStarts[0])) };

			const float* leftGrads   = grads0112;
			const float* rightGrads  = grads0202;
			      float* leftStarts  = starts01;
			      float* rightStarts = starts02;
			if( rightGrads[0] < leftGrads[0] )
			{
				leftGrads   = grads0202;
				rightGrads  = grads0112;
				leftStarts  = starts02;
				rightStarts = starts01;
			}

			leftStarts[0]  += leftGrads[0]  < 0.0f ? leftGrads[0]  : 0.0f;
			leftStarts[1]  += leftGrads[1]  < 0.0f ? leftGrads[1]  : 0.0f;
			rightStarts[0] += rightGrads[0] > 0.0f ? rightGrads[0] : 0.0f;
			rightStarts[1] += rightGrads[1] > 0.0f ? rightGrads[1] : 0.0f;

			/// handle degenerate triangle horizontal line special case
			if( pCorner0->getY() == pCorner2->getY() )
			{
				leftStarts[1]  = float(pCorner0->getX());
				rightStarts[1] = float(pCorner2->getX());
			}

			/// handle degenerate triangle vertical line special case
			if( (pCorner0->getX() == pCorner1->getX()) & (pCorner0->getX() == pCorner2->getX()) )
			{
				rightStarts[0] = float(pCorner0->getX() + 1);
				rightStarts[1] = rightStarts[0];
			}

			/// step through trapezoids
			for( dword t = 0;  t < 2;  ++t )
			{
				float left       = leftStarts[t];
				float right      = rightStarts[t];
				float leftGrad   = leftGrads[t];
				float rightGrad  = rightGrads[t];

				dword y          = trapStarts[t];
				const dword yEnd = trapEnds[t];

				/// step through rows
				for( ;  y < yEnd;  ++y )
				{
					dword x          = hxa7241::utility::fpToInt( float( std::floor( left ) ) );
					const dword xEnd = hxa7241::utility::fpToInt( float( std::ceil( right ) ) );

					/// increment texel count
					count += xEnd - x;

					/// step through texels
					for( ;  x < xEnd;  ++x )
					{
						/// sum texel value
						Float01 specularness;
						getSpecularness( Vector2f(float(x), float(y)) += HALFISH, specularness );
						sum += specularness;
					}

					left  += leftGrad;
					right += rightGrad;
				}
			}
		}
	}
	else
	{
		sum   = baseQualities_m.getSpecularness();
		count = 1;
	}

	/// calculate mean
	if( count > 0 )
	{
		specularnessMean = sum / float(count);
	}
	else
	{
		specularnessMean = Float01();
	}
}


void Material::getEmittedPosition
(
	const Vector2<word> [3],//coords[3],
	const Vector3f&     posRands,
	vector<float>&      ,//storage,
	Vector2f&           barycentrics
) const
{
	float x = posRands.getX();
	float y = posRands.getY();
	if( (x + y) > 1.0f )
	{
		x = 1.0f - x;
		y = 1.0f - y;
	}

	barycentrics.setXY( x, y );
}




/// implementation ---------------------------------------------------------------------------------
void Material::getReflectivity
(
	const Vector2f& coords,
	CieXyz&         rDiffuse,
	CieXyz&         rSpecular
) const
{
	/// colors
	const bool isColor     = getTexelColor( coords, imagePath_m, colorImageName_m,     color_m,     colorConverter_m, isColorLoadFailed_m,     rDiffuse );
	const bool isColorSpec = getTexelColor( coords, imagePath_m, colorSpecImageName_m, colorSpec_m, colorConverter_m, isColorSpecLoadFailed_m, rSpecular );

	/// if one missing, set both to the one supplied
	if( isColor != isColorSpec )
	{
		if( isColor )
		{
			rSpecular = rDiffuse;
		}
		else
		{
			rDiffuse = rSpecular;
		}
	}
	/// if both missing, use base qualities
	else if( !(isColor | isColorSpec) )
	{
		rDiffuse  = baseQualities_m.reflectivityDiffuse_m;
		rSpecular = baseQualities_m.reflectivitySpecular_m;
	}


	/// diffuse/specular ratio
	if( isColor | isColorSpec )
	{
		float ratio;// = 1.0f;
		{
			Float01 specDiffRatio;
			const bool isSpecDiff = getTexel( coords, imagePath_m, specDiffImageName_m, specDiff_m, isSpecDiffLoadFailed_m, specDiffRatio );
			if( isSpecDiff )
			{
				ratio = Float01::moveUp( specDiffRatio );
			}
			else
			{
				/// make ratio from base qualities reflectivities
				const float diffMean = baseQualities_m.reflectivityDiffuse_m.average();
				const float specMean = baseQualities_m.reflectivitySpecular_m.average();
				const float sum      = diffMean + specMean;

				if( sum != 0.0f )
				{
					const float oneOverSum = 1.0f / sum;
					ratio = (diffMean * oneOverSum);
				}
				else
				{
					ratio = 1.0f;
				}
			}
		}

		rDiffuse  *= ratio;
		rSpecular *= (1.0f - ratio);
	}
}


void Material::getRoughness
(
	const Vector2f& coords,
	Float01&        roughnessX,
	Float01&        roughnessY
) const
{
	const bool isRoughnessX = getTexel( coords, imagePath_m, roughnessXImageName_m, roughnessX_m, isRoughnessXLoadFailed_m, roughnessX );
	const bool isRoughnessY = getTexel( coords, imagePath_m, roughnessYImageName_m, roughnessY_m, isRoughnessYLoadFailed_m, roughnessY );

	/// if one missing, set both to the one supplied
	if( isRoughnessX != isRoughnessY )
	{
		if( isRoughnessX )
		{
			roughnessY = roughnessX;
		}
		else
		{
			roughnessX = roughnessY;
		}
	}
	/// if both missing, use base qualities
	else if( !(isRoughnessX | isRoughnessY) )
	{
		roughnessX = baseQualities_m.roughnessX_m;
		roughnessY = baseQualities_m.roughnessY_m;
	}
}


void Material::getTransparency
(
	const Vector2f& coords,
	Float01&        transparency
) const
{
	const bool isTransparency = getTexel( coords, imagePath_m, transparencyImageName_m, transparency_m, isTransparencyLoadFailed_m, transparency );
	/// if missing, use base qualities
	if( !isTransparency )
	{
		transparency = baseQualities_m.transparency_m;
	}
}


void Material::getEmitted
(
	const Vector2f& coords,
	CieXyz&         emitted
) const
{
	//emitted = baseQualities_m.emitted_m;


	/// constant on/off implementation
	if( coords.isZero() )
	{
		emitted = baseQualities_m.emitted_m;
	}
	else
	{
		emitted = CieXyz::ZERO();
	}


/// textured implementation
/// - not used at present...
/// needs getEmittedMean and getEmittedPosition to have textured implementation
/// for the class to work consistently

//	Float01 emittedMask;
//	const bool isEmitted = getTexel( coords, imagePath_m, emittedImageName_m, emitted_m, isEmittedLoadFailed_m, emittedMask );
//
//	emitted = baseQualities_m.emitted_m;
//
//	/// if present, use texel to scale base qualities
//	if( isEmitted )
//	{
//		emitted *= Float01::moveUp( emittedMask );
//	}
}


void Material::getTangent
(
	const Vector2f& coords,
	UnitVector3f&   tangent
) const
{
	UnitVector3f normal;
	const bool isTangent = getTexelVector( coords, imagePath_m, tangentImageName_m, tangent_m, isTangentLoadFailed_m, normal );
	if( isTangent )
	{
		/// rotate 90 deg in plane, to align tangent with texture 'contours'
		tangent.setXYZ( normal.getZ(), 0, -(normal.getX()) );
	}
	else
	{
		tangent = UnitVector3f::ZERO();
	}
}


void Material::getNormal
(
	const Vector2f& coords,
	UnitVector3f&   normal
) const
{
	const bool isNormal = getTexelVector( coords, imagePath_m, normalImageName_m, normal_m, isNormalLoadFailed_m, normal );
	if( !isNormal )
	{
		normal = baseQualities_m.normal_m;
	}
}


void Material::getSpecularness
(
	const Vector2f& coords,
	Float01&        specularness
) const
{
	Float01 roughnessX;
	Float01 roughnessY;
	getRoughness( coords, roughnessX, roughnessY );

	CieXyz reflectivityDiffuse;
	CieXyz reflectivitySpecular;
	getReflectivity( coords, reflectivityDiffuse, reflectivitySpecular );

	specularness = Qualities::getSpecularness( roughnessX, roughnessY, reflectivitySpecular );
}




bool Material::getTexelColor
(
	const Vector2f& coords,
	const string&   imagePath,
	const string&   imageName,
	ImageRgba&      image,
	ImageConverter& converter,
	bool&           isLoadFailed,
	CieXyz&         color
)
{
	const bool isLoaded = loadImage( imagePath, imageName, 0, &image, &converter, isLoadFailed );
	if( isLoaded )
	{
		/// get interpolant info
		dword    smallerOffset;
		Vector2i increment;
		Vector2f subPosition;
		makeInterpolantPositions( coords, Vector2f( image.getWidth(), image.getHeight() ), smallerOffset, increment, subPosition );

		/// interpolate surrounding four elements bilinearly
		using hxa7241::graphics::Vector3f;
		Vector3f value;
		byte     alpha;
		{
			const dword* pSmaller  = image.getMemory() + smallerOffset;

			/// convert elements into rgb fp (unassociated alpha)
			dword rgba[4];

			image.fromDword( *pSmaller, rgba );
			Vector3f a( rgba[0], rgba[1], rgba[2] );
			float    aa( rgba[3] );

			image.fromDword( *(pSmaller + increment.getX()), rgba );
			Vector3f b( rgba[0], rgba[1], rgba[2] );
			float    ba( rgba[3] );

			image.fromDword( *(pSmaller + increment.getY()), rgba );
			Vector3f c( rgba[0], rgba[1], rgba[2] );
			float    ca( rgba[3] );

			image.fromDword( *(pSmaller + increment.getY() + increment.getX()), rgba );
			Vector3f d( rgba[0], rgba[1], rgba[2] );
			float    da( rgba[3] );

			((b -= a) *= subPosition.getX()) += a;
			((d -= c) *= subPosition.getX()) += c;
			(((value = d) -= b) *= subPosition.getY()) += b;
			value *= hxa7241::utility::FRACTIONALIZER::_8BIT;

			ba = ((ba - aa) * subPosition.getX()) + aa;
			da = ((da - ca) * subPosition.getX()) + ca;
			alpha = byte(hxa7241::utility::fpToInt( ((da - ba) * subPosition.getY()) + ba ));


//			dword rgba[4];
//			image.fromDword( image.associateAlphaP(*pSmaller), rgba );
//			Vector3f a( rgba[0], rgba[1], rgba[2] );

//			/// convert elements into rgb fp with associated alpha
//			float rgba[4];
//			Vector3f a( image.fromDword( *pSmaller, rgba ) );
//			a *= rgba[3];
//			Vector3f b( image.fromDword( *(pSmaller + increment.getX()), rgba ) );
//			b *= rgba[3];
//			Vector3f c( image.fromDword( *(pSmaller + increment.getY()), rgba ) );
//			c *= rgba[3];
//			Vector3f d( image.fromDword( *(pSmaller + increment.getY() + increment.getX()), rgba ) );
//			d *= rgba[3];
//			((b -= a) *= subPosition.getX()) += a;
//			((d -= c) *= subPosition.getX()) += c;
//			(((value = d) -= b) *= subPosition.getY()) += b;
		}

		/// convert rgb value to cie xyz
		converter.convertRGBtoXYZ( value, alpha, false, false, color );
	}

	return isLoaded;
}


bool Material::getTexelVector
(
	const Vector2f&                   coords,
	const string&                     imagePath,
	const string&                     imageName,
	Sheet<hxa7241::primitives::byte>& image,
	bool&                             isLoadFailed,
	UnitVector3f&                     vector
)
{
	const bool isLoaded = loadImage( imagePath, imageName, &image, 0, 0, isLoadFailed );
	if( isLoaded )
	{
		/// 3 element wide: a - b + b - c = a - c
		/// with linearly interpolated element height for fp center position and center +/- 0.5, gives smooth profile
		/// (grad is continuous but isnt smooth, virtual height curve is smooth)
		/// adjacent diff * 90 = degrees to nudge normal

		const float QUARTER_CIRCLE = 1.5707963267948966192313216916398f;
		using hxa7241::graphics::Vector3f;

		Vector3f vectX;
		{
			const float heightXLower = readElementInterpolated( Vector2f( coords.getX() - 0.5f, coords.getY()), image );
			const float heightXUpper = readElementInterpolated( Vector2f( coords.getX() + 0.5f, coords.getY()), image );
			const float gradX        = heightXLower - heightXUpper;

			const float angle = gradX * QUARTER_CIRCLE;
			float cosAngle;
			float sinAngle;
			hxa7241::graphics::sinAndcos( angle, sinAngle, cosAngle );

			vectX.setXYZ( sinAngle, cosAngle, 0.0f );
		}

		Vector3f vectZ;
		{
			const float heightZLower = readElementInterpolated( Vector2f( coords.getX(), coords.getY() - 0.5f), image );
			const float heightZUpper = readElementInterpolated( Vector2f( coords.getX(), coords.getY() + 0.5f), image );
			const float gradZ        = heightZLower - heightZUpper;

			const float angle = gradZ * QUARTER_CIRCLE;
			float cosAngle;
			float sinAngle;
			hxa7241::graphics::sinAndcos( angle, sinAngle, cosAngle );

			vectZ.setXYZ( 0.0f, cosAngle, sinAngle );
		}

		const float minY = vectX.getY() < vectZ.getY() ? vectX.getY() : vectZ.getY();
		vector.setXYZ( vectX.getX(), minY, vectZ.getZ() );
	}

	return isLoaded;
}


bool Material::getTexel
(
	const Vector2f&                   coords,
	const string&                     imagePath,
	const string&                     imageName,
	Sheet<hxa7241::primitives::byte>& image,
	bool&                             isLoadFailed,
	Float01&                          texel
)
{
	const bool isLoaded = loadImage( imagePath, imageName, &image, 0, 0, isLoadFailed );
	if( isLoaded )
	{
		texel = readElementInterpolated( coords, image );
	}

	return isLoaded;
}


bool Material::loadImage
(
	const string&                     imagePath,
	const string&                     imageName,
	Sheet<hxa7241::primitives::byte>* pImageByte,
	ImageRgba*                        pImageRgba,
	ImageConverter*                   pConverter,
	bool&                             isLoadFailed
)
{
	bool isLoaded = false;

	if( !imageName.empty() )
	{
		if( ( ((pImageByte != 0) && (pImageByte->getLength() == 0)) ||
		      ((pImageRgba != 0) && (pImageRgba->getLength() == 0)) )
		    & !isLoadFailed )
		{
			bool didRead = false;
			try
			{
				/// open file and request exceptions
				string filePathname( imagePath.size() + imageName.size(), ' ' );
				(filePathname = imagePath) += imageName;
				std::ifstream imageFile( filePathname.c_str(), std::ifstream::binary );
				imageFile.exceptions( imageFile.badbit | imageFile.failbit | imageFile.eofbit );

				/// read file
				if( (pImageRgba != 0) & (pConverter != 0) )
				{
					didRead  = perceptuum2::imaging::ImageReader::readStatic( imageFile, *pImageRgba, *pConverter );
					didRead &= (pImageRgba->getLength() > 0);
				}
				else if( pImageByte != 0 )
				{
					didRead = perceptuum2::imaging::ImageReader::readStatic( imageFile, *pImageByte );
					didRead &= (pImageByte->getLength() > 0);
				}

				if( !didRead )
				{
					hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC1 << CLASS_NAME() << " - " <<
					                              LOAD_FAIL_MESSAGE << " " << filePathname;
				}
			}
			catch( const std::exception& e )
			{
				hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC1 << CLASS_NAME() << " - " <<
				                              LOAD_FAIL_MESSAGE << " " << imagePath << imageName << ",  " << e.what();
			}
			catch( ... )
			{
				hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC1 << CLASS_NAME() << " - " << LOAD_FAIL_MESSAGE;
			}
			isLoadFailed = !didRead;
		}

		isLoaded = !isLoadFailed;
	}

	return isLoaded;
}


float Material::readElementInterpolated
(
	const Vector2f&                         coords,
	const Sheet<hxa7241::primitives::byte>& image
)
{
	/// get interpolant info
	dword    smallerOffset;
	Vector2i increment;
	Vector2f subPosition;
	makeInterpolantPositions( coords, Vector2f( image.getWidth(), image.getHeight() ), smallerOffset, increment, subPosition );

	/// interpolate surrounding four elements bilinearly
	float value;
	{
		const byte* pSmaller  = image.getMemory() + smallerOffset;

		const float a = float( udword(ubyte(*pSmaller)) );
		const float b = float( udword(ubyte(*(pSmaller + increment.getX()))) );
		const float c = float( udword(ubyte(*(pSmaller + increment.getY()))) );
		const float d = float( udword(ubyte(*(pSmaller + increment.getY() + increment.getX()))) );
		const float e = ((b - a) * subPosition.getX()) + a;
		const float f = ((d - c) * subPosition.getX()) + c;
		value         = ((f - e) * subPosition.getY()) + e;
	}

	/// convert 8bit value to float [0,1)
	return  value * hxa7241::utility::FRACTIONALIZER::_8BIT;
}


void Material::makeInterpolantPositions
(
	const Vector2f& coords,
	const Vector2f& imageSize,
	dword&          smallerOffset,
	Vector2i&       increment,
	Vector2f&       subPosition
)
{
	Vector2f clampedCoords( coords );
	clampedCoords -= Vector2f( 0.5f, 0.5f );
	clampedCoords.clampBetween( Vector2f::ZERO(), imageSize - Vector2f::EPSILON() );

	const dword smallerX = hxa7241::utility::fpToIntTrunc( clampedCoords.getX() );
	const dword smallerY = hxa7241::utility::fpToIntTrunc( clampedCoords.getY() );

	smallerOffset = smallerX + (smallerY * imageSize.getX());
	increment.setXY( (smallerX + 1) < imageSize.getX() ? 1                : 0,
	                 (smallerY + 1) < imageSize.getY() ? imageSize.getX() : 0 );
	subPosition.setXY( clampedCoords.getX() - float(smallerX), clampedCoords.getY() - float(smallerY) );
}




void Material::resetImages()
{
	color_m.set( 0, 0, ImageRgba::ARGB );
	colorSpec_m.set( 0, 0, ImageRgba::ARGB );
	specDiff_m.setSize( 0, 0 );
	roughnessX_m.setSize( 0, 0 );
	roughnessY_m.setSize( 0, 0 );
	transparency_m.setSize( 0, 0 );
	emitted_m.setSize( 0, 0 );
	tangent_m.setSize( 0, 0 );
	normal_m.setSize( 0, 0 );

	resetImageFlags();
}


void Material::resetImageFlags()
{
	isColorLoadFailed_m        = false;
	isColorSpecLoadFailed_m    = false;
	isSpecDiffLoadFailed_m     = false;
	isRoughnessXLoadFailed_m   = false;
	isRoughnessYLoadFailed_m   = false;
	isTransparencyLoadFailed_m = false;
	isEmittedLoadFailed_m      = false;
	isTangentLoadFailed_m      = false;
	isNormalLoadFailed_m       = false;
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::modelling::operator<<( SerializerOut& out, const Material& obj )
{
	out.write( "name", obj.name_m );

	out.write( "baseQualities", obj.baseQualities_m );

	out.write( "colorImageName",        obj.colorImageName_m );
	out.write( "colorSpecImageName",    obj.colorSpecImageName_m );
	out.write( "specDiffImageName",     obj.specDiffImageName_m );
	out.write( "roughnessXImageName",   obj.roughnessXImageName_m );
	out.write( "roughnessYImageName",   obj.roughnessYImageName_m );
	out.write( "transparencyImageName", obj.transparencyImageName_m );
	out.write( "emittedImageName",      obj.emittedImageName_m );
	out.write( "tangentImageName",      obj.tangentImageName_m );
	out.write( "normalImageName",       obj.normalImageName_m );

	return out;
}


SerializerIn& perceptuum2::modelling::input( SerializerIn& in, Material& obj, Scene& context )
{
	/// read name
	{
		in.read( "name", obj.name_m );
		context.addNameOfMaterial( obj.name_m );
	}


	in.read( "baseQualities", obj.baseQualities_m );

	in.read( "colorImageName",        obj.colorImageName_m );
	in.read( "colorSpecImageName",    obj.colorSpecImageName_m );
	in.read( "specDiffImageName",     obj.specDiffImageName_m );
	in.read( "roughnessXImageName",   obj.roughnessXImageName_m );
	in.read( "roughnessYImageName",   obj.roughnessYImageName_m );
	in.read( "transparencyImageName", obj.transparencyImageName_m );
	in.read( "emittedImageName",      obj.emittedImageName_m );
	in.read( "tangentImageName",      obj.tangentImageName_m );
	in.read( "normalImageName",       obj.normalImageName_m );

	obj.resetImages();


	/// enforce constraints for non textured emitter implementation
	if( !obj.baseQualities_m.emitted_m.isZero() )
	{
		/// if any image names arent empty, log warning
		/// empty all image names

		bool isNonEmpties = false;

		string*const names[] = { &(obj.colorImageName_m), &(obj.colorSpecImageName_m), &(obj.specDiffImageName_m),
		                         &(obj.roughnessXImageName_m), &(obj.roughnessYImageName_m), &(obj.transparencyImageName_m),
		                         &(obj.emittedImageName_m), &(obj.tangentImageName_m), &(obj.normalImageName_m) };
		for( int i = sizeof(names)/sizeof(names[0]);  i-- > 0;  )
		{
			if( !names[i]->empty() )
			{
				names[i]->resize( 0 );
				isNonEmpties = true;
			}
		}

		if( isNonEmpties )
		{
			hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC1 << obj.CLASS_NAME() << " " << obj.name_m <<
			                              " -  texture mapping on emitters not implemented yet";
		}
	}


	return in;
}


const char* Material::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::modelling::Material";
	return CLASS_NAME;
}


const char* Material::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a material definition of the surface of an object";
	return CLASS_DESCRIPTION;
}
