/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Scene.h"
#include "ObjectRenderableId.h"
#include "Polar.h"
#include "Serialization.h"
#include "Logger.h"

#include "ObjectInstanceMoving.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// statics ----------------------------------------------------------------------------------------
const char ObjectInstanceMoving::DESERIALIZATION_WARNING_1[] = "object nesting too deep";




/// standard object services -----------------------------------------------------------------------
ObjectInstanceMoving::ObjectInstanceMoving
(
	const Scene*  pScene,
	SerializerIn& in,
	const bool    b
)
 :	ObjectInstance( pScene )
	//motions_m()
{
	if( b )
	{
		in.read( "object", *this );
	}
	else
	{
		in >> *this;
	}
}


ObjectInstanceMoving::ObjectInstanceMoving
(
	const Scene* pScene
)
 :	ObjectInstance( pScene )
	//motions_m()
{
}


ObjectInstanceMoving::~ObjectInstanceMoving()
{
}


ObjectInstanceMoving::ObjectInstanceMoving( const ObjectInstanceMoving& other )
 :	ObjectInstance( other ),
	motions_m     ( other.motions_m )
{
}


ObjectInstanceMoving& ObjectInstanceMoving::operator=( const ObjectInstanceMoving& other )
{
	assign( other );

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
bool ObjectInstanceMoving::update
(
	const float time
)
{
	bool isChanging = false;

	const float previousUpdateTime = getUpdateTime();
	if( time != previousUpdateTime )
	{
		/// step thru motions
		{
			deque<Motion>::const_iterator itMotion( motions_m.begin() );
			deque<Motion>::const_iterator itEnd( motions_m.end() );
			deque<SubPart>::iterator      itSubpart( subParts().begin() );
			for( ;  itMotion != itEnd;  ++itMotion, ++itSubpart )
			{
				/// generate subpart transform from motion (if transform will be different for this time)
				isChanging |= itMotion->setTransform( time, previousUpdateTime, true, itSubpart->transform_m );
			}
		}

		/// update subparts
		const bool isSubpartsChanging = ObjectInstance::update( time );

		/// if something is changing, reset base object
		/// unless base already reset itself
		if( isChanging & !isSubpartsChanging )
		{
			ObjectInstance::setBase();
		}

		isChanging |= isSubpartsChanging;

		ObjectInstance::setUpdateNote( time, isChanging );
	}
	else
	{
		isChanging = getUpdateIsChanging();
	}

	return isChanging;
}




/// queries ----------------------------------------------------------------------------------------
const char* ObjectInstanceMoving::getClassName() const
{
	return ObjectInstanceMoving::CLASS_NAME();
}




/// implementation ---------------------------------------------------------------------------------
void ObjectInstanceMoving::assign
(
	const ObjectRenderable& other
)
{
	if( &other != this )
	{
		ObjectInstance::assign( other );

		motions_m.resize( 0 );
	}
}


void ObjectInstanceMoving::assign
(
	const ObjectInstance& other
)
{
	if( &other != this )
	{
		ObjectInstance::assign( other );

		motions_m.resize( 0 );
	}
}


void ObjectInstanceMoving::assign
(
	const ObjectInstanceMoving& other
)
{
	if( &other != this )
	{
		ObjectInstance::assign( other );

		motions_m = other.motions_m;
	}
}


void ObjectInstanceMoving::setBase()
{
	const float time = getScene()->getTimeBegin();

	deque<SubPart>& subparts = subParts();

	/// resize subparts collection
	subparts.resize( motions_m.size() );

	/// step thru motions
	deque<Motion>::const_iterator itMotion( motions_m.begin() );
	deque<SubPart>::iterator      itSubpart( subparts.begin() );
	for( ;  itMotion != motions_m.end();  ++itMotion, ++itSubpart )
	{
		const ObjectInstanceMoving::Motion& motion  = *itMotion;
		ObjectInstance::SubPart&            subpart = *itSubpart;

		/// copy object index to subpart
		subpart.objectIndex_m = motion.getObjectIndex();

		/// generate subpart transform from motion
		motion.setTransform( time, 0.0f, false, subpart.transform_m );
	}

	ObjectInstance::setBase();
}




/// serialization ----------------------------------------------------------------------------------
template<>
inline
const char* SerializerOut::typeName<std::deque<ObjectInstanceMoving::Motion> >()
{
	return "std::deque";
}


template<>
inline
const char* SerializerOut::typeDescription<std::deque<ObjectInstanceMoving::Motion> >()
{
	return "standard deque collection";
}


SerializerOut& perceptuum2::modelling::operator<<( SerializerOut& out, const ObjectInstanceMoving& obj )
{
	out.write( "subParts", obj.motions_m, *(obj.getScene()), 0 );

	return out;
}


SerializerIn& perceptuum2::modelling::operator>>( SerializerIn& in, ObjectInstanceMoving& obj )
{
	/// read object
	{
		/// read sub parts
		obj.motions_m.resize( 0 );
		in.read( "subParts", obj.motions_m, *(obj.getScene()) );

		/// check sub parts
		for( deque<ObjectInstanceMoving::Motion>::iterator it = obj.motions_m.begin();  it != obj.motions_m.end();  ++it )
		{
			/// check object index and object pointer validity
			const ObjectRenderable* const pObject = obj.getScene()->getObject( it->getObjectIndex() );
			if( pObject != 0 )
			{
				/// check nesting limit
				if( !(pObject->getHierarchyLevel() < ObjectRenderableId::getMaxStackSize()) )
				{
					/// remove sub part
					it = obj.motions_m.erase( it );
					hxa7241::utility::Logger() << hxa7241::utility::Logger::EXC1 << obj.DESERIALIZATION_WARNING_1 << " (" << (it - obj.motions_m.begin()) << ")";
				}
			}
			else
			{
				/// remove sub part
				it = obj.motions_m.erase( it );
			}
		}

		if( obj.motions_m.size() > udword(ObjectRenderableId::getMaxStackWidth()) )
		{
			obj.motions_m.resize( ObjectRenderableId::getMaxStackWidth() );
		}
	}

	/// set base
	obj.ObjectInstanceMoving::setBase();

	return in;
}


const char* ObjectInstanceMoving::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::modelling::ObjectInstanceMoving";
	return CLASS_NAME;
}


const char* ObjectInstanceMoving::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "moving transplaced instance(s) of renderable object(s)";
	return CLASS_DESCRIPTION;
}








/// queries ----------------------------------------------------------------------------------------
dword ObjectInstanceMoving::Motion::getObjectIndex() const
{
	return objectIndex_m;
}


bool ObjectInstanceMoving::Motion::setTransform
(
	const float time,
	const float previousTime,
	const bool  isTimeChecked,
	Transplace& transform
) const
{
	bool isChanged = false;

	const float duration = timeEnd_m - timeStart_m;

	if( !isTimeChecked | (duration > 0.0f) )
	{
		float timeClamped = time;
		bool  isPreviousTimeClampedSame = false;

		if( time < timeStart_m )
		{
			timeClamped               = timeStart_m;
			isPreviousTimeClampedSame = previousTime < timeStart_m;
		}
		else if( time > timeEnd_m )
		{
			timeClamped               = timeEnd_m;
			isPreviousTimeClampedSame = previousTime > timeEnd_m;
		}

		if( !isTimeChecked | !isPreviousTimeClampedSame )
		{
			const float timeProportion = (duration > 0.0f) ? (timeClamped - timeStart_m) / duration : 0.0f;

			/// calculate rotation
			const float angle = motionAngle_m * timeProportion;
			Quaternion rotation( angle, motionAxis_m );
			rotation *= rotationStart_m;

			/// calculate position
			Vector3f position;
			(((position = positionEnd_m) -= positionStart_m) *= timeProportion) += positionStart_m;

			/// set transform
			transform.setTransplace( rotation, position );

			isChanged = true;
		}
	}

	return isChanged;
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::modelling::output( SerializerOut& out, const ObjectInstanceMoving::Motion& obj, const Scene& context )
{
	/// write object ref
	{
		const string objectRef( context.getObjectName( obj.objectIndex_m ) );
		out.write( "objectRef", objectRef );
	}

	/// write times
	out.write( "startTime", obj.timeStart_m );
	out.write( "endTime",   obj.timeEnd_m   );

	/// write movements (rotations not done...)
	out.write( "startRotation", Vector3f::ZERO()    );
	out.write( "startPosition", obj.positionStart_m );
	out.write( "endRotation",   Vector3f::ZERO()    );
	out.write( "endPosition",   obj.positionEnd_m   );

	return out;
}

SerializerIn& perceptuum2::modelling::input( SerializerIn& in, ObjectInstanceMoving::Motion& obj, const Scene& context )
{
	/// read object ref
	{
		string objectRef;
		in.read( "objectRef", objectRef );
		obj.objectIndex_m = context.getObjectIndex( objectRef );
	}

	/// read times
	in.read( "startTime", obj.timeStart_m );
	in.read( "endTime",   obj.timeEnd_m   );

	/// read movements
	{
		Vector3f rotationStart;
		Vector3f rotationEnd;
		in.read( "startRotation", rotationStart       );
		in.read( "startPosition", obj.positionStart_m );
		in.read( "endRotation",   rotationEnd         );
		in.read( "endPosition",   obj.positionEnd_m   );

		/// convert degrees to radians
		{
			const float dToR = hxa7241::graphics::degreesToRadians( 1.0f );
			rotationStart *= dToR;
			rotationEnd   *= dToR;
		}

		obj.rotationStart_m.setRotations( rotationStart.getX(), rotationStart.getY(), rotationStart.getZ() );

		/// calculate intermediate rotation
		Quaternion intermediateRotation( obj.rotationStart_m );
		intermediateRotation.invert();
		intermediateRotation *= Quaternion( rotationEnd.getX(), rotationEnd.getY(), rotationEnd.getZ() );

		intermediateRotation.getAngleAndAxis( obj.motionAngle_m, obj.motionAxis_m );
	}

	/// enforce constraints
	{
		if( obj.timeEnd_m < obj.timeStart_m )
		{
			obj.timeEnd_m = obj.timeStart_m;
		}
	}

	return in;
}


const char* ObjectInstanceMoving::Motion::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::modelling::ObjectInstanceMoving::Motion";
	return CLASS_NAME;
}


const char* ObjectInstanceMoving::Motion::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "internal element from an ObjectInstanceMoving";
	return CLASS_DESCRIPTION;
}
