/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Serialization.h"
#include "FpToInt.h"

#include "ObjectMeshBlock.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// standard object services -----------------------------------------------------------------------
ObjectMeshBlock::ObjectMeshBlock
(
	const Scene*  pScene,
	SerializerIn& in,
	const bool    b
)
 :	ObjectMesh( pScene ),
	dimensions_m(),
	isNet_m(),
	textureCoords_m(),
	textureScale_m()
{
	if( b )
	{
		in.read( "object", *this );
	}
	else
	{
		in >> *this;
	}
}


ObjectMeshBlock::ObjectMeshBlock
(
	const Scene* pScene
)
 :	ObjectMesh( pScene ),
	dimensions_m(),
	isNet_m(),
	textureCoords_m(),
	textureScale_m()
{
}


ObjectMeshBlock::~ObjectMeshBlock()
{
}


ObjectMeshBlock::ObjectMeshBlock( const ObjectMeshBlock& other )
 :	ObjectMesh     ( other ),
	dimensions_m   ( other.dimensions_m ),
	isNet_m        ( other.isNet_m ),
	textureCoords_m( other.textureCoords_m ),
	textureScale_m ( other.textureScale_m )
{
}


ObjectMeshBlock& ObjectMeshBlock::operator=( const ObjectMeshBlock& other )
{
	assign( other );

	return *this;
}




/// implementation ---------------------------------------------------------------------------------
void ObjectMeshBlock::assign
(
	const ObjectRenderable& other
)
{
	if( &other != this )
	{
		ObjectMesh::assign( other );

		dimensions_m    = Vector3f();
		isNet_m         = false;
		textureCoords_m = Vector2<word>();
		textureScale_m  = 0.0f;
	}
}


void ObjectMeshBlock::assign
(
	const ObjectMesh& other
)
{
	if( &other != this )
	{
		ObjectMesh::assign( other );

		dimensions_m    = Vector3f();
		isNet_m         = false;
		textureCoords_m = Vector2<word>();
		textureScale_m  = 0.0f;
	}
}


void ObjectMeshBlock::assign
(
	const ObjectMeshBlock& other
)
{
	if( &other != this )
	{
		ObjectMesh::assign( other );

		dimensions_m    = other.dimensions_m;
		isNet_m         = other.isNet_m;
		textureCoords_m = other.textureCoords_m;
		textureScale_m  = other.textureScale_m;
	}
}


void ObjectMeshBlock::makeBlock
(
	const Vector3f&               dimensions,
	const bool                    isNet,
	const Vector2<word>&          textureCoord,
	const vector<Vector2<word> >& textureCoords,
	const float                   textureScale,
	vector<Vector3f>&             vertexPositions,
	vector<TriangleImplicit>&     triangles
)
{
	/// make vertexs
	/// 000 001 010 011
	/// 100 101 110 111
	vertexPositions.resize( 8 );
	for( dword v = 8;  v-- > 0; )
	{
		Vector3f& vertPos = vertexPositions[v];

		vertPos.setXYZ( float(v & 0x01), float((v & 0x02) >> 1), float((v & 0x04) >> 2) );
		vertPos *= dimensions;
	}


	/// make triangles
	/// 012 123 456 567  ()
	/// 024 246 135 357  (first row digits 3 bits rotated left 1)
	/// 041 415 263 637  (first row digits 3 bits rotated left 2)
	triangles.resize( 12 );
	for( dword i = 3;  i-- > 0; )
	{
		for( dword j = 4;  j-- > 0; )
		{
			const dword a = ((j & 0x02) << 1) | (j & 0x01);
			dword t[3];
			for( dword k = 3;  k-- > 0; )
			{
				dword v = k + a;
				for( dword r = i;  r-- > 0; )
				{
					v = ((v << 1) & 0x07) | ((v >> 2) & 0x01);
				}

				t[k] = v;
			}
			TriangleImplicit& triangle = triangles[(i * 4) + j];
			triangle.set( word(t[0]), word(t[1]), word(t[2]) );
			triangle.set( false );
		}
	}


	/// set triangle texture coords
	if( isNet )
	{
		static const byte NETVERTS[24] = { 0x10, 0x50, 0x11, 0x51,   //00010000 01010000 00010001 01010001,    /// y,0     x+y,0       y,y       x+y,y
										   0x11, 0x51, 0x15, 0x55,   //00010001 01010001 00010101 01010101,    /// y,y     x+y,y       y,y+z     x+y,y+z
										   0x15, 0x55, 0x16, 0x56,   //00010101 01010101 00010110 01010110,    /// y,y+z   x+y,y+z     y,2y+z    x+y,2y+z
										   0x16, 0x56, 0x1A, 0x5A,   //00010110 01010110 00011010 01011010,    /// y,2y+z  x+y,2y+z    y,2(y+z)  x+y,2(y+z)
										   0x01, 0x11, 0x05, 0x15,   //00000001 00010001 00000101 00010101,    /// 0,y     y,y         0,y+z     y,y+z
										   0x51, 0x61, 0x55, 0x65 }; //01010001 01100001 01010101 01100101     /// x+y,y   x+2y,y      x+y,y+z   x+2y,y+z
		static const byte TRIANGLES_TO_NETVERTS[12][3] = { { 10, 11,  8 }, { 11,  8,  9 }, {  0,  1,  2 }, {  1,  2,  3 },
														   { 18, 19, 16 }, { 19, 16, 17 }, { 23, 22, 21 }, { 22, 21, 20 },
														   { 12, 14, 13 }, { 14, 13, 15 }, {  6,  4,  7 }, {  4,  7,  5 } };

		if( textureScale != 0.0f )
		{
			const Vector3f dimsAbs( dimensions.abs() );
			const Vector3f dimsScaled( dimsAbs * (textureScale / dimsAbs.getX()) );

			for( dword i = triangles.size();  i-- > 0; )
			{
				Vector2<word> txy[3];
				for( dword j = 3;  j-- > 0; )
				{
					using hxa7241::graphics::Vector2f;
					using hxa7241::utility::fpToInt;

					const dword bits  = dword(NETVERTS[ TRIANGLES_TO_NETVERTS[i][j] ]) & 0xFF;
					const dword xBits = bits >> 4;
					const dword yBits = bits &  0x0F;
					Vector2f xy( (float(xBits >> 2) * dimsScaled.getX()) + (float(xBits & 0x03) * dimsScaled.getY()),
								 (float(yBits >> 2) * dimsScaled.getZ()) + (float(yBits & 0x03) * dimsScaled.getY()) );
					xy += Vector2f( textureCoord.getX(), textureCoord.getY() );
					xy.clampBetween( Vector2f::ZERO(), Vector2f( WORD_MAX, WORD_MAX ) );

					txy[j].setXY( word(fpToInt( xy.getX() )), word(fpToInt( xy.getY() )) );
				}

				triangles[i].set( txy );
			}
		}
	}
	else
	{
		const dword end = triangles.size();
		for( dword i = 0, j = 0;  i < end;  ++i, j += (i & 1) ? 1 : 3 )
		{
			const Vector2<word> txy[3] = { textureCoords[j], textureCoords[j+1], textureCoords[j+2] };
			triangles[i].set( txy );
		}
	}
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::modelling::operator<<( SerializerOut& out, const ObjectMeshBlock& obj )
{
	/// write dimensions
	out.write( "dimensions", obj.dimensions_m );

	/// write texture coords
	out.write( "netOrFaces", obj.isNet_m );
	if( obj.isNet_m )
	{
		out.write( "textureCoords", obj.textureCoords_m );
		out.write( "textureScale",  obj.textureScale_m  );
	}
	else
	{
		const vector<TriangleImplicit>& triangles = obj.triangles();
		const dword end = triangles.size() & 0xFFFFFFFE;
		for( dword i = 0;  i < end;  ++i )
		{
			Vector2<word> txy[3];

			triangles[i].getTextureCoords( txy );
			out.write( "textureCoords", txy[0] );
			out.write( "textureCoords", txy[1] );

			++i;
			triangles[i].getTextureCoords( txy );
			out.write( "textureCoords", txy[1] );
			out.write( "textureCoords", txy[2] );
		}
	}

	/// write material id
	out.write( "materialRef", obj.getMaterialName() );

	return out;
}


SerializerIn& perceptuum2::modelling::operator>>( SerializerIn& in, ObjectMeshBlock& obj )
{
	try
	{
		/// read
		vector<Vector2<word> > textureCoords;
		dword materialId;// = 0;
		{
			/// read dimensions
			in.read( "dimensions", obj.dimensions_m );

			/// read texture coords
			{
				in.read( "netOrFaces", obj.isNet_m );
				if( obj.isNet_m )
				{
					in.read( "textureCoords", obj.textureCoords_m );
					in.read( "textureScale",  obj.textureScale_m  );

					if( obj.textureScale_m != 0.0f )
					{
						obj.textureCoords_m.clampBetween( Vector2<word>::ZERO(), Vector2<word>::MAX() );

						if( obj.textureScale_m < 0.0f )
						{
							obj.textureScale_m = -obj.textureScale_m;
						}
					}
				}
				else
				{
					textureCoords.resize( 24 );
					for( dword i = 0;  i < 24;  ++i )
					{
						in.read( "textureCoords", textureCoords[i] );
					}
				}
			}

			/// read material id
			{
				string materialName;
				in.read( "materialRef", materialName );
				materialId = obj.getMaterialIndex( materialName );
			}
		}


		/// make
		vector<Vector3f>         vertexs;
		vector<TriangleImplicit> faces;
		obj.makeBlock( obj.dimensions_m, obj.isNet_m, obj.textureCoords_m, textureCoords, obj.textureScale_m,
					   vertexs, faces );

		/// set base
		obj.setMesh( vertexs, faces, materialId );


		return in;
	}
	catch( ... )
	{
		/// zero object
		obj.dimensions_m    = Vector3f::ZERO();
		obj.isNet_m         = false;
		obj.textureCoords_m = Vector2<word>::ZERO();
		obj.textureScale_m  = 0.0f;

		obj.zeroObject();

		throw;
	}
}


const char* ObjectMeshBlock::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::modelling::ObjectMeshBlock";
	return CLASS_NAME;
}


const char* ObjectMeshBlock::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a rectilinear block, made of a ObjectMesh";
	return CLASS_DESCRIPTION;
}
