/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "UnitVector3f.h"
#include "Serialization.h"
#include <map>

#include "ObjectMeshSphere.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// standard object services -----------------------------------------------------------------------
ObjectMeshSphere::ObjectMeshSphere
(
	const Scene*  pScene,
	SerializerIn& in,
	const bool    b
)
 :	ObjectMesh( pScene ),
	radius_m(),
	resolution_m()
{
	if( b )
	{
		in.read( "object", *this );
	}
	else
	{
		in >> *this;
	}
}


ObjectMeshSphere::ObjectMeshSphere
(
	const Scene* pScene
)
 :	ObjectMesh( pScene ),
	radius_m(),
	resolution_m()
{
}


ObjectMeshSphere::~ObjectMeshSphere()
{
}


ObjectMeshSphere::ObjectMeshSphere( const ObjectMeshSphere& other )
 :	ObjectMesh  ( other ),
	radius_m    ( other.radius_m ),
	resolution_m( other.resolution_m )
{
}


ObjectMeshSphere& ObjectMeshSphere::operator=( const ObjectMeshSphere& other )
{
	assign( other );

	return *this;
}




/// implementation ---------------------------------------------------------------------------------
void ObjectMeshSphere::assign
(
	const ObjectRenderable& other
)
{
	if( &other != this )
	{
		ObjectMesh::assign( other );

		radius_m     = 0.0f;
		resolution_m = 0.0f;
	}
}


void ObjectMeshSphere::assign
(
	const ObjectMesh& other
)
{
	if( &other != this )
	{
		ObjectMesh::assign( other );

		radius_m     = 0.0f;
		resolution_m = 0.0f;
	}
}


void ObjectMeshSphere::assign
(
	const ObjectMeshSphere& other
)
{
	if( &other != this )
	{
		ObjectMesh::assign( other );

		radius_m     = other.radius_m;
		resolution_m = other.resolution_m;
	}
}


void ObjectMeshSphere::makeSphere
(
	const float               radius,
	const dword               resolution,
	vector<Vector3f>&         vertexs,
	vector<TriangleImplicit>& faces
)
{
	/// reserve storage
	/// choose polyhedron and subdivision
	/// subdiv  0  1   2    3    4      res  0  1   2   3   4   5   6   7   8   9
	/// tetra   4 16  64  256 1024
	/// oct     8 32 128  512 2048      oct  8     32     128     512    2048
	/// icos   20 80 320 1280 5120      icos   20      80     320    1280    5120
	const bool  octaOrIcosa = bool(resolution & 1);
	const dword subdivision = resolution >> 1;
	{
		/// f = initfaces * (4 ^ subdivision)
		const dword initialFaceCount = octaOrIcosa ? 20 : 8;
		const dword faceCount        = initialFaceCount << (subdivision * 2);
		faces.reserve( faceCount );

		/// v = (0.5 * f) + 2
		const dword vertexCount = (faceCount >> 1) + 2;
		vertexs.reserve( vertexCount );
	}


	/// make and subdivide
	makeTrianglePolyhedron( octaOrIcosa, radius, vertexs, faces );
	subdivideTrianglePolyhedron( subdivision, radius, vertexs, faces );
}


void ObjectMeshSphere::makeTrianglePolyhedron
(
	const bool                octaOrIcosa,
	const float               radius,
	vector<Vector3f>&         vertexs,
	vector<TriangleImplicit>& faces
)
{
	/// constant polyhedra definitions
	//static const float TETRAHEDRON_VERTEXS[4][3]  = { {+1, +1, +1}, {-1, -1, +1}, {-1, +1, -1}, {+1, -1, -1} };
	//static const word  TETRAHEDRON_FACES[4][3]    = { { 0,  1,  2}, { 0,  1,  3}, { 0,  2,  3}, { 1,  2,  3} };
	//static const dword TETRAHEDRON_VERTEX_COUNT   = sizeof(TETRAHEDRON_VERTEXS) / sizeof(float[3]);
	//static const dword TETRAHEDRON_FACE_COUNT     = sizeof(TETRAHEDRON_FACES)   / sizeof(word[3]);
	static const float OCTAHEDRON_VERTEXS[6][3]   = { {+1, 0, 0}, {-1, 0, 0}, {0, +1, 0}, {0, -1, 0}, {0, 0, +1}, {0, 0, -1} };
	static const word  OCTAHEDRON_FACES[8][3]     = { { 0, 2, 4 }, { 0, 2, 5 }, { 0, 3, 4 }, { 0, 3, 5 }, { 1, 2, 4 }, { 1, 2, 5 }, { 1, 3, 4 }, { 1, 3, 5 } };
	static const dword OCTAHEDRON_VERTEX_COUNT    = sizeof(OCTAHEDRON_VERTEXS) / sizeof(float[3]);
	static const dword OCTAHEDRON_FACE_COUNT      = sizeof(OCTAHEDRON_FACES)   / sizeof(word[3]);
	static const float ICOSAHEDRON_VERTEXS[12][3] = { {1.61803398875f, 1, 0}, {-1.61803398875f, 1, 0}, {1.61803398875f, -1, 0}, {-1.61803398875f, -1, 0},
													  {1, 0, 1.61803398875f}, {1, 0, -1.61803398875f}, {-1, 0, 1.61803398875f}, {-1, 0, -1.61803398875f},
													  {0, 1.61803398875f, 1}, {0, -1.61803398875f, 1}, {0, 1.61803398875f, -1}, {0, -1.61803398875f, -1} };
	static const word  ICOSAHEDRON_FACES[20][3]   = { { 0, 2,  4 }, { 0, 2,  5 }, { 0, 4,  8 }, { 0, 5, 10 }, { 0,  8, 10 },
													  { 1, 3,  6 }, { 1, 3,  7 }, { 1, 6,  8 }, { 1, 7, 10 }, { 1,  8, 10 },
													  { 2, 4,  9 }, { 2, 5, 11 }, { 2, 9, 11 }, { 3, 6,  9 }, { 3,  7, 11 },
													  { 3, 9, 11 }, { 4, 6,  8 }, { 4, 6,  9 }, { 5, 7, 10 }, { 5, 11,  7 } };
	static const dword ICOSAHEDRON_VERTEX_COUNT   = sizeof(ICOSAHEDRON_VERTEXS) / sizeof(float[3]);
	static const dword ICOSAHEDRON_FACE_COUNT     = sizeof(ICOSAHEDRON_FACES)   / sizeof(word[3]);

	typedef const float VertexData[3];
	typedef const word  FaceData[3];
	static const VertexData* VERTEXS[] = { OCTAHEDRON_VERTEXS, ICOSAHEDRON_VERTEXS };
	static const FaceData*   FACES[]   = { OCTAHEDRON_FACES,   ICOSAHEDRON_FACES };
	static const dword VERTEX_COUNTS[] = { OCTAHEDRON_VERTEX_COUNT, ICOSAHEDRON_VERTEX_COUNT };
	static const dword FACE_COUNTS[]   = { OCTAHEDRON_FACE_COUNT,   ICOSAHEDRON_FACE_COUNT };


	/// copy to vectors
	{
		const dword poly       = dword(octaOrIcosa);
		const VertexData*const pInitialVertexData = VERTEXS[poly];
		const FaceData*const   pInitialFaceData   = FACES[poly];
		const dword            initialVertexCount = VERTEX_COUNTS[poly];
		const dword            initialFaceCount   = FACE_COUNTS[poly];

		/// init vertexs and faces
		{
			vertexs.resize( 0 );
			faces.resize( 0 );

			const float scale = radius / Vector3f( pInitialVertexData[0][0], pInitialVertexData[0][1], pInitialVertexData[0][2] ).length();
			for( int i = 0;  i < initialVertexCount;  ++i )
			{
				const float* v = pInitialVertexData[i];
				vertexs.push_back( Vector3f( v[0], v[1], v[2] ) *= scale );
			}

			for( int i = 0;  i < initialFaceCount;  ++i )
			{
				const word* f = pInitialFaceData[i];
				faces.push_back( TriangleImplicit( f[0], f[1], f[2], true ) );
			}
		}
	}
}


void ObjectMeshSphere::subdivideTrianglePolyhedron
(
	const dword               subdivision,
	const float               radius,
	vector<Vector3f>&         vertexs,
	vector<TriangleImplicit>& faces
)
{
	/// subdivide polyhedron: each triangle face into four, each edge into two, putting new vertexs on circumsphere
	/// (doesnt produce equal size faces)

	std::map<dword, word> edges;
	for( int i = subdivision;  i-- > 0; )
	{
		/// make subdivided edges and new vertexs
		edges.clear();
		for( int f = faces.size();  f-- > 0; )
		{
			const dword vi[] = { faces[f].getVertexIndex0(), faces[f].getVertexIndex1(), faces[f].getVertexIndex2() };
			for( int v = 3;  v-- > 0; )
			{
				/// make edge key
				const dword vOne = vi[v];
				const dword vTwo = vi[(v + 1) %3];
				const dword edgeKey = makeEdgeKey( vOne, vTwo );

				/// check if edge already there
				if( edges.find( edgeKey ) == edges.end() )
				{
					/// make and append middle vertex
					using hxa7241::graphics::UnitVector3f;
					const Vector3f middleVertex( UnitVector3f( vertexs[vOne] + vertexs[vTwo] ).getVector() * radius );
					vertexs.push_back( middleVertex );

					/// make and append edge
					const dword vMid = vertexs.size() - 1;
					edges[ edgeKey ] = word(vMid);
				}
			}
		}

		/// subdivide faces
		for( int f = 0, end = faces.size();  f < end;  ++f )
		{
			const dword vi[] = { faces[f].getVertexIndex0(), faces[f].getVertexIndex1(), faces[f].getVertexIndex2() };

			word vi2[3];
			for( int v = 3;  v-- > 0; )
			{
				const word midOne = edges[ makeEdgeKey( vi[(v + 2) %3], vi[v] ) ];
				const word midTwo = edges[ makeEdgeKey( vi[v], vi[(v + 1) %3] ) ];
				vi2[v] = midOne;

				/// append new face, made with corner and two subdivided edge vertexs
				faces.push_back( TriangleImplicit( midOne, word(vi[v]), midTwo, true ) );
			}

			/// modify old face to be center face, made with subdivided edge vertexs
			faces[f].set( vi2[0], vi2[1], vi2[2] );
		}
	}
}


dword ObjectMeshSphere::makeEdgeKey
(
	const dword one,
	const dword two
)
{
	/// sort to consistent order
	if( two < one )
	{
		const temp = one;
		one = two;
		two = temp;
	}

	return (one << 16) | (two & 0xFFFF);
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::modelling::operator<<( SerializerOut& out, const ObjectMeshSphere& obj )
{
	/// write radius
	out.write( "diameter", obj.radius_m * 2.0f );

	/// write resolution
	out.write( "resolution", obj.resolution_m );

	/// write material id
	out.write( "materialRef", obj.getMaterialName() );

	return out;
}


SerializerIn& perceptuum2::modelling::operator>>( SerializerIn& in, ObjectMeshSphere& obj )
{
	try
	{
		/// read
		dword materialId;// = 0;
		{
			/// read radius
			{
				float diameter = 0.0f;
				in.read( "diameter", diameter );
				if( diameter <= 0.0f )
				{
					diameter = 1.0f;
				}
				obj.radius_m = diameter * 0.5f;
			}

			/// read resolution
			in.read( "resolution", obj.resolution_m );
			//static const dword RESOLUTION_MAX = 13;   /// 65536 faces (must have < 16bit vertex count)
			static const dword RESOLUTION_MAX = 7;   /// 65536 faces (must have < 16bit vertex count)
			if( obj.resolution_m < 0 )
			{
				obj.resolution_m = 0;
			}
			else if( obj.resolution_m > RESOLUTION_MAX )
			{
				obj.resolution_m = RESOLUTION_MAX;
			}

			/// read material id
			{
				string materialName;
				in.read( "materialRef", materialName );
				materialId = obj.getMaterialIndex( materialName );
			}
		}


		/// make
		vector<Vector3f>         vertexs;
		vector<TriangleImplicit> faces;
		obj.makeSphere( obj.radius_m, obj.resolution_m, vertexs, faces );


		/// set base
		obj.setMesh( vertexs, faces, materialId );


		return in;
	}
	catch( ... )
	{
		/// zero object
		obj.radius_m     = 0.0f;
		obj.resolution_m = 0;

		obj.zeroObject();

		throw;
	}
}


const char* ObjectMeshSphere::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::modelling::ObjectMeshSphere";
	return CLASS_NAME;
}


const char* ObjectMeshSphere::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a sphere, made of a ObjectMesh";
	return CLASS_DESCRIPTION;
}
