/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Serialization.h"
#include "Transform.h"
#include <string>

#include "Qualities.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// standard object services -----------------------------------------------------------------------
Qualities::Qualities()
 :	normal_m              (),
	tangent_m             ( UnitVector3f::X() ),
	reflectivityDiffuse_m ( 0.5f, 0.5f, 0.5f ),
	reflectivitySpecular_m( 0.5f, 0.5f, 0.5f ),
	roughnessX_m          ( 1.0f ),
	roughnessY_m          ( 1.0f ),
	transparency_m        ( 0.0f ),
	refractiveIndex_m     ( 1.0f ),
	emitted_m             ( CieXyz::ZERO() ),
	emittedRoughness_m    ( 1.0f )
{
	brdfClassName_m[0] = '\0';
}


Qualities::~Qualities()
{
}


Qualities::Qualities( const Qualities& other )
 :	normal_m              ( other.normal_m ),
	tangent_m             ( other.tangent_m ),
	reflectivityDiffuse_m ( other.reflectivityDiffuse_m ),
	reflectivitySpecular_m( other.reflectivitySpecular_m ),
	roughnessX_m          ( other.roughnessX_m ),
	roughnessY_m          ( other.roughnessY_m ),
	transparency_m        ( other.transparency_m ),
	refractiveIndex_m     ( other.refractiveIndex_m ),
	emitted_m             ( other.emitted_m ),
	emittedRoughness_m    ( other.emittedRoughness_m )
{
	copyString( other.brdfClassName_m, sizeof(brdfClassName_m)/sizeof(*brdfClassName_m), brdfClassName_m );
}


Qualities& Qualities::operator=( const Qualities& other )
{
	if( &other != this )
	{
		normal_m               = other.normal_m;
		tangent_m              = other.tangent_m;
		reflectivityDiffuse_m  = other.reflectivityDiffuse_m;
		reflectivitySpecular_m = other.reflectivitySpecular_m;
		roughnessX_m           = other.roughnessX_m;
		roughnessY_m           = other.roughnessY_m;
		transparency_m         = other.transparency_m;
		refractiveIndex_m      = other.refractiveIndex_m;
		emitted_m              = other.emitted_m;
		emittedRoughness_m     = other.emittedRoughness_m;

		copyString( other.brdfClassName_m, sizeof(brdfClassName_m)/sizeof(*brdfClassName_m), brdfClassName_m );
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void Qualities::setToDefault()
{
	normal_m           = UnitVector3f::ZERO();
	tangent_m          = UnitVector3f::X();
	reflectivityDiffuse_m.setXYZ( 0.5f, 0.5f, 0.5f );
	reflectivitySpecular_m.setXYZ( 0.5f, 0.5f, 0.5f );
	roughnessX_m       = 1.0f;
	roughnessY_m       = 1.0f;
	transparency_m     = 0.0f;
	refractiveIndex_m  = 1.0f;
	emitted_m          = CieXyz::ZERO();
	emittedRoughness_m = 1.0f;

	brdfClassName_m[0] = '\0';
}


Qualities& Qualities::transformEq
(
	const Transform& trans
)
{
	trans.transform( normal_m );
	trans.transform( tangent_m );

	return *this;
}




/// queries ----------------------------------------------------------------------------------------
Float01 Qualities::getSpecularness() const
{
	return getSpecularness( roughnessX_m, roughnessY_m, reflectivitySpecular_m );

//	const float smoothness = 1.0f - ((roughnessX_m + roughnessY_m) * 0.5f);
//
//	return Float01( reflectivitySpecular_m.average() * smoothness );
}


Float01 Qualities::getSpecularness
(
	const Float01 roughnessX,
	const Float01 roughnessY,
	const CieXyz& reflectivitySpecular
)
{
	const float smoothness = 1.0f - ((roughnessX + roughnessY) * 0.5f);

	return Float01( reflectivitySpecular.average() * smoothness );


//	/// mean
//	return Float01( FLOAT_ALMOST_ONE - ((roughnessX + roughnessY) * 0.5f) );
//
//
//	/// min
//	float roughnessMin = roughnessX;
//	if( roughnessMin > roughnessY )
//	{
//		roughnessMin = roughnessY;
//	}
//
//	return Float01( FLOAT_ALMOST_ONE - roughnessMin );
}




/// implementation ---------------------------------------------------------------------------------
void Qualities::copyString( const char* pSource, const dword length, char* pTarget )
{
	for( dword i = length;  i-- > 0; )
	{
		pTarget[i] = pSource[i];
	}
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::modelling::operator<<( SerializerOut& out, const Qualities& obj )
{
//	out.write( "normal"              , obj.normal_m  );
	out.write( "tangent"             , obj.tangent_m );

	out.write( "reflectivityDiffuse" , obj.reflectivityDiffuse_m     );
	out.write( "reflectivitySpecular", obj.reflectivitySpecular_m    );
	out.write( "roughnessX"          , float(obj.roughnessX_m)       );
	out.write( "roughnessY"          , float(obj.roughnessY_m)       );
	out.write( "transparency"        , float(obj.transparency_m)     );
	out.write( "refractiveIndex"     , float(obj.refractiveIndex_m)  );
	out.write( "emitted"             , obj.emitted_m                 );
	out.write( "emittedRoughness"    , float(obj.emittedRoughness_m) );
	out.write( "brdfClassName"       , obj.brdfClassName_m           );

	return out;
}


SerializerIn& perceptuum2::modelling::operator>>( SerializerIn& in, Qualities& obj )
{
	/// read
	{
		obj.normal_m  = UnitVector3f::ZERO();
		//in.read( "normal"              , obj.normal_m  );
		in.read( "tangent"             , obj.tangent_m );

		in.read( "reflectivityDiffuse" , obj.reflectivityDiffuse_m  );
		in.read( "reflectivitySpecular", obj.reflectivitySpecular_m );

		float f;
		in.read( "roughnessX", f );
		obj.roughnessX_m = f;
		in.read( "roughnessY", f );
		obj.roughnessY_m = f;
		in.read( "transparency", f );
		obj.transparency_m = f;

		in.read( "refractiveIndex", obj.refractiveIndex_m );
		in.read( "emitted"        , obj.emitted_m         );

		in.read( "emittedRoughness", f );
		obj.emittedRoughness_m = f;

		{
			std::string brdfClassName( sizeof(obj.brdfClassName_m)/sizeof(*obj.brdfClassName_m), '\0' );
			in.read( "brdfClassName", brdfClassName );

			brdfClassName.resize( sizeof(obj.brdfClassName_m)/sizeof(*obj.brdfClassName_m) - 1 );
			obj.copyString( brdfClassName.c_str(), sizeof(obj.brdfClassName_m)/sizeof(*obj.brdfClassName_m), obj.brdfClassName_m );
		}
	}

	/// enforce constraints
	{
		/// make sure reflectivitys sum to less than one
		{
			CieXyz reflectivity;
			(reflectivity = obj.reflectivityDiffuse_m) += obj.reflectivitySpecular_m;
			const float largest = reflectivity.largest();

			if( largest >= 1.0f )
			{
				const float rescaling = FLOAT_ALMOST_ONE / largest;
				obj.reflectivityDiffuse_m  *= rescaling;
				obj.reflectivitySpecular_m *= rescaling;
			}
		}

		/// disclude superluminal phase velocity
		if( obj.refractiveIndex_m < 1.0f )
		{
			obj.refractiveIndex_m = 1.0f;
		}

		/// disallow negative emission
		obj.emitted_m.clampToMinOf( hxa7241::graphics::Vector3f::ZERO() );

		/// isotropicalise roughness if no tangent supplied
		if( obj.tangent_m.isZero() )
		{
			const float isoRoughness = (obj.roughnessX_m + obj.roughnessY_m) * 0.5f;
			obj.roughnessX_m = isoRoughness;
			obj.roughnessY_m = isoRoughness;
		}
	}

	return in;
}


const char* Qualities::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::modelling::Qualities";
	return CLASS_NAME;
}


const char* Qualities::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "a material definition of a point on a surface";
	return CLASS_DESCRIPTION;
}
