/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Serialization.h"
#include "ObjectRenderableId.h"
#include "UriFile.h"
#include "Logger.h"
#include <sstream>

#include "Scene.h"


using namespace perceptuum2::modelling;




/// statics
const char Scene::START_TIME_LABEL[]     = "start time";
const char Scene::END_TIME_LABEL[]       = "end time";
const char Scene::OBJECT_COUNT_LABEL[]   = "number of objects";
const char Scene::MATERIAL_COUNT_LABEL[] = "number of materials";




/// standard object services -----------------------------------------------------------------------
Scene::Scene()
 :	lightsourceThreshold_m(),
	materialNames_m(),
	objectNames_m(),
	objects_m(),
	defaultObject_m( this ),
	materials_m(),
	defaultMaterial_m(),
	timeBegin_m(),
	timeEnd_m()
{
}


Scene::~Scene()
{
	const deque<ObjectInstantiator>::iterator end( objects_m.end() );
	for( deque<ObjectInstantiator>::iterator it( objects_m.begin() );  it != end;  ++it )
	{
		delete it->getObject();
	}

//	for( int i = objects_m.size();  i-- > 0; )
//	{
//		delete objects_m[i].getObject();
//	}
}




/// commands ---------------------------------------------------------------------------------------
void Scene::setLightsourceEmittanceThreshold
(
	const float lightsourceThreshold
)
{
	lightsourceThreshold_m = lightsourceThreshold;
}


void Scene::addNameOfMaterial
(
	const string& name
)
{
	addName( name, "material", materialNames_m );
}


void Scene::addNameOfObject
(
	const string& name
)
{
	addName( name, "object", objectNames_m );
}




/// queries ----------------------------------------------------------------------------------------
float Scene::getLightsourceThreshold() const
{
	return lightsourceThreshold_m;
}


void Scene::getSummary( string& s ) const
{
//	s = "\nscene summary not implemented yet...";

	std::ostringstream oss;

	oss << '\n';
	oss << START_TIME_LABEL << ": " << timeBegin_m << '\n';
	oss << END_TIME_LABEL   << ": " << timeEnd_m   << '\n';

	oss << '\n';
	oss << OBJECT_COUNT_LABEL << ": " << objects_m.size() << '\n';

	oss << '\n';
	oss << MATERIAL_COUNT_LABEL << ": " << materials_m.size() << '\n';

	s = oss.str();
}


float Scene::getTimeBegin() const
{
	return timeBegin_m;
}


float Scene::getTimeEnd() const
{
	return timeEnd_m;
}


bool Scene::update
(
	const float time
) const
{
	bool isChanged = false;

	ObjectRenderable*const pRootObject = objects_m.back().getObject();
	if( pRootObject != 0 )
	{
		isChanged = pRootObject->update( time );
	}

	return isChanged;
}


ObjectRenderableId Scene::getRootObject() const
{
	return ObjectRenderableId( objects_m.back().getObject() );
}


ObjectRenderableId Scene::getDefaultObject() const
{
	return ObjectRenderableId( &defaultObject_m );
}


//const vector<const ObjectRenderable*>& Scene::getObjects() const
//{
//	return *( reinterpret_cast< const vector<const ObjectRenderable*>* >( &objects_m ) );
//}


const ObjectRenderable* Scene::getObject
(
	const dword index
) const
{
	const ObjectRenderable* pObject = 0;

	/// check for valid index
	if( (index >= 0) & (udword(index) < objects_m.size()) )
	{
		pObject = objects_m[ index ].getObject();
	}

	return pObject;
}


ObjectRenderable* Scene::getObjectNonConst
(
	const dword index
) const
{
	ObjectRenderable* pObject = 0;

	/// check for valid index
	if( (index >= 0) & (udword(index) < objects_m.size()) )
	{
		pObject = objects_m[ index ].getObject();
	}

	return pObject;
}


const Material& Scene::getMaterial
(
	const dword index
) const
{
	const Material* pMaterial;

	if( (index >= 0) & (udword(index) < materials_m.size()) )
	{
		pMaterial = &(materials_m[ index ]);
	}
	else
	{
		pMaterial = &defaultMaterial_m;
	}

	return *pMaterial;
}


dword Scene::getMaterialIndex
(
	const string& name
) const
{
	return getIndex( name, "material", materialNames_m );
}


string Scene::getMaterialName
(
	const dword index
) const
{
	return getName( index, materialNames_m );
}


dword Scene::getObjectIndex
(
	const string& name
) const
{
	return getIndex( name, "object", objectNames_m );
}


string Scene::getObjectName
(
	const dword index
) const
{
	return getName( index, objectNames_m );
}




/// implementation ---------------------------------------------------------------------------------
void Scene::addName
(
	const string&       name,
	const char          type[],
	map<string, dword>& mapping
)
{
	/// check for empty name
	if( name.empty() )
	{
		/// prepare to throw exception
		string message;
		message += "null ";
		message += type;
		message += " name";

		throw hxa7241::utility::Exception( message.c_str(), CLASS_NAME(), "addName" );
	}

	/// add name to map, and check if its already present
	if( !mapping.insert( map<string, dword>::value_type(name, mapping.size()) ).second )
	{
		throwNameException( name, type, "duplicate", "addName" );
	}
}


dword Scene::getIndex
(
	const string&             name,
	const char                type[],
	const map<string, dword>& mapping
)
{
	const map<string, dword>::const_iterator it = mapping.find( name );

	/// check if name is not in map
	if( it == mapping.end() )
	{
		throwNameException( name, type, "unfound", "getIndex" );
	}

	return it->second;
}


string Scene::getName
(
	const dword               index,
	const map<string, dword>& mapping
)
{
	string name;

	map<string, dword>::const_iterator it = mapping.begin();
	std::advance( it, index );

	if( it != mapping.end() )
	{
		name = it->first;
	}

	return name;
}


void Scene::throwNameException
(
	const string& name,
	const char    type[],
	const char    problem[],
	const char    method[]
)
{
	string message;
	message += problem;
	message += " ";
	message += type;
	message += " name '";
	message += name;
	message += "'";

	throw hxa7241::utility::Exception( message.c_str(), CLASS_NAME(), method );
}


void Scene::enforceConstraints()
{
}




/// serialization ----------------------------------------------------------------------------------
template<>
inline
const char* SerializerOut::typeName<std::deque<Material> >()
{
	return "std::deque";
}

template<>
inline
const char* SerializerOut::typeDescription<std::deque<Material> >()
{
	return "standard deque collection";
}


template<>
inline
const char* SerializerOut::typeName<std::deque<ObjectInstantiator> >()
{
	return "std::deque";
}

template<>
inline
const char* SerializerOut::typeDescription<std::deque<ObjectInstantiator> >()
{
	return "standard deque collection";
}


SerializerOut& perceptuum2::modelling::operator<<( SerializerOut& out, const Scene& obj )
{
	/// write times
	{
		out.write( "timeBegin", obj.timeBegin_m );
		out.write( "timeEnd",   obj.timeEnd_m   );
	}

	/// write medium
	{
		const float mediumTransparency    = obj.defaultObject_m.getTransparency( ObjectRenderableId(), Vector3f::ZERO() );
		const float mediumRefractiveIndex = obj.defaultObject_m.getRefractiveIndex( ObjectRenderableId(), Vector3f::ZERO() );
		out.write( "mediumTransparency",    mediumTransparency,   "0-1" );
		out.write( "mediumRefractiveIndex", mediumRefractiveIndex );
	}

	/// write materials
	out.write( "materials", obj.materials_m, "some material definitions" );

	/// write objects
	out.write( "objects", obj.objects_m, "some object definitions" );

	return out;
}


SerializerIn& perceptuum2::modelling::operator>>( SerializerIn& in, Scene& obj )
{
	obj.materialNames_m.clear();


	/// read times
	{
		in.read( "timeBegin", obj.timeBegin_m );
		in.read( "timeEnd",   obj.timeEnd_m   );
	}


	/// read medium
	{
		float mediumTransparency;
		float mediumRefractiveIndex;
		in.read( "mediumTransparency",    mediumTransparency    );
		in.read( "mediumRefractiveIndex", mediumRefractiveIndex );

		obj.defaultObject_m.setMedium( mediumTransparency, mediumRefractiveIndex );
	}


//	if( obj.pParent_m != 0 )
//	{
//		perceptuum2::modelling::operator>>( in, *obj.pParent_m );
//	}
//	else
//	{

	/// read includes
	{
		/// if top layer scene, resize to 0
		//obj.includes_m.resize( 0 );
		//iterator latest = includes.end

		//in.read( "includes", obj.includes_m );
//for( unsigned i = 0;  i < obj.includes_m.size();  ++i )
//	hxa7241::utility::Logger() << "include:  " << obj.includes_m[i];   /// DEBUG ///

		/// if top layer scene, delete and resize to 0
			/// delete current objects
			/// resize objects to 0

		/// step thru latest includes
			/// add to set
			/// if already in set, erase from deque
		/// step thru latest includes
			// make UriFile from include string (throws if not file scheme)
			//Scene subScene( &obj );
			//XmlSerializerIn::serializeIn( subScene, uriFile.getPathname().c_str() );

		/// if top layer scene, empty includes set

		/// (in read objects section, dont delete or resize objects)
	}


	/// read materials
	/// if top layer scene, resize to 0
	obj.materials_m.resize( 0 );
	in.read( "materials", obj.materials_m, obj );


	/// read objects
	{
		/// delete current objects
		const deque<ObjectInstantiator>::iterator end( obj.objects_m.end() );
		for( deque<ObjectInstantiator>::iterator it( obj.objects_m.begin() );  it != end;  ++it )
		{
			delete it->getObject();
		}

		obj.objects_m.resize( 0 );
		in.read( "objects", obj.objects_m, obj );
	}


	//obj.enforceConstraints();

	return in;
}


const char* Scene::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::modelling::Scene";
	return CLASS_NAME;
}


const char* Scene::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "the geometrical and material description of a renderable subject";
	return CLASS_DESCRIPTION;
}
