/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Vector3f.h"
#include "UnitVector3f.h"
#include "FpToInt.h"

#include "Vertex.h"   /// own header is included last


using namespace perceptuum2::modelling;




/// static
const float Vertex::POSITION_SCALING     = 1.0f / 16382.0f;
const float Vertex::NORMAL_SCALING       = 1.0f / 63.5f;
const float Vertex::POSITION_SCALING_INV = 16382.0f;
const float Vertex::NORMAL_SCALING_INV   = 63.5f;




/// standard object services -----------------------------------------------------------------------
Vertex::Vertex()
 :	posXposYnormZlsb_m(),
	posZnormXnormYnormZmsb_m()
{
}


Vertex::Vertex
(
	const Vector3f&     meshPosition,
	const Vector3f&     oneOverMeshScale,
	const Vector3f&     position,
	const UnitVector3f& normal
)
{
	Vertex::set( meshPosition, oneOverMeshScale, position, normal );
}


Vertex::~Vertex()
{
}


Vertex::Vertex( const Vertex& other )
 :	posXposYnormZlsb_m      ( other.posXposYnormZlsb_m ),
	posZnormXnormYnormZmsb_m( other.posZnormXnormYnormZmsb_m )
{
}


Vertex& Vertex::operator=( const Vertex& other )
{
	posXposYnormZlsb_m       = other.posXposYnormZlsb_m;
	posZnormXnormYnormZmsb_m = other.posZnormXnormYnormZmsb_m;

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void Vertex::set
(
	const Vector3f&     meshPosition,
	const Vector3f&     oneOverMeshScale,
	const Vector3f&     position,
	const UnitVector3f& normal
)
{
	/// ((position - meshPosition) * oneOverMeshScale) must map into the [-0.5, 0.5] interval

	/// compress
	const dword positionX = hxa7241::utility::fpToInt( (position.x_m - meshPosition.x_m) * oneOverMeshScale.x_m * POSITION_SCALING_INV );
	const dword positionY = hxa7241::utility::fpToInt( (position.y_m - meshPosition.y_m) * oneOverMeshScale.y_m * POSITION_SCALING_INV );
	const dword positionZ = hxa7241::utility::fpToInt( (position.z_m - meshPosition.z_m) * oneOverMeshScale.z_m * POSITION_SCALING_INV );

	const dword normalX   = hxa7241::utility::fpToInt( (normal.getX() + 1.0f) * NORMAL_SCALING_INV );
	const dword normalY   = hxa7241::utility::fpToInt( (normal.getY() + 1.0f) * NORMAL_SCALING_INV );
	const dword normalZ   = hxa7241::utility::fpToInt( (normal.getZ() + 1.0f) * NORMAL_SCALING_INV );

	/// twiddle into bits
	posXposYnormZlsb_m       = (positionX << 18) | ((positionY & 0x3FFF) << 4) | (normalZ & 0x07);
	posZnormXnormYnormZmsb_m = (positionZ << 18) | ((normalX & 0x7F) << 11) | ((normalY & 0x7F) << 4) | ((normalZ & 0x7F) >> 3);
}




/// queries ----------------------------------------------------------------------------------------
void Vertex::getPosition
(
	const Vector3f& meshPosition,
	const Vector3f& meshScale,
	Vector3f&       position
) const
{
	/// twiddle out of bits
	const dword positionX = posXposYnormZlsb_m >> 18;
	const dword positionY = (posXposYnormZlsb_m << 14) >> 18;
	const dword positionZ = posZnormXnormYnormZmsb_m >> 18;

	/// decompress
	position.x_m = (float(positionX) * POSITION_SCALING * meshScale.x_m) + meshPosition.x_m;
	position.y_m = (float(positionY) * POSITION_SCALING * meshScale.y_m) + meshPosition.y_m;
	position.z_m = (float(positionZ) * POSITION_SCALING * meshScale.z_m) + meshPosition.z_m;
}


void Vertex::getNormal
(
	UnitVector3f& normal
) const
{
	/// twiddle out of bits
	const dword normalX = (posZnormXnormYnormZmsb_m >> 11) & 0x7F;
	const dword normalY = (posZnormXnormYnormZmsb_m >>  4) & 0x7F;
	const dword normalZ = ((posZnormXnormYnormZmsb_m << 3) & 0x7F) | (posXposYnormZlsb_m & 0x07);

	/// decompress
	normal.setXYZ( (float(normalX) * NORMAL_SCALING) - 1.0f,
	               (float(normalY) * NORMAL_SCALING) - 1.0f,
	               (float(normalZ) * NORMAL_SCALING) - 1.0f );
}
