/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Float01.h"
#include "Serialization.h"

#include "SamplingControlDiffuse.h"   /// own header is included last


using namespace perceptuum2::transport;




/// standard object services -----------------------------------------------------------------------
SamplingControlDiffuse::SamplingControlDiffuse
(
	const dword sobolDimensions[COORDS_COUNT]
)
 :	coordCache_m()
{
	for( dword i = sizeof(sobols_m)/sizeof(*sobols_m);  i-- > 0; )
	{
		sobols_m[i].setDimension( sobolDimensions[i] );
	}
}


SamplingControlDiffuse::~SamplingControlDiffuse()
{
}


SamplingControlDiffuse::SamplingControlDiffuse
(
	const SamplingControlDiffuse& other
)
 :	coordCache_m( other.coordCache_m )
{
	for( dword i = sizeof(sobols_m)/sizeof(*sobols_m);  i-- > 0; )
	{
		sobols_m[i] = other.sobols_m[i];
	}
}


SamplingControlDiffuse& SamplingControlDiffuse::operator=
(
	const SamplingControlDiffuse& other
)
{
	if( &other != this )
	{
		coordCache_m = other.coordCache_m;

		for( dword i = sizeof(sobols_m)/sizeof(*sobols_m);  i-- > 0; )
		{
			sobols_m[i] = other.sobols_m[i];
		}
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void SamplingControlDiffuse::advance()
{
	coordCache_m.resize( 0 );
}




/// queries ----------------------------------------------------------------------------------------
void SamplingControlDiffuse::getSampleCoords
(
	dword   index,
	Float01 coords[COORDS_COUNT]
) const
{
	/// precondition input
	if( index < 0 )
	{
		index = 0;
	}


	const dword coordCacheSetLength = dword(coordCache_m.size()) / COORDS_COUNT;

	/// check if cache needs to be extended
	if( index >= coordCacheSetLength )
	{
		/// add more of the sequence on the end
		const dword extend = (index - coordCacheSetLength) + 1;
		for( dword i = extend;  i-- > 0; )
		{
			for( dword j = 0, end = COORDS_COUNT;  j < end;  ++j )
			{
				coordCache_m.push_back( sobols_m[j].getItem() );
				sobols_m[j].next();
			}
		}
	}

	/// read coords from cache
	for( dword i = COORDS_COUNT;  i-- > 0; )
	{
		coords[i] = coordCache_m[(index * COORDS_COUNT) + i];
	}
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& perceptuum2::transport::operator<<( SerializerOut& out, const SamplingControlDiffuse& obj )
{
	for( dword i = 0, end = sizeof(obj.sobols_m)/sizeof(*obj.sobols_m);  i < end;  ++i )
	{
		out.write( "sobol", obj.sobols_m[i] );
	}

	{
		const dword cacheLength = obj.coordCache_m.size();
		out.write( "cacheLength", cacheLength );

		for( dword i = 0;  i < cacheLength;  ++i )
		{
			out.write( "coord", obj.coordCache_m[i] );
		}
	}

	return out;
}


SerializerIn& perceptuum2::transport::operator>>( SerializerIn& in, SamplingControlDiffuse& obj )
{
	for( dword i = 0, end = sizeof(obj.sobols_m)/sizeof(*obj.sobols_m);  i < end;  ++i )
	{
		in.read( "sobol", obj.sobols_m[i] );
	}

	{
		dword cacheLength = 0;
		in.read( "cacheLength", cacheLength );
		cacheLength = cacheLength >= 0 ? cacheLength : 0;

		obj.coordCache_m.resize( cacheLength );
		for( dword i = 0;  i < cacheLength;  ++i )
		{
			in.read( "coord", obj.coordCache_m[i] );
		}
	}

	return in;
}


const char* SamplingControlDiffuse::CLASS_NAME()
{
	static const char CLASS_NAME[] = "perceptuum2::transport::SamplingControlDiffuse";
	return CLASS_NAME;
}


const char* SamplingControlDiffuse::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "grouping of stochastic sequences used in diffuse gathers";
	return CLASS_DESCRIPTION;
}
