/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "RenderingTime.h"
#include "SamplingControl.h"
#include "Scene.h"
#include "Ray.h"
#include "Intersection.h"
#include "ObjectRenderable.h"

#include "SpacetimeIndex.h"   /// own header is included last


using namespace perceptuum2::transport;




/// standard object services -----------------------------------------------------------------------
SpacetimeIndex::SpacetimeIndex
(
	const Scene& scene
)
 :	pScene_m( &scene ),
	time_m( scene.getTimeBegin(), scene.getTimeEnd() )
{
}


SpacetimeIndex::~SpacetimeIndex()
{
}


SpacetimeIndex::SpacetimeIndex( const SpacetimeIndex& other )
{
	SpacetimeIndex::assign( other );
}


SpacetimeIndex& SpacetimeIndex::operator=( const SpacetimeIndex& other )
{
	assign( other );

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void SpacetimeIndex::update
(
	const SamplingControl& samplingControl
)
{
	time_m.setTimePoint( samplingControl.getTimePoint() );

	pScene_m->update( time_m.getTime() );
}




/// queries ----------------------------------------------------------------------------------------
const Scene& SpacetimeIndex::getScene() const
{
	return *pScene_m;
}


bool SpacetimeIndex::getNextIntersection
(
	const Ray&    ray,
	Intersection& intersection
) const
{
	const perceptuum2::modelling::ObjectRenderableId rootObject( pScene_m->getRootObject() );
	float distance = FLOAT_MAX;

	const bool isHit = rootObject->getIntersection( rootObject, ray, distance, intersection );

	return isHit & (distance < FLOAT_MAX);
}


bool SpacetimeIndex::isShadowed
(
	const Ray&                rayFromSource,
	const Vector3f&           ,//positionDestination,
	const ObjectRenderableId& objectDestination
) const
{
	/// do easy check if query point is in half space behind source point
	const bool isSelfShadowing = rayFromSource.isInside();

	bool isInShadow = isSelfShadowing;
	if( !isInShadow )
	{
		/// do general intersection test between source and query points
		Intersection intersection;
		const bool isHit = getNextIntersection( rayFromSource, intersection );

		/// intersection with query point primitive is not counted
		isInShadow = (isHit && (intersection.object_m != objectDestination));
	}

	return isInShadow;
}




/// implementation ---------------------------------------------------------------------------------
void SpacetimeIndex::assign
(
	const SpacetimeIndex& other
)
{
	if( &other != this )
	{
		pScene_m = other.pScene_m;
		time_m   = other.time_m;
	}
}
