/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "perceptuum2.rc.h"
#include "Perceptuum2.h"
#include "Win32Exception.h"
#include "Win32FileDialog.h"
#include "StopDialog.h"
#include "ToneDialog.h"
#include "InfoDialog.h"
#include "AboutDialog.h"
#include <string>
#include <vector>
#include "commctrl.h"

#include "ButtonBar.h"   /// own header is included last


using namespace perceptuum2::platform_ui;




/// statics
const char ButtonBar::CLASS_NAME[] = "perceptuum2::platform_ui::ButtonBar";
const char ButtonBar::KEYS[]       = "OPSGTIA";




/// standard object services -----------------------------------------------------------------------
ButtonBar::ButtonBar
(
	HWND hWndParent
)
 :	hInstance_m( getHInstance( hWndParent ) ),
	hWnd_m( createButtonBar( hWndParent, hInstance_m ) )
{
}


ButtonBar::~ButtonBar()
{
}




/// commands ---------------------------------------------------------------------------------------
bool ButtonBar::enableButton
(
	const int  id,
	const bool enableOrNot
) const
{
	return bool( ::SendMessage( hWnd_m, TB_ENABLEBUTTON, (WPARAM)id, (LPARAM)(BOOL)enableOrNot ) );
}


bool ButtonBar::enableButtonsForNoRender() const
{
	static const int  ids[]          = { IDM_GLARE, IDM_SAVE, IDM_TONEMAP, IDM_INFO };
	static const bool enableOrNots[] = { false,     false,    false,       false };

	return enableButtons( ids, enableOrNots, sizeof(ids) / sizeof(int) );
}


bool ButtonBar::enableButtonsForStart() const
{
	static const int  ids[]          = { IDM_START, IDM_STOP, IDM_SAVE, IDM_GLARE, IDM_TONEMAP, IDM_INFO };
	static const bool enableOrNots[] = { false,     true,     true,     false,     true,        true };

	return enableButtons( ids, enableOrNots, sizeof(ids) / sizeof(int) );
}


bool ButtonBar::enableButtonsForStop() const
{
	static const int  ids[]          = { IDM_START, IDM_STOP, IDM_GLARE };
	static const bool enableOrNots[] = { true,      false,    true };

	return enableButtons( ids, enableOrNots, sizeof(ids) / sizeof(int) );
}


void ButtonBar::onSize
(
	WPARAM wParam,
	LPARAM lParam
)
{
	::SendMessage( hWnd_m, WM_SIZE, lParam, wParam );
}


bool ButtonBar::onCommand
(
	WPARAM       wParam,
	Perceptuum2& p2
)
{
	bool didHandleMessage = true;

	switch( wParam )
	{
		case IDM_START :
		{
			onStart( p2 );
			break;
		}
		case IDM_STOP :
		{
			onStop( p2 );
			break;
		}
		case IDM_SAVE :
		{
			onSave( p2 );
			break;
		}
		case IDM_GLARE :
		{
			onGlare( p2 );
			break;
		}
		case IDM_TONEMAP :
		{
			onTone( p2 );
			break;
		}
		case IDM_INFO :
		{
			onInfo( p2 );
			break;
		}
		case IDM_ABOUT :
		{
			onAbout();
			break;
		}
		default :
		{
			didHandleMessage = false;
			break;
		}
	}

	return didHandleMessage;
}


bool ButtonBar::onCommandKey
(
	WPARAM       wParam,
	Perceptuum2& p2
)
{
	bool didHandleMessage = true;

	if( wParam == WPARAM(KEYS[START_KEY]) )
	{
		if( isButtonEnabled( IDM_START ) )
		{
			onStart( p2 );
		}
	}
	else if( wParam == WPARAM(KEYS[STOP_KEY]) )
	{
		if( isButtonEnabled( IDM_STOP ) )
		{
			onStop( p2 );
		}
	}
	else if( wParam == WPARAM(KEYS[SAVE_KEY]) )
	{
		if( isButtonEnabled( IDM_SAVE ) )
		{
			onSave( p2 );
		}
	}
	else if( wParam == WPARAM(KEYS[GLARE_KEY]) )
	{
		if( isButtonEnabled( IDM_GLARE ) )
		{
			onGlare( p2 );
		}
	}
	else if( wParam == WPARAM(KEYS[TONEMAP_KEY]) )
	{
		if( isButtonEnabled( IDM_TONEMAP ) )
		{
			onTone( p2 );
		}
	}
	else if( wParam == WPARAM(KEYS[INFO_KEY]) )
	{
		if( isButtonEnabled( IDM_INFO ) )
		{
			onInfo( p2 );
		}
	}
	else if( wParam == WPARAM(KEYS[ABOUT_KEY]) )
	{
		if( isButtonEnabled( IDM_ABOUT ) )
		{
			onAbout();
		}
	}
	else
	{
		didHandleMessage = false;
	}

	return didHandleMessage;
}



/// queries ----------------------------------------------------------------------------------------
bool ButtonBar::getWindowRect
(
	RECT& rect
) const
{
	return bool( ::GetWindowRect( hWnd_m, &rect) );
}


bool ButtonBar::isButtonEnabled
(
	const int id
) const
{
	return bool( ::SendMessage( hWnd_m, TB_ISBUTTONENABLED, (WPARAM)id, (LPARAM)0 ) );
}




/// construction -----------------------------------------------------------------------------------
HINSTANCE ButtonBar::getHInstance
(
	HWND hWnd
)
{
	static const char METHOD_NAME[] = "getHInstance";

	HINSTANCE hInstance = (HINSTANCE)::GetWindowLong( hWnd, GWL_HINSTANCE );
	if( hInstance == 0 )
	{
		const DWORD error = ::GetLastError();
		throw hxa7241::win32::Win32Exception( IDS_EXC_HINSTANCE, error, CLASS_NAME, METHOD_NAME );
	}

	return hInstance;
}


HWND ButtonBar::createButtonBar
(
	HWND      hWndParent,
	HINSTANCE hInstance
)
{
	static const char METHOD_NAME[] = "createButtonBar";

	static const int noOfButtons   = 7;
	static const int separator1Pos = 2;
	static const int separator2Pos = 7;


	/// create window
	HWND hWndButtonBar = ::CreateWindowEx( 0, TOOLBARCLASSNAME, NULL, WS_CHILD | WS_VISIBLE, 0, 0, 0, 0, hWndParent, (HMENU)IDR_BUTTONBAR, hInstance, NULL );
	if( hWndButtonBar == 0 )
	{
		const DWORD error = ::GetLastError();
		throw hxa7241::win32::Win32Exception( IDS_EXC_BUTTONBAR_WINDOW, error, CLASS_NAME, METHOD_NAME );
	}

	::SendMessage( hWndButtonBar, TB_BUTTONSTRUCTSIZE, (WPARAM)sizeof(TBBUTTON), 0 );


	/// fill in button details
	TBBUTTON tbButtons[ noOfButtons ];
	{
		static const UINT  strings[]  = { IDS_START, IDS_STOP, IDS_SAVE, IDS_GLARE, IDS_TONEMAP, IDS_INFO, IDS_ABOUT };
		static const DWORD commands[] = { IDM_START, IDM_STOP, IDM_SAVE, IDM_GLARE, IDM_TONEMAP, IDM_INFO, IDM_ABOUT };
		TCHAR buf[10];
		for( int i = 0;  i < noOfButtons;  ++i )
		{
			const bool lsok = bool( ::LoadString( hInstance, strings[i], (LPSTR)&buf, sizeof(buf)/sizeof(TCHAR) ) );
			LRESULT iString = -1;
			if( lsok )
			{
				iString = ::SendMessage( hWndButtonBar, TB_ADDSTRING, 0, (LPARAM)(LPSTR)buf );
			}

			TBBUTTON* ptbButton  = &(tbButtons[i]);
			ptbButton->iBitmap   = i;
			ptbButton->idCommand = commands[i];
			ptbButton->fsState   = TBSTATE_ENABLED;
			ptbButton->fsStyle   = TBSTYLE_BUTTON;
			ptbButton->dwData    = 0;
			ptbButton->iString   = iString;
		}
	}
	TBBUTTON tbSeparator;
	{
		tbSeparator.iBitmap   = 0;
		tbSeparator.idCommand = 0;
		tbSeparator.fsState   = 0;
		tbSeparator.fsStyle   = TBSTYLE_SEP;
		tbSeparator.dwData    = 0;
		tbSeparator.iString   = 0;
	}
	TBADDBITMAP tbab = { hInstance, IDB_TOOLBAR_IMAGES };


	/// set buttons in bar
	const bool ok = bool( ::SendMessage( hWndButtonBar, TB_ADDBUTTONS, (WPARAM)noOfButtons, (LPARAM)(LPTBBUTTON)&tbButtons ) );
	if( !ok )
	{
		const DWORD error = ::GetLastError();

		::DestroyWindow( hWndButtonBar );
		throw hxa7241::win32::Win32Exception( IDS_EXC_BUTTONBAR_BUTTONS, error, CLASS_NAME, METHOD_NAME );
	}
	::SendMessage( hWndButtonBar, TB_ADDBITMAP, (WPARAM)noOfButtons, (LPARAM)&tbab );
	::SendMessage( hWndButtonBar, TB_INSERTBUTTON, (WPARAM)separator1Pos, (LPARAM)(LPTBBUTTON)&tbSeparator );
	::SendMessage( hWndButtonBar, TB_INSERTBUTTON, (WPARAM)separator2Pos, (LPARAM)(LPTBBUTTON)&tbSeparator );
	::SendMessage( hWndButtonBar, TB_AUTOSIZE, (WPARAM)0, (LPARAM)0 );


	return hWndButtonBar;
}


bool ButtonBar::enableButtons
(
	const int  ids[],
	const bool enableOrNots[],
	const int  len
) const
{
	bool ok = true;

	for( int i = len;  i-- > 0; )
	{
		ok &= bool( ::SendMessage( hWnd_m, TB_ENABLEBUTTON, (WPARAM)(ids[i]), (LPARAM)(BOOL)(enableOrNots[i]) ) );
	}

	/// always disable glare button - glare not implemented yet
	ok &= bool( ::SendMessage( hWnd_m, TB_ENABLEBUTTON, (WPARAM)(IDM_GLARE), (LPARAM)(BOOL)(false) ) );

	return ok;
}




/// message handlers -------------------------------------------------------------------------------
void ButtonBar::onStart
(
	Perceptuum2& p2
)
{
	/// get file filter
	std::vector<std::string> names;
	p2.getStartFileNames( names );


	/// make and run dialog
	TCHAR pDialogTitle[64];
	::LoadString( hInstance_m, IDS_START_DLGTITLE, (LPSTR)&pDialogTitle, sizeof(pDialogTitle)/sizeof(TCHAR) );

	Win32FileDialog dlg( pDialogTitle, 0, names, false );

	std::string filePathname;
	int         namePos;
	int         extensionPos;
	const bool ok = dlg.run( hInstance_m, hWnd_m, filePathname, namePos, extensionPos );
	if( ok )
	{
		/// command app
		p2.start( filePathname, namePos, extensionPos );
	}
}


void ButtonBar::onStop
(
	Perceptuum2& p2
)
{
	/// stop app timer
	p2.enableTimer( false );

	/// make and run dialog
	StopDialog dialog( hInstance_m, hWnd_m, p2 );
	const int stop = dialog.run();

	if( stop != dialog.STOP_CANCEL )
	{
		p2.stop( stop );
	}
	else
	{
		/// restart app timer
		p2.enableTimer( true );
	}
}


void ButtonBar::onSave
(
	Perceptuum2& p2
)
{
	/// get file filter
	std::vector<std::string> names;
	p2.getFormattingNames( names );


	/// make and run dialog
	TCHAR pDialogTitle[64];
	::LoadString( hInstance_m, IDS_SAVE_DLGTITLE, (LPSTR)&pDialogTitle, sizeof(pDialogTitle)/sizeof(TCHAR) );

	Win32FileDialog dlg( pDialogTitle, 0, names );

	std::string filePathname;
	int         namePos;
	int         extensionPos;
	const bool ok = dlg.run( hInstance_m, hWnd_m, filePathname, namePos, extensionPos );
	if( ok )
	{
		p2.save( filePathname, extensionPos );
	}
}


void ButtonBar::onGlare
(
	Perceptuum2&
)
{
	::MessageBox( hWnd_m, "not done yet", "Perceptuum2", MB_OK | MB_ICONINFORMATION );
}


void ButtonBar::onTone
(
	Perceptuum2& p2
)
{
	float f = p2.getTonemapAdaptationLuminance();

	ToneDialog dialog( hInstance_m, hWnd_m, &p2 );
	dialog.run( f );

//	/// for non-interactive dialog
//	const bool ok = bool( dialog.run( f ) );
//	if( ok )
//	{
//		p2.setTonemapAdaptationLuminance( f );
//	}
}


void ButtonBar::onInfo
(
	Perceptuum2& p2
)
{
	std::vector<std::string> names;
	p2.getSummaryNames( names );

	std::vector<std::string> summaries;
	p2.getSummaries( summaries );

	InfoDialog dialog( hInstance_m, hWnd_m );
	dialog.run( &names, &summaries );
}


void ButtonBar::onAbout()
{
	AboutDialog about( hInstance_m, hWnd_m, ::GetDC( NULL ) );
	about.run();
}
