/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Win32BackBuffer.h"
#include "Win32Dib.h"
#include "Logger.h"
#include "Win32Exception.h"
#include "perceptuum2.rc.h"
#include <string>

#include "Win32ImagePaneSlideable.h"   /// own header is included last


using namespace perceptuum2::platform_ui;




/// statics
const TCHAR Win32ImagePaneSlideable::WND_CLASS_NAME[] = TEXT("ImagePaneWindowClass");
const char  Win32ImagePaneSlideable::CLASS_NAME[]     = "perceptuum2::platform_ui::Win32ImagePaneSlideable";
const char  Win32ImagePaneSlideable::ERROR_MESSAGE1[] = "allocation for backbuffer failed";

const COLORREF Win32ImagePaneSlideable::BACKGROUND_DEFAULT = 0x00404040;//0x00728080;




/// standard object services -----------------------------------------------------------------------
Win32ImagePaneSlideable::Win32ImagePaneSlideable
(
	HINSTANCE hInstance,
	HWND      hWndParent
)
 :	hWnd_m( 0 ),
	pBackBuffer_m( 0 ),
	hBrushBackground_m( 0 ),
	brushColor_m( 0 ),
	imageIsGripped_m( false ),
	normalCursor_m( 0 ),
	handCursor_m( 0 ),
	handGrippedCursor_m( 0 ),
	mousePos_m(),
	imagePos_m()
{
	Win32ImagePaneSlideable::setBackground( BACKGROUND_DEFAULT );

	/// init cursors
	{
		static const char METHOD_NAME[] = "constructor";

		using hxa7241::win32::Win32Exception;

		normalCursor_m      = ::LoadCursor( NULL, IDC_ARROW );
		if( normalCursor_m == NULL )
		{
			throw Win32Exception( IDS_EXC_CURSORLOAD, ::GetLastError(), CLASS_NAME, METHOD_NAME );
		}

		handCursor_m        = ::LoadCursor( hInstance, MAKEINTRESOURCE(IDC_HAND1) );
		if( handCursor_m == NULL )
		{
			throw Win32Exception( IDS_EXC_CURSORLOAD, ::GetLastError(), CLASS_NAME, METHOD_NAME );
		}

		handGrippedCursor_m = ::LoadCursor( hInstance, MAKEINTRESOURCE(IDC_HAND2) );
		if( handGrippedCursor_m == NULL )
		{
			throw Win32Exception( IDS_EXC_CURSORLOAD, ::GetLastError(), CLASS_NAME, METHOD_NAME );
		}
	}

	/// default positions
	{
		mousePos_m.x = 0;
		mousePos_m.y = 0;
		imagePos_m.x = 0;
		imagePos_m.y = 0;
	}

	registerWindowClass( hInstance );
	hWnd_m = createWindow( hInstance, hWndParent );
}


Win32ImagePaneSlideable::~Win32ImagePaneSlideable()
{
	::DeleteObject( hBrushBackground_m );
	delete pBackBuffer_m;
}




/// commands ---------------------------------------------------------------------------------------
void Win32ImagePaneSlideable::moveWindow
(
	const POINT& leftTop,
	const POINT& widthHeight
)
{
	::MoveWindow( hWnd_m, leftTop.x, leftTop.y, widthHeight.x, widthHeight.y, TRUE );
}


void Win32ImagePaneSlideable::setImageSize
(
	const dword width,
	const dword height
)
{
	static const char METHOD_NAME[] = "setImageSize";

	delete pBackBuffer_m;
	pBackBuffer_m = 0;

	HDC dc = ::GetDC( hWnd_m );
	if( dc == NULL )
	{
		throw hxa7241::win32::Win32Exception( IDS_EXC_BACKBUFFER_DC, ::GetLastError(), CLASS_NAME, METHOD_NAME );
	}

	pBackBuffer_m = new Win32BackBuffer( dc, width, height );
	if( pBackBuffer_m == 0 )
	{
		throw hxa7241::utility::Exception( ERROR_MESSAGE1, CLASS_NAME, METHOD_NAME );
	}


	if( isCursorInPane( hWnd_m ) )
	{
		setCursor( hWnd_m );
	}

	updateImage();
}


void Win32ImagePaneSlideable::updateImage
(
	const dword lineNumber
)
{
	HDC dc = ::GetDC( hWnd_m );
	if( dc != 0 )
	{
		RECT rect;
		const BOOL ok = ::GetClientRect( hWnd_m, &rect );
		if( ok )
		{
			/// blit line
			const POINT targetSize = { rect.right, rect.bottom };
			const RECT  sourceRect = { 0, lineNumber, pBackBuffer_m->getWidth(), lineNumber + 1 };
			pBackBuffer_m->blitToPositionWithBorder( dc, targetSize, &sourceRect, imagePos_m, hBrushBackground_m );
		}
	}
}


void Win32ImagePaneSlideable::updateImage()
{
	HDC dc = ::GetDC( hWnd_m );
	if( dc != 0 )
	{
		RECT rect;
		const BOOL ok = ::GetClientRect( hWnd_m, &rect );
		if( ok )
		{
			const POINT targetSize = { rect.right, rect.bottom };
			pBackBuffer_m->blitToPositionWithBorder( dc, targetSize, 0, imagePos_m, hBrushBackground_m );
		}
	}

//	::InvalidateRect( hWnd_m, NULL, FALSE );
//	::UpdateWindow( hWnd_m );
}


void Win32ImagePaneSlideable::setBackground
(
	COLORREF colorref
)
{
	::DeleteObject( hBrushBackground_m );
	hBrushBackground_m = ::CreateSolidBrush( colorref );
	brushColor_m       = colorref;

	::InvalidateRect( hWnd_m, NULL, FALSE );
}




/// queries ----------------------------------------------------------------------------------------
void Win32ImagePaneSlideable::getImage
(
	dword*& pixels,
	dword&  width,
	dword&  height
) const
{
	if( pBackBuffer_m != 0 )
	{
		using hxa7241::win32::Win32Dib;
		const Win32Dib* pDib = pBackBuffer_m->getWin32Dib();

		pixels = static_cast<dword*>( pDib->getPixels() );
		width  = pDib->getWidth();
		height = pDib->getHeight();
	}
	else
	{
		width  = 0;
		height = 0;
		pixels = 0;
	}
}


COLORREF Win32ImagePaneSlideable::getBackground() const
{
	return brushColor_m;
}




/// implementation ---------------------------------------------------------------------------------
LRESULT CALLBACK imagePaneProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam );


void Win32ImagePaneSlideable::registerWindowClass
(
	HINSTANCE hInstance
)
{
	static const char METHOD_NAME[] = "registerWindowClass";

	static bool isRegistered = false;
	if( !isRegistered )
	{
		WNDCLASSEX wc;

		wc.cbSize        = sizeof(WNDCLASSEX);
		wc.style         = CS_OWNDC | CS_DBLCLKS | CS_SAVEBITS | CS_HREDRAW | CS_VREDRAW;
		wc.lpfnWndProc   = (WNDPROC)imagePaneProc;
		wc.cbClsExtra    = 0;
		wc.cbWndExtra    = 0;
		wc.hInstance     = hInstance;
		wc.hIcon         = NULL;
		wc.hCursor       = NULL;
		wc.hbrBackground = NULL;
		wc.lpszMenuName  = NULL;
		wc.lpszClassName = WND_CLASS_NAME;
		wc.hIconSm       = NULL;

		if( ::RegisterClassEx( &wc ) == 0 )
		{
			throw hxa7241::win32::Win32Exception( IDS_EXC_IMAGEPANE_REG, ::GetLastError(), CLASS_NAME, METHOD_NAME );
		}

		isRegistered = true;
	}
}


HWND Win32ImagePaneSlideable::createWindow
(
	HINSTANCE hInstance,
	HWND      hWndParent
)
{
	static const char METHOD_NAME[] = "createWindow";


	/// madness, but how else am i supposed to get a pointer to the object associated with the window ???
	/// (the void pointer member of createstruct doesnt work)
	const dword thisAsDword  = reinterpret_cast<dword>( this );
	const TCHAR thisAsName[] = { TCHAR( (thisAsDword      ) & 0xFF ), TCHAR( (thisAsDword >>  8) & 0xFF ),
	                             TCHAR( (thisAsDword >> 16) & 0xFF ), TCHAR( (thisAsDword >> 24) & 0xFF ), TEXT('\0') };

	HWND hWndImagePane = ::CreateWindow( WND_CLASS_NAME, thisAsName,
										 WS_CHILDWINDOW | WS_VISIBLE,
										 0, 0, 0, 0,
										 hWndParent, (HMENU)NULL, hInstance, NULL );
	if( hWndImagePane == NULL )
	{
		throw hxa7241::win32::Win32Exception( IDS_EXC_IMAGEPANE_CREATE, ::GetLastError(), CLASS_NAME, METHOD_NAME );
	}

	return hWndImagePane;
}


LRESULT CALLBACK imagePaneProc
(
	HWND   hWnd,
	UINT   message,
	WPARAM wParam,
	LPARAM lParam
)
{
	/// set up association between hWnd and object
	/// (should be the first message (i hope))
	if( message == WM_NCCREATE )
	{
		/// extract object ptr
		const TCHAR* wndName   = ((LPCREATESTRUCT)lParam)->lpszName;
		const dword  objectPtr = dword(ubyte(wndName[0])) | (dword(ubyte(wndName[1])) << 8) | (dword(ubyte(wndName[2])) << 16) | (dword(ubyte(wndName[3])) << 24);

		::SetLastError( 0 );
		/// init hWnd with object ptr
		const LONG ok = ::SetWindowLong( hWnd, GWL_USERDATA, static_cast<LONG>(objectPtr) );

		/// quit if fail
		if( (ok == 0) && (::GetLastError() != 0) )
		{
			return FALSE;
		}
	}

	/// dispatch message
	LRESULT r;
	{
		/// get the object ptr for this hWnd
		const LONG objectPtrAsLong = ::GetWindowLong( hWnd, GWL_USERDATA );
		Win32ImagePaneSlideable* pObject = reinterpret_cast<Win32ImagePaneSlideable*>( objectPtrAsLong );
		if( pObject != 0 )
		{
			r = pObject->windowProc( hWnd, message, wParam, lParam );
		}
		else
		{
			r = ::DefWindowProc( hWnd, message, wParam, lParam );
		}
	}

	return r;
}




/// message handling -------------------------------------------------------------------------------
LRESULT Win32ImagePaneSlideable::windowProc
(
	HWND   hWnd,
	UINT   message,
	WPARAM wParam,
	LPARAM lParam
)
{
	bool didHandleMessage = true;

	try
	{
		switch( message )
		{
			case WM_CREATE :
			{
				onCreate( hWnd );
				didHandleMessage = false;
				break;
			}
			case WM_PAINT :
			{
				onPaint( hWnd );
				break;
			}
			case WM_LBUTTONDOWN :
			{
				onLButtonDown( hWnd );
				break;
			}
			case WM_LBUTTONUP :
			{
				onLButtonUp( hWnd );
				break;
			}
			case WM_MOUSEMOVE :
			{
				onMouseMove( hWnd, message, wParam, lParam );
				break;
			}
			default :
			{
				didHandleMessage = false;
				break;
			}
		}
	}
	catch( const std::exception& e )
	{
		using hxa7241::utility::Logger;
		hxa7241::utility::Logger() << Logger::EXC2 << e.what();
	}
	catch( ... )
	{
		hxa7241::utility::Exception e;
		using hxa7241::utility::Logger;
		Logger() << Logger::EXC2 << e.what();
	}

	LRESULT lResult = 0;
	if( !didHandleMessage )
	{
		lResult = ::DefWindowProc( hWnd, message, wParam, lParam );
	}

	return lResult;
}


void Win32ImagePaneSlideable::onCreate
(
	HWND hWnd
) const
{
	if( isCursorInPane( hWnd ) )
	{
		::SetCursor( normalCursor_m );
	}
}


void Win32ImagePaneSlideable::onPaint
(
	HWND hWnd
)
{
	PAINTSTRUCT ps;
	HDC dc = ::BeginPaint( hWnd, &ps );

	if( dc != 0 )
	{
		bool ok = false;

		if( pBackBuffer_m != 0 )
		{
			RECT rect;
			ok = bool( ::GetClientRect( hWnd, &rect ) );
			if( ok )
			{
				const POINT targetSize = { rect.right, rect.bottom };
				pBackBuffer_m->blitToPositionWithBorder( dc, targetSize, 0, imagePos_m, hBrushBackground_m );
			}
		}

		if( !ok )
		{
			::FillRect( dc, &ps.rcPaint, hBrushBackground_m );
		}
	}

	::EndPaint( hWnd, &ps );
}


void Win32ImagePaneSlideable::onLButtonDown
(
	HWND hWnd
)
{
	if( isPaneSmallerThanImage( hWnd ) )
	{
		imageIsGripped_m = true;
		::SetCapture( hWnd );
		::SetCursor( handGrippedCursor_m );
	}
}


void Win32ImagePaneSlideable::onLButtonUp
(
	HWND hWnd
)
{
	::ReleaseCapture();
	imageIsGripped_m = false;
	setCursor( hWnd );
}


void Win32ImagePaneSlideable::onMouseMove
(
	HWND   hWnd,
	UINT   ,//message,
	WPARAM ,//wParam,
	LPARAM lParam
)
{
	setCursor( hWnd );


	const POINT mousePos = { (LONG(lParam) << 16) >> 16, LONG(lParam) >> 16 };

	if( imageIsGripped_m )
	{
		const POINT mouseMovement = { (mousePos.x - mousePos_m.x), (mousePos.y - mousePos_m.y) };

		imagePos_m.x += mouseMovement.x;
		imagePos_m.y += mouseMovement.y;

		::InvalidateRect( hWnd, NULL, FALSE );
		//::UpdateWindow( hWnd );
	}

	mousePos_m = mousePos;
}


bool Win32ImagePaneSlideable::isCursorInPane
(
	HWND hWnd
) const
{
	POINT pos;
	const BOOL ok1 = ::GetCursorPos( &pos );

	RECT pane;
	const BOOL ok2 = ::GetWindowRect( hWnd, &pane );

	return bool(ok1 && ok2) && bool(::PtInRect( &pane, pos ));
}


bool Win32ImagePaneSlideable::isPaneSmallerThanImage
(
	HWND hWnd
) const
{
	RECT rect;
	const BOOL ok = ::GetClientRect( hWnd, &rect );

	return !bool(ok) || ( (pBackBuffer_m != 0) &&
	       ( (rect.right  < pBackBuffer_m->getWidth()) ||
	         (rect.bottom < pBackBuffer_m->getHeight()) ) );
}


void Win32ImagePaneSlideable::setCursor
(
	HWND hWnd
) const
{
	HCURSOR cursor = normalCursor_m;

	if( isPaneSmallerThanImage( hWnd ) )
	{
		if( imageIsGripped_m )
		{
			cursor = handGrippedCursor_m;
		}
		else
		{
			cursor = handCursor_m;
		}
	}

	::SetCursor( cursor );
}
