/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Float01.h"
//#include <algorithm>

#include "Histogram.h"   /// own header is included last


using namespace hxa7241::utility;




/// standard object services -----------------------------------------------------------------------
Histogram::Histogram
(
	vector<float>& storage,
	const dword    length
)
 :	composition_m( storage )
{
	composition_m.resize( 0 );
	composition_m.reserve( length );
}


Histogram::~Histogram()
{
}




/// commands ---------------------------------------------------------------------------------------
void Histogram::addElement
(
	float element
)
{
	/// ensure vector of ascending (or flat) values only
	if( element < 0.0f )
	{
		element = 0.0f;
	}

	/// get current total
	float elementSum = 0.0f;
	if( composition_m.size() > 0 )
	{
		/// get last element
		elementSum = composition_m.back();
	}

	/// accumulate to current element
	composition_m.push_back( elementSum + element );
}




/// queries ----------------------------------------------------------------------------------------
bool Histogram::findElement
(
	const Float01 point,
	dword&        index
) const
{
	float discarded;
	return findElement( point, index, discarded );
}


bool Histogram::findElement
(
	const Float01 point,
	dword&        index,
	float&        subElementPosition
) const
{
	bool isFound = false;

	const dword length = composition_m.size();
	if( length > 0 )
	{
		/// make point between 0 and largest element
		const float largest     = composition_m.back();
		const float pointScaled = point * largest;

		if( pointScaled < largest )
		{
			/// find index of first element in the composition that is *greater* than scaled point
			vector<float>::iterator pointIter = std::upper_bound( composition_m.begin(), composition_m.end(), pointScaled );
			const dword indexMaybe = std::distance( composition_m.begin(), pointIter );

			if( indexMaybe < length )
			{
				index   = indexMaybe;
				isFound = true;
			}
		}
		else
		{
			if( largest > 0.0f )
			{
				/// for point equal to non-zero end (largest element), result is last element
				index   = length - 1;
				isFound = true;
			}
		}

		/// rescale point
		if( isFound )
		{
			const float lowerElement = (index != 0) ? composition_m[index - 1] : 0.0f;
			subElementPosition = (pointScaled - lowerElement) / (composition_m[index] - lowerElement);
		}
	}

	return isFound;
}


dword Histogram::getSize() const
{
	return composition_m.size();
}


Float01 Histogram::getElement
(
	const dword index
) const
{
	return Float01(composition_m[index] / composition_m.back());
}
