/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Serialization.h"

#include "RandomGood.h"   /// own header is included last


using namespace hxa7241::utility;




/// statics
const float RandomGood::AM = float( 1.0 / 2147483648.0 );




/// standard object services -----------------------------------------------------------------------
RandomGood::RandomGood
(
	const dword seed
)
 :	Random()
{
	RandomGood::setSeed( seed );
}


RandomGood::~RandomGood()
{
}


RandomGood::RandomGood( const RandomGood& other )
 :	Random()
{
	RandomGood::operator=( other );
}


RandomGood& RandomGood::operator=( const RandomGood& other )
{
	if( &other != this )
	{
		Random::operator=( other );

		idum_m = other.idum_m;
		iy_m   = other.iy_m;

		for( int j = NTAB;  j-- > 0; )
		{
			iv_m[ j ] = other.iv_m[ j ];
		}
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void RandomGood::setSeed
(
	dword seed
)
{
	seed = (seed + 1) & 0x7FFFFFFF;
	seed = seed != 0 ? seed : 1;                    /// prevent  idum = 0

	idum_m = seed;

	for( int j = NTAB + 7;  j >= 0;  j-- )          /// load the shuffle table (after 8 warm-ups)
	{
		const dword k = idum_m / IQ;                ///
		idum_m = IA * (idum_m - k * IQ) - IR * k;   /// (same as sequence part)
		idum_m += (idum_m >> 31) & IM;              ///
		//if( idum_m < 0 ) idum_m += IM;            ///

		if( j < NTAB ) iv_m[ j ] = idum_m;
	}

	iy_m = iv_m[ 0 ];


	next();
}


void RandomGood::next()
{
	const dword k = idum_m / IQ;                /// Compute  idum  =  (IA * idum) % IM  without
	idum_m = IA * (idum_m - k * IQ) - IR * k;   ///    overflow, by Shrage's method
	idum_m += (idum_m >> 31) & IM;              /// (avoid jump)
	//if( idum_m < 0 ) idum_m += IM;            ///

	const int j = iy_m / NDIV;                  /// In the range  0...NTAB-1
	iy_m        = iv_m[ j ];                    /// Output previously stored value,
	iv_m[ j ]   = idum_m;                       ///    and refill the shuffle table
}




/// queries ----------------------------------------------------------------------------------------
dword RandomGood::getDword() const
{
	return iy_m;                                /// result not quite 0 ... SDWORD max
}


float RandomGood::getFloat() const
{
	return float(udword(iy_m)) * AM;
}




/// serialization ----------------------------------------------------------------------------------
SerializerOut& hxa7241::utility::operator<<( SerializerOut& out, const RandomGood& obj )
{
	out.write( "idum", obj.idum_m );
	out.write( "iy",   obj.iy_m );

	for( dword i = 0;  i < obj.NTAB;  ++i )
	{
		out.write( "iv", obj.iv_m[i] );
	}

	return out;
}


SerializerIn& hxa7241::utility::operator>>( SerializerIn& in, RandomGood& obj )
{
	in.read( "idum", obj.idum_m );
	in.read( "iy",   obj.iy_m );

	for( dword i = 0;  i < obj.NTAB;  ++i )
	{
		in.read( "iv", obj.iv_m[i] );
	}

	return in;
}


const char* RandomGood::CLASS_NAME()
{
	static const char CLASS_NAME[] = "hxa7241::utility::RandomGood";
	return CLASS_NAME;
}


const char* RandomGood::CLASS_DESCRIPTION()
{
	static const char CLASS_DESCRIPTION[] = "the random number generator from 'Numerical Recipes In C', ch7 p280-281";
	return CLASS_DESCRIPTION;
}
