/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "RandomMT.h"   /// own header is included last


using namespace hxa7241::utility;




/// standard object services -----------------------------------------------------------------------
RandomMT::RandomMT
(
	const dword seed
)
 :	Random(),
	state_m( new uint32[N+1] ),
	next_m(),
	left_m( -1 ),
	random_m()
{
	RandomMT::setSeed( seed );
}


RandomMT::~RandomMT()
{
	delete[] state_m;
}


RandomMT::RandomMT( const RandomMT& other )
 :	Random(),
	state_m( new uint32[N+1] ),
	next_m(),
	left_m( -1 ),
	random_m()
{
	RandomMT::operator=( other );
}


RandomMT& RandomMT::operator=( const RandomMT& other )
{
	if( &other != this )
	{
		Random::operator=( other );

		next_m = state_m + (other.next_m - other.state_m);
		left_m = other.left_m;

		for( int j = N+1;  j-- > 0; )
		{
			state_m[ j ] = other.state_m[ j ];
		}

		random_m = other.random_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void  RandomMT::setSeed
(
	const dword seed_
)
{
	//
	// We initialize state[0..(N-1)] via the generator
	//
	//   x_new = (69069 * x_old) mod 2^32
	//
	// from Line 15 of Table 1, p. 106, Sec. 3.3.4 of Knuth's
	// _The Art of Computer Programming_, Volume 2, 3rd ed.
	//
	// Notes (SJC): I do not know what the initial state requirements
	// of the Mersenne Twister are, but it seems this seeding generator
	// could be better.  It achieves the maximum period for its modulus
	// (2^30) iff x_initial is odd (p. 20-21, Sec. 3.2.1.2, Knuth); if
	// x_initial can be even, you have sequences like 0, 0, 0, ...;
	// 2^31, 2^31, 2^31, ...; 2^30, 2^30, 2^30, ...; 2^29, 2^29 + 2^31,
	// 2^29, 2^29 + 2^31, ..., etc. so I force seed to be odd below.
	//
	// Even if x_initial is odd, if x_initial is 1 mod 4 then
	//
	//   the          lowest bit of x is always 1,
	//   the  next-to-lowest bit of x is always 0,
	//   the 2nd-from-lowest bit of x alternates      ... 0 1 0 1 0 1 0 1 ... ,
	//   the 3rd-from-lowest bit of x 4-cycles        ... 0 1 1 0 0 1 1 0 ... ,
	//   the 4th-from-lowest bit of x has the 8-cycle ... 0 0 0 1 1 1 1 0 ... ,
	//    ...
	//
	// and if x_initial is 3 mod 4 then
	//
	//   the          lowest bit of x is always 1,
	//   the  next-to-lowest bit of x is always 1,
	//   the 2nd-from-lowest bit of x alternates      ... 0 1 0 1 0 1 0 1 ... ,
	//   the 3rd-from-lowest bit of x 4-cycles        ... 0 0 1 1 0 0 1 1 ... ,
	//   the 4th-from-lowest bit of x has the 8-cycle ... 0 0 1 1 1 1 0 0 ... ,
	//    ...
	//
	// The generator's potency (min. s>=0 with (69069-1)^s = 0 mod 2^32) is
	// 16, which seems to be alright by p. 25, Sec. 3.2.1.3 of Knuth.  It
	// also does well in the dimension 2..5 spectral tests, but it could be
	// better in dimension 6 (Line 15, Table 1, p. 106, Sec. 3.3.4, Knuth).
	//
	// Note that the random number user does not see the values generated
	// here directly since reloadMT() will always munge them first, so maybe
	// none of all of this matters.  In fact, the seed values made here could
	// even be extra-special desirable if the Mersenne Twister theory says
	// so-- that's why the only change I made is to restrict to odd seeds.
	//

	uint32 seed = uint32(seed_);

	register uint32 x = (seed | 1U) & 0xFFFFFFFFU, *s = state_m;
	register int    j;

    for(left_m=0, *s++=x, j=N; --j;
        *s++ = (x*=69069U) & 0xFFFFFFFFU);
}


void RandomMT::next()
{
    if(--left_m < 0)
    {
        random_m = dword(reloadMT());
	}
	else
	{
	    uint32 y;
		y  = *next_m++;
		y ^= (y >> 11);
		y ^= (y <<  7) & 0x9D2C5680U;
		y ^= (y << 15) & 0xEFC60000U;
		random_m = dword(y ^ (y >> 18));
	}
}




/// queries ----------------------------------------------------------------------------------------
dword RandomMT::getDword() const
{
	return random_m;
}


float RandomMT::getFloat() const
{
	static const float DIV = float( 1.0 / 4294967296.0 );
	return float(udword(random_m)) * DIV;
}




/// implementation ---------------------------------------------------------------------------------
uint32 RandomMT::reloadMT()
{
	register uint32 *p0=state_m, *p2=state_m+2, *pM=state_m+M, s0, s1;
	register int    j;

	if(left_m < -1)
		setSeed( dword(4357U) );

	left_m=N-1, next_m=state_m+1;

	for(s0=state_m[0], s1=state_m[1], j=N-M+1; --j; s0=s1, s1=*p2++)
		*p0++ = *pM++ ^ (mixBits(s0, s1) >> 1) ^ (loBit(s1) ? K : 0U);

	for(pM=state_m, j=M; --j; s0=s1, s1=*p2++)
		*p0++ = *pM++ ^ (mixBits(s0, s1) >> 1) ^ (loBit(s1) ? K : 0U);

	s1=state_m[0], *p0 = *pM ^ (mixBits(s0, s1) >> 1) ^ (loBit(s1) ? K : 0U);
	s1 ^= (s1 >> 11);
	s1 ^= (s1 <<  7) & 0x9D2C5680U;
	s1 ^= (s1 << 15) & 0xEFC60000U;
	return(s1 ^ (s1 >> 18));
}
