/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef RandomMT_h
#define RandomMT_h


#include "Primitives.h"
#include "Random.h"




#include "hxa7241_utility.h"
namespace hxa7241 { namespace utility {
	using namespace hxa7241::primitives;
	typedef udword uint32;


/**
 * mersenne twister MT19937 random number generator.<br/><br/>
 *
 * gives a signed dword in the range [MIN_VALUE, MAX_VALUE], and a float in the
 * range [0, 1).<br/><br/>
 *
 * adapted from code by <cite>shawn cokus</cite>, which was adapted from
 * <cite>takuji nishimura, http://www.math.keio.ac.jp/~matumoto/emt.html</cite>
 */

// "The code is for the uniform integer in 0..(2^32 - 1) case, and runs about
// 2.15 times faster than the code on the web page. (I'm on a DEC Alpha running
// OSF/1 and I am using the GNU C compiler at optimization level -O3.) With
// the faster code, I can generate about 12.5 million random numbers per
// second on the machine here. The code is not completely ANSI: I have a
// tendency to use `//'-style comments, and I do use the `inline' keyword of
// GNU C once. If you change the comments to `/* ... */'-style and remove the
// `inline' keyword (or, better yet, change the small function tagged with it
// to a macro), then I'm pretty sure the code would be ANSI (except for some
// assumptions about the representation of unsigned numbers due to the &'s,
// |'s, and ^'s). Feel free to do whatever you'd like with this implementation
// -- and thanks for putting your information on the web."

// This is the ``Mersenne Twister'' random number generator MT19937, which
// generates pseudorandom integers uniformly distributed in 0..(2^32 - 1)
// starting from any odd seed in 0..(2^32 - 1).  This version is a recode
// by Shawn Cokus (Cokus@math.washington.edu) on March 8, 1998 of a version by
// Takuji Nishimura (who had suggestions from Topher Cooper and Marc Rieffel in
// July-August 1997).
//
// Effectiveness of the recoding (on Goedel2.math.washington.edu, a DEC Alpha
// running OSF/1) using GCC -O3 as a compiler: before recoding: 51.6 sec. to
// generate 300 million random numbers; after recoding: 24.0 sec. for the same
// (i.e., 46.5% of original time), so speed is now about 12.5 million random
// number generations per second on this machine.
//
// According to the URL <http://www.math.keio.ac.jp/~matumoto/emt.html>
// (and paraphrasing a bit in places), the Mersenne Twister is ``designed
// with consideration of the flaws of various existing generators,'' has
// a period of 2^19937 - 1, gives a sequence that is 623-dimensionally
// equidistributed, and ``has passed many stringent tests, including the
// die-hard test of G. Marsaglia and the load test of P. Hellekalek and
// S. Wegenkittl.''  It is efficient in memory usage (typically using 2506
// to 5012 bytes of static data, depending on data type sizes, and the code
// is quite short as well).  It generates random numbers in batches of 624
// at a time, so the caching and pipelining of modern systems is exploited.
// It is also divide- and mod-free.
//
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as published by
// the Free Software Foundation (either version 2 of the License or, at your
// option, any later version).  This library is distributed in the hope that
// it will be useful, but WITHOUT ANY WARRANTY, without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
// the GNU Library General Public License for more details.  You should have
// received a copy of the GNU Library General Public License along with this
// library; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA 02111-1307, USA.
//
// The code as Shawn received it included the following notice:
//
//   Copyright (C) 1997 Makoto Matsumoto and Takuji Nishimura.  When
//   you use this, send an e-mail to <matumoto@math.keio.ac.jp> with
//   an appropriate reference to your work.
//
// It would be nice to CC: <Cokus@math.washington.edu> when you write.
//
class RandomMT
	: public Random
{
public:
/// standard object services -----------------------------------------------------------------------
	explicit RandomMT( dword seed =0 );

	virtual ~RandomMT();
	         RandomMT( const RandomMT& );
	RandomMT& operator=( const RandomMT& );


/// commands ---------------------------------------------------------------------------------------
	virtual void  setSeed( dword );
	virtual void  next();


/// queries ----------------------------------------------------------------------------------------
	virtual dword getDword() const;
	virtual float getFloat() const;


/// implementation ---------------------------------------------------------------------------------
private:
	         uint32 reloadMT();


/// fields -----------------------------------------------------------------------------------------
private:
	#define N              (624)                 // length of state vector
	#define M              (397)                 // a period parameter
	#define K              (0x9908B0DFU)         // a magic constant

	#define hiBit(u)       ((u) & 0x80000000U)   // mask all but highest   bit of u
	#define loBit(u)       ((u) & 0x00000001U)   // mask all but lowest    bit of u
	#define loBits(u)      ((u) & 0x7FFFFFFFU)   // mask     the highest   bit of u
	#define mixBits(u, v)  (hiBit(u)|loBits(v))  // move hi bit of u to hi bit of v

	//uint32  state_m[N+1];   // state vector + 1 extra to not violate ANSI C
	uint32* state_m;        // (moved to heap)
	uint32* next_m;         // next random value is computed from here
	int     left_m;         // can *next++ this many times before reloading

	dword   random_m;
};


}}




#endif//RandomMT_h
