/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef SerializerIn_h
#define SerializerIn_h


#include "Primitives.h"
#include "SerializerOut.h"
#include "Exception.h"
#include <iostream>
#include <string>
#include <deque>




#include "utility.h"
namespace hxa7241 { namespace utility {
	using namespace hxa7241::primitives;
	using std::istream;
	using std::string;
	using std::deque;


/**
 * abstract base of an inward serializer class.<br/><br/>
 *
 * derivatives need to provide a way of supplying an object and stream, and
 * should override operator>>, read header footer pre object and post object
 * methods.
 *
 * @see SerializerOut
 */
class SerializerIn
{
/// standard object services -----------------------------------------------------------------------
protected:
	         SerializerIn( istream&, const char* fileName );

	virtual ~SerializerIn();
private:
	         SerializerIn( const SerializerIn& );
	SerializerIn& operator=( const SerializerIn& );


/// commands ---------------------------------------------------------------------------------------
protected:
	template<class TYPE>
	        bool  read( TYPE& object );

public:
	template<class TYPE>
	        bool  read( const char* name, TYPE& object );
	template<class TYPE>
	        bool  read( const char* name, deque<TYPE>& objects );
	template<class TYPE, class TYPE2>
	        bool  read( const char* name, deque<TYPE>& objects, TYPE2& context );
	template<class TYPE, class TYPE2>
	        bool  read( const char* name, TYPE& object, TYPE2& context );
	template<class TYPE>
	        bool  readElement( const char* collectionName, const char* name, TYPE& object );
	template<class TYPE, class TYPE2>
	        bool  readElement( const char* collectionName, const char* name, TYPE& object, TYPE2& context );
//	template<class TYPE>
//	        bool  readp( const char* name, TYPE* pObject );
//	template<class TYPE>
//	        bool  readpa( const char* name, TYPE*& pObject );
//	        bool  readca( const char* name, char*& pChars );


/// queries ----------------------------------------------------------------------------------------
public:


/// implementation ---------------------------------------------------------------------------------
protected:
	virtual SerializerIn& operator>>( void*& );
	virtual SerializerIn& operator>>( bool& );
	virtual SerializerIn& operator>>( hxa7241::utility::byte& );
	virtual SerializerIn& operator>>( ubyte& );
	virtual SerializerIn& operator>>( word& );
	virtual SerializerIn& operator>>( uword& );
	virtual SerializerIn& operator>>( dword& );
	virtual SerializerIn& operator>>( udword& );
	virtual SerializerIn& operator>>( long& );
	virtual SerializerIn& operator>>( unsigned long& );
	virtual SerializerIn& operator>>( float& );
	virtual SerializerIn& operator>>( double& );
	virtual SerializerIn& operator>>( char& );
	virtual SerializerIn& operator>>( char* );
	virtual SerializerIn& operator>>( string& );
	template<class TYPE>
	        SerializerIn& input( const char* collectionName, deque<TYPE>& );
	template<class TYPE, class TYPE2>
	        SerializerIn& input( const char* collectionName, deque<TYPE>&, TYPE2& context );

	virtual void  readHeader( const char* name );
	virtual void  readFooter( const char* name );

	virtual bool  readPreObject ( const char* name );
	virtual void  readPostObject( const char* name );
	virtual bool  readPreElement( const char* collectionName, const char* name, bool& isCollectionEnd );


/// fields -----------------------------------------------------------------------------------------
protected:
	istream& in_m;
	string   buffer_m;
	string   fileName_m;

	std::ios_base::iostate  previousExceptionFlags_m;


private:
	static const char CLASS_NAME[];
	static const char EXCEPTION_MESSAGE[];
};








/// TEMPLATES ///


/// commands ---------------------------------------------------------------------------------------
template<class TYPE>
bool SerializerIn::read
(
	TYPE& object
)
{
	try
	{
		readHeader( SerializerOut::typeName<TYPE>() );

		const bool content = read( SerializerOut::typeName<TYPE>(), object );

		if( content )
		{
			readFooter( SerializerOut::typeName<TYPE>() );
		}

		return content;
	}
	catch( std::ios_base::failure e )
	{
		string msg( EXCEPTION_MESSAGE );
		throw hxa7241::utility::Exception( ((msg += " - ") += e.what()).c_str(), CLASS_NAME, 0 );
	}
}




template<class TYPE>
bool SerializerIn::read
(
	const char*const name,
	TYPE&            object
)
{
	const bool isContent = readPreObject( name );
	if( isContent )
	{
		*this >> object;
		readPostObject( name );
	}

	return isContent;
}


template<class TYPE>
bool SerializerIn::read
(
	const char*const name,
	deque<TYPE>&     objects
)
{
	const bool isContent = readPreObject( name );
	if( isContent )
	{
		input( name, objects );
	}

	return isContent;
}


template<class TYPE, class TYPE2>
bool SerializerIn::read
(
	const char*const name,
	deque<TYPE>&     objects,
	TYPE2&           context
)
{
	const bool isContent = readPreObject( name );
	if( isContent )
	{
		input( name, objects, context );
	}

	return isContent;
}


template<class TYPE, class TYPE2>
bool SerializerIn::read
(
	const char*const name,
	TYPE&            object,
	TYPE2&           context
)
{
	const bool isContent = readPreObject( name );
	if( isContent )
	{
		::input( *this, object, context );
		readPostObject( name );
	}

	return isContent;
}


template<class TYPE>
bool SerializerIn::readElement
(
	const char*const collectionName,
	const char*const name,
	TYPE&            object
)
{
	bool isCollectionEnd = false;
	const bool isContent = readPreElement( collectionName, name, isCollectionEnd );
	if( isContent )
	{
		*this >> object;
		readPostObject( name );
	}

	return isCollectionEnd;
}


template<class TYPE, class TYPE2>
bool SerializerIn::readElement
(
	const char*const collectionName,
	const char*const name,
	TYPE&            object,
	TYPE2&           context
)
{
	bool isCollectionEnd = false;
	const bool isContent = readPreElement( collectionName, name, isCollectionEnd );
	if( isContent )
	{
		::input( *this, object, context );
		readPostObject( name );
	}

	return isCollectionEnd;
}


//template<class TYPE>
//inline
//bool SerializerIn::readp
//(
//	const char*const name,
//	TYPE*            pObject
//)
//{
//	return read( name, *pObject );
//}
//
//
//template<class TYPE>
//bool SerializerIn::readpa
//(
//	const char*const name,
//	TYPE*&           pObject
//)
//{
//	delete pObject;
//	pObject = new TYPE();
//
//	return read( name, *pObject );
//}
//
//
//bool SerializerIn::readca
//(
//	const char*const name,
//	char*&           pChars
//)
//{
//	string s;
//	const bool content = read( name, s );
//
//	delete[] pChars;
//	pChars = new char[ s.length() + 1 ];
//
//	strcpy( pChars, s.c_str() );
//
//	return content;
//}




/// implementation ---------------------------------------------------------------------------------
template<class TYPE>
SerializerIn& SerializerIn::input
(
	const char*const collectionName,
	deque<TYPE>&     collection
)
{
	/// member template functions cant be virtual, so choose implementation manually
	static const string XML( "XmlSerializerIn" );

	const string typeName( typeid(*this).name() );

	if( typeName.find( XML ) != typeName.npos )
	{
		do
		{
			collection.resize( collection.size() + 1 );
		}
		while( !readElement( collectionName, SerializerOut::COLLECTION_ELEMENT_NAME(), collection.back() ) );

		collection.pop_back();
	}
	else
	{
		/// read size
		dword size = 0;
		read( SerializerOut::COLLECTION_SIZE_NAME(), size );

		/// read elements
		for( dword i = 0;  i < size;  ++i )
		{
			collection.resize( collection.size() + 1 );

			read( SerializerOut::COLLECTION_ELEMENT_NAME(), collection.back() );
		}
	}

	return *this;
}


template<class TYPE, class TYPE2>
SerializerIn& SerializerIn::input
(
	const char*const collectionName,
	deque<TYPE>&     collection,
	TYPE2&           context
)
{
	/// member template functions cant be virtual, so choose implementation manually
	static const string XML( "XmlSerializerIn" );

	const string typeName( typeid(*this).name() );

	if( typeName.find( XML ) != typeName.npos )
	{
		do
		{
			collection.resize( collection.size() + 1 );
		}
		while( !readElement( collectionName, SerializerOut::COLLECTION_ELEMENT_NAME(), collection.back(), context ) );

		collection.pop_back();
	}
	else
	{
		/// read size
		dword size = 0;
		read( SerializerOut::COLLECTION_SIZE_NAME(), size );

		/// read elements
		for( dword i = 0;  i < size;  ++i )
		{
			collection.resize( collection.size() + 1 );

			read( SerializerOut::COLLECTION_ELEMENT_NAME(), collection.back(), context );
		}
	}

	return *this;
}


}}




#endif//SerializerIn_h
