/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef SerializerOut_h
#define SerializerOut_h


#include "Primitives.h"
#include "Exception.h"
#include <iostream>
#include <string>
#include <deque>




#include "utility.h"
namespace hxa7241 { namespace utility {
	using namespace hxa7241::primitives;
	using std::ostream;
	using std::string;
	using std::deque;
	typedef char*       charptr;
	typedef const char* constcharptr;


/**
 * abstract base of an outward serializer class.<br/><br/>
 *
 * derivatives need to provide a way of supplying an object and stream, and
 * should override operator<<, write header footer pre object and post object
 * methods.
 *
 * @see SerializerIn
 */
class SerializerOut
{
/// standard object services -----------------------------------------------------------------------
protected:
	explicit SerializerOut( ostream& );

	virtual ~SerializerOut();
private:
	         SerializerOut( const SerializerOut& );
	SerializerOut& operator=( const SerializerOut& );


/// commands ---------------------------------------------------------------------------------------
protected:
	template<class TYPE>
	        void  write( const TYPE& object );

public:
	template<class TYPE>
	        void  write ( const char* name, const TYPE& object,  const char* comment =0 );
	template<class TYPE, class TYPE2>
	        void  write ( const char* name, const deque<TYPE>& objects, const TYPE2& context, const char* comment );
	template<class TYPE, class TYPE2>
	        void  write ( const char* name, const TYPE& object, const TYPE2& context, const char* comment );
//	template<class TYPE>
//	        void  writep( const char* name, const TYPE* pObject, const char* comment =0 );
//	template<class TYPE>
//	        void  writea( const char* name, const TYPE* pObjects, int arrayLength );


/// queries ----------------------------------------------------------------------------------------
public:
	template<class TYPE>
	static  const char* typeName();
	template<class TYPE>
	static  const char* typeDescription();


/// implementation ---------------------------------------------------------------------------------
protected:
	virtual SerializerOut& operator<<( const void* );
	virtual SerializerOut& operator<<( bool );
	virtual SerializerOut& operator<<( hxa7241::utility::byte );
	virtual SerializerOut& operator<<( ubyte );
	virtual SerializerOut& operator<<( word );
	virtual SerializerOut& operator<<( uword );
	virtual SerializerOut& operator<<( dword );
	virtual SerializerOut& operator<<( udword );
	virtual SerializerOut& operator<<( long );
	virtual SerializerOut& operator<<( unsigned long );
	virtual SerializerOut& operator<<( float );
	virtual SerializerOut& operator<<( double );
	virtual SerializerOut& operator<<( char );
	virtual SerializerOut& operator<<( const char* );
	virtual SerializerOut& operator<<( const string& );
	template<class TYPE>
	        SerializerOut& operator<<( const deque<TYPE>& );
	template<class TYPE, class TYPE2>
	        SerializerOut& output( const deque<TYPE>&, const TYPE2& context );

	virtual void  writeHeader( const char* name, const char* typeName, const char* comment );
	virtual void  writeFooter( const char* name, const char* typeName, const char* comment );

	virtual void  writePreObject ( const char* name, const char* typeName, const char* comment );
	virtual void  writePostObject( const char* name, const char* typeName, const char* comment );


/// fields -----------------------------------------------------------------------------------------
protected:
	ostream& out_m;
	string   buffer_m;

	std::ios_base::iostate  previousExceptionFlags_m;


private:
	static const char CLASS_NAME[];
	static const char EXCEPTION_MESSAGE[];
public:
	static const char* COLLECTION_ELEMENT_NAME();
	static const char* COLLECTION_SIZE_NAME();
};








/// TEMPLATES ///


/// commands ---------------------------------------------------------------------------------------
template<class TYPE>
void SerializerOut::write
(
	const TYPE& object
)
{
	try
	{
		writeHeader( 0, typeName<TYPE>(), typeDescription<TYPE>() );

		write( typeName<TYPE>(), object, typeDescription<TYPE>() );

		writeFooter( 0, typeName<TYPE>(), typeDescription<TYPE>() );
	}
	catch( std::ios_base::failure e )
	{
		string msg( EXCEPTION_MESSAGE );
		throw hxa7241::utility::Exception( ((msg += " - ") += e.what()).c_str(), CLASS_NAME, 0 );
	}
}




template<class TYPE>
void SerializerOut::write
(
	const char*const name,
	const TYPE&      object,
	const char*const comment
)
{
	writePreObject( name, typeName<TYPE>(), comment );

	*this << object;

	writePostObject( name, typeName<TYPE>(), comment );
}


template<class TYPE, class TYPE2>
void SerializerOut::write
(
	const char*const   name,
	const deque<TYPE>& objects,
	const TYPE2&       context,
	const char*const   comment
)
{
	writePreObject( name, typeName<TYPE>(), comment );

	output( objects, context );

	writePostObject( name, typeName<TYPE>(), comment );
}


template<class TYPE, class TYPE2>
void SerializerOut::write
(
	const char*const name,
	const TYPE&      object,
	const TYPE2&     context,
	const char*const comment
)
{
	writePreObject( name, typeName<TYPE>(), comment );

	::output( *this, object, context );

	writePostObject( name, typeName<TYPE>(), comment );
}


//template<class TYPE>
//inline
//void SerializerOut::writep
//(
//	const char*const name,
//	const TYPE*      pObject,
//	const char*const comment
//)
//{
//	write( name, *pObject, comment );
//}


//template<class TYPE>
//void SerializerOut::writea
//(
//	const char*const name,
//	const TYPE*      pObjects,
//	const int        arrayLength
//)
//{
//	writeTagOpen( name );
//	for( int i = 0;  i < arrayLength;  ++i )
//	{
//		write( ARRAY_ELEMENT_NAME, pObjects[i] );
//	}
//	writeTagClose( name );
//}




/// queries ----------------------------------------------------------------------------------------
template<class TYPE>
inline
const char* SerializerOut::typeName()
{
	return TYPE::CLASS_NAME();
}


template<class TYPE>
inline
const char* SerializerOut::typeDescription()
{
	return TYPE::CLASS_DESCRIPTION();
}




/// specialisations
template<>
inline
const char* SerializerOut::typeName<bool>()
{
	return "boolean";
}


template<>
inline
const char* SerializerOut::typeName<char>()
{
	return "character";
}


template<>
inline
const char* SerializerOut::typeName<byte>()
{
	return "bytesigned";
}


template<>
inline
const char* SerializerOut::typeName<ubyte>()
{
	return "byteunsigned";
}


template<>
inline
const char* SerializerOut::typeName<word>()
{
	return "wordsigned";
}


template<>
inline
const char* SerializerOut::typeName<uword>()
{
	return "wordunsigned";
}


template<>
inline
const char* SerializerOut::typeName<dword>()
{
	return "dwordsigned";
}


template<>
inline
const char* SerializerOut::typeName<udword>()
{
	return "dwordunsigned";
}


template<>
inline
const char* SerializerOut::typeName<long>()
{
	return "dwordsigned";
}


template<>
inline
const char* SerializerOut::typeName<unsigned long>()
{
	return "dwordunsigned";
}


template<>
inline
const char* SerializerOut::typeName<float>()
{
	return "floatingpointsingle";
}


template<>
inline
const char* SerializerOut::typeName<double>()
{
	return "floatingpointdouble";
}


//template<>
//inline
//const char* SerializerOut::typeName<char*>()
//{
//	return "characters";
//}


template<>
inline
const char* SerializerOut::typeName<const char*>()
{
	return "characters";
}


template<>
inline
const char* SerializerOut::typeName<std::string>()
{
	return "std::string";
}


///	need to be provided locally, since the compiler doesnt like what i believe
///	is the correct partial specialisation definition
///	very irritating
///
//template<class TYPE>
//inline
//const char* SerializerOut::typeName< std::deque<TYPE> >()
//{
//	return "std::deque";
//}
//
//
//template<class TYPE>
//inline
//const char* SerializerOut::typeDescription< std::deque<TYPE> >()
//{
//	return "a standard deque collection";
//}




/// implementation ---------------------------------------------------------------------------------
template<class TYPE>
SerializerOut& SerializerOut::operator<<( const deque<TYPE>& collection )
{
	/// member template functions cant be virtual, so choose implementation manually
	static const string XML( "XmlSerializerOut" );

	const string typeName( typeid(*this).name() );

	if( typeName.find( XML ) == typeName.npos )
	{
		/// write size
		write( COLLECTION_SIZE_NAME(), dword(collection.size()) );
	}

	const deque<TYPE>::const_iterator end( collection.end() );
	for( deque<TYPE>::const_iterator it( collection.begin() );  it != end;  ++it )
	{
		write( COLLECTION_ELEMENT_NAME(), *it );
	}
//	const int length = collection.size();
//	for( int i = 0;  i < length;  ++i )
//	{
//		write( COLLECTION_ELEMENT_NAME(), collection[i] );
//	}

	return *this;
}


template<class TYPE, class TYPE2>
SerializerOut& SerializerOut::output
(
	const deque<TYPE>& collection,
	const TYPE2&       context
)
{
	/// member template functions cant be virtual, so choose implementation manually
	static const string XML( "XmlSerializerOut" );

	const string typeName( typeid(*this).name() );

	if( typeName.find( XML ) == typeName.npos )
	{
		/// write size
		write( COLLECTION_SIZE_NAME(), dword(collection.size()) );
	}

	const deque<TYPE>::const_iterator end( collection.end() );
	for( deque<TYPE>::const_iterator it( collection.begin() );  it != end;  ++it )
	{
		write( COLLECTION_ELEMENT_NAME(), *it, context, 0 );
	}
//	const int length = collection.size();
//	for( int i = 0;  i < length;  ++i )
//	{
//		write( COLLECTION_ELEMENT_NAME(), collection[i], context, 0 );
//	}

	return *this;
}


}}




#endif//SerializerOut_h
