/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "UriFile.h"   /// own header is included last


using namespace hxa7241::utility;




/// statics
const char UriFile::SCHEME[]         = "file";
const char UriFile::SCHEME_SEP[]     = "://";
const char UriFile::SCHEME_AND_SEP[] = "file://";
const char UriFile::SEPARATORS[]     = "/\\";




/// standard object services -----------------------------------------------------------------------
UriFile::UriFile()
 :	string_m (),
	mode_m   ( DEFAULT_MODE ),
	pathPos_m( 0 ),
	namePos_m( 0 )
{
}


UriFile::UriFile( const string& str )
 :	string_m ( str ),
	mode_m   ( DEFAULT_MODE )
{
	UriFile::validate();
}


UriFile::UriFile( const char chs[] )
 :	string_m ( chs ),
	mode_m   ( DEFAULT_MODE )
{
	UriFile::validate();
}


UriFile::~UriFile()
{
}


UriFile::UriFile( const UriFile& other )
{
	UriFile::operator=( other );
}


UriFile& UriFile::operator=( const UriFile& other )
{
	if( &other != this )
	{
		string_m  = other.string_m;
		mode_m    = other.mode_m;
		pathPos_m = other.pathPos_m;
		namePos_m = other.namePos_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
void UriFile::setMode
(
	const EMode mode
)
{
	if( mode_m != mode )
	{
		mode_m = mode;
		modulate( mode_m, string_m );
	}
}




/// queries ----------------------------------------------------------------------------------------
const char* UriFile::getScheme()
{
	return SCHEME;
}


const char* UriFile::getSchemeWithSeparator()
{
	return SCHEME_AND_SEP;
}


const char* UriFile::getSchemeSeparator()
{
	return SCHEME_SEP;
}


bool UriFile::isValid() const
{
	return !string_m.empty();
}


void UriFile::getString
(
	string& str
) const
{
	str = string_m;
}


void UriFile::getPath
(
	string& str
) const
{
	str = string_m.substr( pathPos_m, namePos_m - pathPos_m );
}


void UriFile::getName
(
	string& str
) const
{
	str = string_m.substr( namePos_m, string_m.size() - namePos_m );
}


void UriFile::getPathname
(
	string& str
) const
{
	str = string_m.substr( pathPos_m, string_m.size() - pathPos_m );
}


UriFile::EMode UriFile::getMode() const
{
	return mode_m;
}




/// implementation ---------------------------------------------------------------------------------
void UriFile::validate()
{
	/// match SCHEME_AND_SEP at pos 0
	const bool isValid = (string_m.rfind( string(SCHEME_AND_SEP), 1 ) == 0);

	if( isValid )
	{
		pathPos_m = (sizeof(SCHEME_AND_SEP) / sizeof(SCHEME_AND_SEP[0])) - 1;

		/// find last separator
		const dword lastSeparatorPos = string_m.find_last_of( SEPARATORS );
		namePos_m = lastSeparatorPos + 1;

		modulate( mode_m, string_m );
	}
	else
	{
		string_m.resize( 0 );

		pathPos_m = 0;
		namePos_m = 0;
	}
}


void UriFile::modulate
(
	const EMode mode,
	string&     str
) const
{
	const char notWanted = SEPARATORS[ (int(mode) + 1) & 1 ];
	const char wanted    = SEPARATORS[ int(mode) ];

	const dword start = (sizeof(SCHEME_AND_SEP) / sizeof(SCHEME_AND_SEP[0])) - 1;
	const dword end   = str.size();
	for( dword i = start;  i < end;  ++i )
	{
		if( str[i] == notWanted )
		{
			str[i] = wanted;
		}
	}
}
