/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#ifndef Vector_h
#define Vector_h


#include "Primitives.h"
#include "Exception.h"
#include <vector>
#include "hxa7241vector.h"




#include "hxa7241_utility.h"
namespace hxa7241 { namespace utility {
	using namespace hxa7241::primitives;


/**
 * a vector with more reserve control, but only a partly similar interface.
 * <br/><br/>
 *
 * every time a std::vector is resized, elements are constructed or destructed.
 * this vector allows another 'used size' to be set and altered, without (always)
 * causing element con/de-struction.<br/><br/>
 *
 * also adds some further control over reserve size.
 *
 * @see hxa7241vector.h
 *
 * @implementation
 * made with std::vector.
 *
 * @invariants
 * usedSize_m >= 0
 * usedSize_m <= vector_m.size()
 * vector_m.size() <= DWORD_MAX
 */
template<class TYPE>
class Vector
{
/// standard object services -----------------------------------------------------------------------
public:
	         Vector();
	explicit Vector( dword size );

	        ~Vector();
	         Vector( const Vector& );
	Vector& operator=( const Vector& );


/// commands ---------------------------------------------------------------------------------------
public:
	        void  setVector( const std::vector<TYPE>& );

	        void  setUsedSize( dword );
	        void  resize( dword );
	        void  reserve( dword );

	        void  push_back( const TYPE& );
	        void  pop_back();
	        void  pushBackUsed( const TYPE& );
	        void  popBackUsed();

	        void  rezero();
	        void  unreserve();
	        void  resizeReserve( dword size );
	        void  reserveExactClear( dword size );
	        void  assignUnreserve( const Vector<TYPE>& other );

	        TYPE& operator[]( dword );
	        TYPE* getFront();
	        TYPE* getEnd();
	        TYPE* getEndUsed();


/// queries ----------------------------------------------------------------------------------------
public:
	        const std::vector<TYPE>& getVector() const;

	        dword       getUsedSize()            const;
	        dword       size()                   const;
	        dword       capacity()               const;
	        bool        empty()                  const;

	        const TYPE& operator[]( dword )      const;
	        const TYPE* getFront()               const;
	        const TYPE* getEnd()                 const;
	        const TYPE* getEndUsed()             const;


/// implementation ---------------------------------------------------------------------------------
protected:
	        void  conditionUsedSize();


/// fields -----------------------------------------------------------------------------------------
private:
	std::vector<TYPE> vector_m;
	dword             usedSize_m;
};








/// standard object services -----------------------------------------------------------------------
template<class TYPE>
Vector<TYPE>::Vector()
 :	vector_m  (),
	usedSize_m( 0 )
{
}


template<class TYPE>
Vector<TYPE>::Vector
(
	const dword size
)
 :	vector_m  ( size >= 0 ? size : 0 ),
	usedSize_m( 0 )
{
}


template<class TYPE>
inline
Vector<TYPE>::~Vector()
{
}


template<class TYPE>
Vector<TYPE>::Vector
(
	const Vector<TYPE>& other
)
 :	vector_m  ( other.vector_m ),
	usedSize_m( other.usedSize_m )
{
}


template<class TYPE>
Vector<TYPE>& Vector<TYPE>::operator=
(
	const Vector<TYPE>& other
)
{
	if( &other != this )
	{
		vector_m   = other.vector_m;
		usedSize_m = other.usedSize_m;
	}

	return *this;
}




/// commands ---------------------------------------------------------------------------------------
template<class TYPE>
void Vector<TYPE>::setVector
(
	const std::vector<TYPE>& v
)
{
	vector_m   = v;
	usedSize_m = v.size();
}


template<class TYPE>
void Vector<TYPE>::setUsedSize
(
	dword size
)
{
	if( size < 0 )
	{
		size = 0;
	}

	if( size > dword(vector_m.size()) )
	{
		vector_m.resize( size );
	}

	usedSize_m = size;
}


template<class TYPE>
void Vector<TYPE>::resize
(
	dword size
)
{
	if( size < 0 )
	{
		size = 0;
	}

	vector_m.resize( size );

	conditionUsedSize();
}


template<class TYPE>
void Vector<TYPE>::reserve
(
	dword size
)
{
	if( size < 0 )
	{
		size = 0;
	}

	vector_m.reserve( size );
}


template<class TYPE>
void Vector<TYPE>::push_back
(
	const TYPE& value
)
{
	const udword size = vector_m.size();
	if( size >= udword(DWORD_MAX) )
	{
		throw Exception( "array length overflow", "hxa7241::utility::Vector<>", "push_back" );
	}

	vector_m.push_back( value );

	usedSize_m = dword(vector_m.size());
}


template<class TYPE>
void Vector<TYPE>::pop_back()
{
	vector_m.pop_back();

	conditionUsedSize();
}


template<class TYPE>
void Vector<TYPE>::pushBackUsed
(
	const TYPE& value
)
{
	if( usedSize_m == DWORD_MAX )
	{
		throw Exception( "array length overflow", "hxa7241::utility::Vector<>", "pushBackUsed" );
	}

	if( usedSize_m < dword(vector_m.size()) )
	{
		vector_m[usedSize_m] = value;
	}
	else
	{
		vector_m.push_back( value );
	}

	++usedSize_m;
}


template<class TYPE>
void Vector<TYPE>::popBackUsed()
{
	if( usedSize_m > 0 )
	{
		--usedSize_m;
	}
}


template<class TYPE>
void Vector<TYPE>::rezero()
{
	rezero( vector_m );

	usedSize_m = 0;
}


template<class TYPE>
void Vector<TYPE>::unreserve()
{
	unreserve( vector_m );
}


template<class TYPE>
void Vector<TYPE>::resizeReserve
(
	const dword size
)
{
	resizeReserve( vector_m, size );

	conditionUsedSize();
}


template<class TYPE>
void Vector<TYPE>::reserveExactClear
(
	const dword size
)
{
	reserveExactClear( vector_m, size );

	usedSize_m = 0;
}


template<class TYPE>
void Vector<TYPE>::assignUnreserve
(
	const Vector<TYPE>& other
)
{
	assignUnreserve( other.vector_m, vector_m );

	usedSize_m = other.usedSize_m;
}


template<class TYPE>
inline
TYPE& Vector<TYPE>::operator[]
(
	dword index
)
{
	return vector_m[ index ];
}


template<class TYPE>
inline
TYPE* Vector<TYPE>::getFront()
{
	return &( vector_m.front() );
}


template<class TYPE>
inline
TYPE* Vector<TYPE>::getEnd()
{
	return &( vector_m.back() ) + 1;
}


template<class TYPE>
inline
TYPE* Vector<TYPE>::getEndUsed()
{
	return &( vector_m[usedSize_m] );
}




/// queries ----------------------------------------------------------------------------------------
template<class TYPE>
inline
const std::vector<TYPE>& Vector<TYPE>::getVector() const
{
	return vector_m;
}


template<class TYPE>
inline
dword Vector<TYPE>::getUsedSize() const
{
	return usedSize_m;
}


template<class TYPE>
inline
dword Vector<TYPE>::size() const
{
	return dword(vector_m.size());
}


template<class TYPE>
inline
dword Vector<TYPE>::capacity() const
{
	return dword(vector_m.capacity());
}


template<class TYPE>
inline
bool Vector<TYPE>::empty() const
{
	return vector_m.empty();
}


template<class TYPE>
inline
const TYPE& Vector<TYPE>::operator[]
(
	dword index
) const
{
	return vector_m[ index ];
}


template<class TYPE>
inline
const TYPE* Vector<TYPE>::getFront() const
{
	return &( vector_m.front() );
}


template<class TYPE>
inline
const TYPE* Vector<TYPE>::getEnd() const
{
	return &( vector_m.back() ) + 1;
}


template<class TYPE>
inline
const TYPE* Vector<TYPE>::getEndUsed() const
{
	return &( vector_m[usedSize_m] );
}




/// implementation ---------------------------------------------------------------------------------
template<class TYPE>
void Vector<TYPE>::conditionUsedSize()
{
	const dword size = dword(vector_m.size());
	if( usedSize_m > size )
	{
		usedSize_m = size;
	}
}


}}




#endif//Vector_h
