/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "XmlSerializerOut.h"
#include "Exception.h"
#include <sstream>
#include <cctype>

#include "XmlSerializerIn.h"   /// own header is included last


using namespace hxa7241::utility;




/// statics
const char XmlSerializerIn::EXCEPTION_MESSAGE1[] = "xml serialization version not supported";
const char XmlSerializerIn::EXCEPTION_MESSAGE2[] = "xml structure doesnt match class";
const char XmlSerializerIn::EXCEPTION_MESSAGE3[] = "data not between open and close tags";
const char XmlSerializerIn::EXCEPTION_MESSAGE4[] = "no CDATA for string item";

const char XmlSerializerIn::CLASS_NAME[] = "hxa7241::utility::XmlSerializerIn";




/// standard object services -----------------------------------------------------------------------
XmlSerializerIn::XmlSerializerIn
(
	istream& in,
	const char* fileName
)
 :	SerializerIn( in, fileName ),
	nodeBuffer_m(),
	lineNumber_m( 1 ),
	charPos_m( 1 ),
	lastCharWasNewline_m( false )
{
}


XmlSerializerIn::~XmlSerializerIn()
{
}




/// implementation ---------------------------------------------------------------------------------
XmlSerializerIn& XmlSerializerIn::operator>>( void*& a )
{
	in_m >> a;

	return *this;
}


XmlSerializerIn& XmlSerializerIn::operator>>( bool& a )
{
	in_m >> a;

	return *this;
}


XmlSerializerIn& XmlSerializerIn::operator>>( hxa7241::utility::byte& a )
{
	in_m >> a;

	return *this;
}


XmlSerializerIn& XmlSerializerIn::operator>>( ubyte& a )
{
	in_m >> a;

	return *this;
}


XmlSerializerIn& XmlSerializerIn::operator>>( word& a )
{
	in_m >> a;

	return *this;
}


XmlSerializerIn& XmlSerializerIn::operator>>( uword& a )
{
	in_m >> a;

	return *this;
}


XmlSerializerIn& XmlSerializerIn::operator>>( dword& a )
{
	in_m >> a;

	return *this;
}


XmlSerializerIn& XmlSerializerIn::operator>>( udword& a )
{
	in_m >> a;

	return *this;
}


XmlSerializerIn& XmlSerializerIn::operator>>( long& a )
{
	in_m >> a;

	return *this;
}


XmlSerializerIn& XmlSerializerIn::operator>>( unsigned long& a )
{
	in_m >> a;

	return *this;
}


XmlSerializerIn& XmlSerializerIn::operator>>( float& a )
{
	in_m >> a;

	return *this;
}


XmlSerializerIn& XmlSerializerIn::operator>>( double& a )
{
	in_m >> a;

	return *this;
}


XmlSerializerIn& XmlSerializerIn::operator>>( char& a )
{
	char d[] = " ";
	readChars( d );
	a = d[0];

	return *this;
}


XmlSerializerIn& XmlSerializerIn::operator>>( char* cp )
{
	readChars( cp );

	return *this;
}


XmlSerializerIn& XmlSerializerIn::operator>>( string& s )
{
	readString( s );

	return *this;
}




/// implementation ---------------------------------------------------------------------------------
void XmlSerializerIn::readHeader
(
	const char* name
)
{
	readProlog( name );
}


bool XmlSerializerIn::readPreObject
(
	const char* name
)
{
	return readTagOpen( name );
}


void XmlSerializerIn::readPostObject
(
	const char* name
)
{
	readTagClose( name );
}


bool XmlSerializerIn::readPreElement
(
	const char*const collectionName,
	const char*const name,
	bool&            isCollectionEnd
)
{
	return readTagOpenElement( collectionName, name, isCollectionEnd );
}




/// implementation ---------------------------------------------------------------------------------
void XmlSerializerIn::readProlog
(
	const char*const //className
)
{
	static const char METHOD_NAME[] = "readProlog";

	const char* matches[] = { XmlSerializerOut::XML_DECLARATION, XmlSerializerOut::SERIALIZER_DECLARATION };

	for( int i = 0;  i < 2;  ++i )
	{
		readNode();

		buffer_m = matches[i];
		nodeBuffer_m.content_m = nodeBuffer_m.content_m.substr( 0, buffer_m.length() );

		if( (nodeBuffer_m.type_m != Node::PI) ||
			(nodeBuffer_m.content_m != buffer_m) )
		{
			throwException( EXCEPTION_MESSAGE1, METHOD_NAME );
		}
	}
}


bool XmlSerializerIn::readTagOpen
(
	const char*const name
)
{
	static const char METHOD_NAME[] = "readTagOpen";

	readNode();

	XmlSerializerOut::maskName( name, buffer_m );

	if( ((nodeBuffer_m.type_m != Node::OPEN) & (nodeBuffer_m.type_m != Node::EMPTY)) ||
	    (name == 0) ||
	    (nodeBuffer_m.content_m != buffer_m) )
	{
		string message( EXCEPTION_MESSAGE2 );
		message += ": expected ";

		if( (nodeBuffer_m.type_m != Node::OPEN) & (nodeBuffer_m.type_m != Node::EMPTY) )
		{
			message += "opening tag";
		}
		else
		{
			message += "'";
			message += name;
			message += "' tag, but found '";
			message += nodeBuffer_m.content_m;
			message += "' tag";
		}

		throwException( message.c_str(), METHOD_NAME );
	}

	return (nodeBuffer_m.type_m == Node::OPEN);
}


void XmlSerializerIn::readTagClose
(
	const char*const name
)
{
	static const char METHOD_NAME[] = "readTagClose";

	readNode();

	XmlSerializerOut::maskName( name, buffer_m );

	if( (nodeBuffer_m.type_m != Node::CLOSE) ||
	    (name == 0) ||
	    (nodeBuffer_m.content_m != buffer_m) )
	{
		string message( EXCEPTION_MESSAGE2 );
		message += ": expected ";

		if( nodeBuffer_m.type_m != Node::CLOSE )
		{
			message += "closing tag";
		}
		else
		{
			message += "'";
			message += name;
			message += "' tag, but found '";
			message += nodeBuffer_m.content_m;
			message += "' tag";
		}

		throwException( message.c_str(), METHOD_NAME );
	}
}


bool XmlSerializerIn::readTagOpenElement
(
	const char*const collectionName,
	const char*const name,
	bool&            isCollectionEnd
)
{
	static const char METHOD_NAME[] = "readTagOpenElement";

	isCollectionEnd = false;

	readNode();

	XmlSerializerOut::maskName( name, buffer_m );
	const bool isElementOpen = ((nodeBuffer_m.type_m == Node::OPEN) | (nodeBuffer_m.type_m == Node::EMPTY)) &
	                           (nodeBuffer_m.content_m == buffer_m);
	if( !isElementOpen )
	{
		XmlSerializerOut::maskName( collectionName, buffer_m );
		const bool isCollectionClose = (nodeBuffer_m.type_m == Node::CLOSE) &
		                               (nodeBuffer_m.content_m == buffer_m);
		if( !isCollectionClose )
		{
			string message( EXCEPTION_MESSAGE2 );
			message += ": expected '";
			message += name;
			message += "' opening tag or '";
			message += collectionName;
			message += "' closing tag, but found '";
			message += nodeBuffer_m.content_m;
			message += "' tag";

			throwException( message.c_str(), METHOD_NAME );
		}

		isCollectionEnd = isCollectionClose;
	}

	return (nodeBuffer_m.type_m == Node::OPEN);
}


void XmlSerializerIn::readString
(
	string& s
)
{
	checkCdataOpen();

	s = "";
	readUntilTriple( XmlSerializerOut::CDATA_CLOSE, &s );
}


void XmlSerializerIn::readChars
(
	char* cp
)
{
	checkCdataOpen();

	buffer_m = "";
	readUntilTriple( XmlSerializerOut::CDATA_CLOSE, &buffer_m );

	const char* sp = buffer_m.c_str();
	while( (*cp != 0) & (*sp != 0) )
	{
		*(cp++) = *(sp++);
	}
}




void XmlSerializerIn::readNode()
{
//	const char METHOD_NAME[] = "readNode";

	do
	{
		const char c = readChar();

		/// skip between markup
		if( c != '<' )
		{
			while( readChar() != '<' );

//			char d;
//			while( (d = readChar()) != '<' )
//			{
//				if( !isBlank( d ) )
//				{
//					throwException( EXCEPTION_MESSAGE3, METHOD_NAME );
//				}
//			}
		}

		readMarkup();
	}
	while( nodeBuffer_m.type_m == Node::NOTUSED );
}


void XmlSerializerIn::readMarkup()
{
	nodeBuffer_m.content_m = "";

	const char c1 = readChar();
	switch( c1 )
	{
		case '!' :
		{
			nodeBuffer_m.type_m = Node::NOTUSED;

			const char c2 = readChar();
			switch( c2 )
			{
				/// cdata
				case '[' :
				{
					/// move to after "]]>"   <![CDATA[...]]>
					readUntilTriple( XmlSerializerOut::CDATA_CLOSE, 0 );
					break;
				}
				/// comment
				case '-' :
				{
					/// move to after "-->"   <!--...-->
					readChar();
					readUntilTriple( XmlSerializerOut::COMMENT_CLOSE +1, 0 );
					break;
				}
				/// doctype
				case 'D' :
				/// unknown
				default :
				{
					/// move to after '>'
					for( char c = c2;  c != '>';  c = readChar() );
					break;
				}
			}
			break;
		}
		/// pi
		case '?' :
		{
			nodeBuffer_m.type_m = Node::PI;

			/// move to after "?>"   <?...?>
			readUntilPair( XmlSerializerOut::PI_CLOSE, nodeBuffer_m.content_m );
			break;
		}
		/// close
		case '/' :
		{
			nodeBuffer_m.type_m = Node::CLOSE;

			/// move to after ">"
			char c;
			bool blankFound = false;
			while( (c = readChar()) != '>' )
			{
				/// add to content, stop adding when blank found
				blankFound |= isBlank( c );
				if( !blankFound )
				{
					nodeBuffer_m.content_m += c;
				}
			}
			break;
		}
		/// open or empty
		default  :
		{
			nodeBuffer_m.content_m += c1;

			/// move to after ">"
			char c;
			bool blankFound = false;
			bool slashFound = false;
			while( (c = readChar()) != '>' )
			{
				/// add to content, stop adding when blank or '/' found
				blankFound |= isBlank( c );
				if( !(blankFound | slashFound) )
				{
					nodeBuffer_m.content_m += c;
				}
				slashFound = (c == '/');
			}

			nodeBuffer_m.type_m = slashFound ? Node::EMPTY : Node::OPEN;
			break;
		}
	}
}


//void XmlSerializerIn::readCharacterData
//(
//	const istream& in,
//	Node&          node
//)
//{
//	char c;
//	/// read until markup found
//	while( (c = char(getChar( in ))) != '<' )
//	{
//		switch( c )
//		{
//			/// reference
//			case '&' :
//			{
//				readUntilChar( in, ';' );
////				readReference( in, node );
//				break;
//			}
//			/// normal char
//			default :
//			{
//				/// add to content
//				node.content_m += c;
//				break;
//			}
//		}
//	}
//
//	in.putback( c );
//}


//void XmlSerializerIn::readReference
//(
//	const istream& in,
//	Node&          node
//)
//{
//	const char c = char(getChar( in ));
//	switch( c )
//	{
//		/// numerical char
//		case '#' :
//		{
//			char d = char(getChar( in ));
//			if( d == 'x' )
//			{
//				/// add '\xHEX' to content
//				/// move to after ';'
//			}
//			else
//			{
//				/// add '\xDEC' to content
//				/// move to after ';'
//			}
//			break;
//		}
//		case 'a' :
//		{
//			switch( char(getChar( in )) )
//			{
//				case 'm' :
//				{
//					add "&" to content
//					move + 2
//					break;
//				}
//				case 'p' :
//				{
//					add "'" to content
//					move + 3
//					break;
//				}
//				case ';' :
//				{
//					break;
//				}
//				default :
//				{
//					/// move to after ';'
//					readUntilChar( in, ';' );
//					break;
//				}
//			}
//			/// move to after ';'
//			readUntilChar( in, ';' );
//			break;
//		}
//		/// greater-than
//		case 'g' :
//		{
//			if( char(getChar( in )) == 't' )
//			{
//				node.content_m += '>';
//			}
//			/// move to after ';'
//			readUntilChar( in, ';' );
//			break;
//		}
//		/// less-than
//		case 'l' :
//		{
//			if( char(getChar( in )) == 't' )
//			{
//				node.content_m += '<';
//			}
//			/// move to after ';'
//			readUntilChar( in, ';' );
//			break;
//		}
//		/// quote
//		case 'q' :
//		{
//			const char d[4];
//			in.read( d, sizeof(d)/sizeof(char) );
//			if( (d[0] == 'u') & (d[1] == 'o') & (d[2] == 't') & (d[2] == ';') );
//			{
//				node.content_m += '"';
//			}
//			else
//			{
//			}
//			/// move to after ';'
//			readUntilChar( in, ';' );
//			break;
//		}
//		/// empty
//		case ';' :
//		{
//			break;
//		}
//		/// unknown
//		default :
//		{
//			/// move to after ';'
//			readUntilChar( in, ';' );
//			break;
//		}
//	}
//}


//void XmlSerializerIn::readUntilChar
//(
//	const istream& in,
//	const char     c,
//)
//{
//	while( char(getChar( in )) != c );
//}


void XmlSerializerIn::readUntilPair
(
	const char*const pair,
	string&          content
)
{
	char d[2];
	d[0] = readChar();
	d[1] = readChar();

	while( !((d[0] == pair[0]) & (d[1] == pair[1])) )
	{
		/// add to content
		content += d[0];

		d[0] = d[1];
		d[1] = readChar();
	}
}


void XmlSerializerIn::readUntilTriple
(
	const char*const triple,
	string*          content
)
{
	char d[3];
	d[0] = readChar();
	d[1] = readChar();
	d[2] = readChar();

	while( !((d[0] == triple[0]) & (d[1] == triple[1]) & (d[2] == triple[2])) )
	{
		/// add to content
		if( content != 0 )
		{
			*content += d[0];
		}

		d[0] = d[1];
		d[1] = d[2];
		d[2] = readChar();
	}
}


char XmlSerializerIn::readChar()
{
	const char c = char(in_m.get());

	/// track position
	if( lastCharWasNewline_m )
	{
		++lineNumber_m;
		charPos_m = 1;
	}
	else
	{
		++charPos_m;
	}
	lastCharWasNewline_m = (c == '\n');

	return c;
}


void XmlSerializerIn::checkCdataOpen()
{
	static const char METHOD_NAME[] = "checkCdataOpen";

	/// check cdata start
	for( int i = 0;  XmlSerializerOut::CDATA_OPEN[i] != 0;  ++i )
	{
		if( readChar() != XmlSerializerOut::CDATA_OPEN[i] )
		{
			throwException( EXCEPTION_MESSAGE4, METHOD_NAME );
		}
	}
}


void XmlSerializerIn::throwException
(
	const char*const pMessage,
	const char*const pMethodName
)
{
	std::ostringstream ss;
	ss << pMessage << "\n(file: " << fileName_m << ", line: " << lineNumber_m << ")";
	//ss << pMessage << "\n(file: " << fileName_m << ", line: " << lineNumber_m << ", char: " << charPos_m << ")";

	throw hxa7241::utility::Exception( ss.str().c_str(), CLASS_NAME, pMethodName );
}


bool XmlSerializerIn::isBlank( const char c )
{
	/// xml blank chars
	return (c == char(0x20)) | (c == char(0x09)) | (c == char(0x0D)) | (c == char(0x0A));
}


bool XmlSerializerIn::isBlank( const string& s )
{
	bool blank = true;
	for( string::size_type i = s.length();  (i-- > 0) && blank; )
	{
		blank = isBlank( s[i] );
	}

	return blank;
}
