/*--------------------------------------------------------------------

   Perceptuum renderer, version 2.1
   Copyright (c) 2002-2003,  Harrison Ainsworth.

   www.hxa7241.org

--------------------------------------------------------------------*/

/*--------------------------------------------------------------------

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307, USA

--------------------------------------------------------------------*/


#include "Win32Exception.h"

#include "Win32Dib.h"


using namespace hxa7241::win32;




/// statics
const char Win32Dib::CLASS_NAME[]         = "hxa7241::win32::Win32Dib";
const char Win32Dib::EXCEPTION_MESSAGE1[] = "width, height negative";
const char Win32Dib::EXCEPTION_MESSAGE2[] = "Couldn't get memory for BitmapInfo";
const char Win32Dib::EXCEPTION_MESSAGE3[] = "::CreateDIBSection() failed";




/// standard object services -----------------------------------------------------------------------
Win32Dib::Win32Dib
(
	const dword width,
	const dword height
)
 :	pBitmapInfo_m( 0 ),
	bitmapHandle_m( 0 ),
	pPixels_m( 0 ),
	widthInBytes_m( 0 )
{
	Win32Dib::constructBitmap( width, height, e32BitsPerPixel, 0,
	                           widthInBytes_m, pBitmapInfo_m, bitmapHandle_m, pPixels_m );
}


Win32Dib::Win32Dib
(
	const dword width,
	const dword height,
	const dword bitsPerPixel,
	const RGBQUAD palette[256]
)
 :	pBitmapInfo_m( 0 ),
	bitmapHandle_m( 0 ),
	pPixels_m( 0 ),
	widthInBytes_m( 0 )
{
	Win32Dib::constructBitmap( width, height, bitsPerPixel, palette,
	                           widthInBytes_m, pBitmapInfo_m, bitmapHandle_m, pPixels_m );
}


Win32Dib::~Win32Dib()
{
	::DeleteObject( bitmapHandle_m );
	delete[] reinterpret_cast<byte*>( pBitmapInfo_m );
}




/// construction implementation --------------------------------------------------------------------
void Win32Dib::constructBitmap
(
	const dword   width,
	const dword   height,
	      dword   bitsPerPixel,
	const RGBQUAD palette[256],
	dword&        widthInBytes,  /// out
	BITMAPINFO*&  pBitmapInfo,   /// out
	HBITMAP&      bitmapHandle,  /// out
	void*&        pPixels        /// out
)  /// static
{
	static const char METHOD_NAME[] = "constructBitmap";

	if( (width < 0) || (height < 0) )
	{
		throw hxa7241::utility::Exception( EXCEPTION_MESSAGE1, CLASS_NAME, METHOD_NAME );
	}

	bitsPerPixel &= 0xFFFFFFF8;
	bitsPerPixel = (bitsPerPixel <  8) ?  8 : bitsPerPixel;
	bitsPerPixel = (bitsPerPixel > 32) ? 32 : bitsPerPixel;


	/// rows are padded at the end to make the width a dword multiple
	widthInBytes = ((width * dword(bitsPerPixel / 8)) + 3) & ~3;


	/// allocate data structure
	static const dword bitmapInfoExtraBytes[] = { 256 * sizeof(RGBQUAD), 3 * sizeof(dword), 0, 0 };

	const dword bitmapInfoSizeInBytes = sizeof(BITMAPINFO) + bitmapInfoExtraBytes[ (bitsPerPixel / 8) - 1 ];
	pBitmapInfo = reinterpret_cast<BITMAPINFO*>(new byte[ bitmapInfoSizeInBytes ]);

	if( pBitmapInfo == 0 )
	{
		throw hxa7241::utility::Exception( EXCEPTION_MESSAGE2, CLASS_NAME, METHOD_NAME );
	}


	/// fill data structure
	pBitmapInfo->bmiHeader.biSize           =  sizeof( BITMAPINFOHEADER );   /// BitmapInfoSizeInBytes;
	pBitmapInfo->bmiHeader.biWidth          =  long(width);
	pBitmapInfo->bmiHeader.biHeight         =  -long(height);                /// neg means top down
	pBitmapInfo->bmiHeader.biPlanes         =  1;                            /// always 1
	pBitmapInfo->bmiHeader.biBitCount       =  word(bitsPerPixel);           /// 1, 4, 8, 16, 24, 32
	pBitmapInfo->bmiHeader.biCompression    =  BI_RGB;
	pBitmapInfo->bmiHeader.biSizeImage      =  0;                            /// 0 = full size no compress
	pBitmapInfo->bmiHeader.biXPelsPerMeter  =  1;                            /// ?
	pBitmapInfo->bmiHeader.biYPelsPerMeter  =  1;                            /// ?
	pBitmapInfo->bmiHeader.biClrUsed        =  0;                            /// 0 = all possible
	pBitmapInfo->bmiHeader.biClrImportant   =  0;                            /// 0 = all

	if( bitsPerPixel == e16BitsPerPixel )
	{
	   reinterpret_cast<DWORD*const>(pBitmapInfo->bmiColors)[ 0 ] = DWORD(0x001F001F);   // blue   0000000000011111,0000000000011111
	   reinterpret_cast<DWORD*const>(pBitmapInfo->bmiColors)[ 1 ] = DWORD(0x07E007E0);   // green  0000011111100000,0000011111100000
	   reinterpret_cast<DWORD*const>(pBitmapInfo->bmiColors)[ 2 ] = DWORD(0xF800F800);   // red    1111100000000000,1111100000000000
	}
	else if( bitsPerPixel == e8BitsPerPixel )
	{
		if( palette != 0 )
		{
		   for( int i = 256;  i-- != 0; )
		   {
		      pBitmapInfo->bmiColors[ i ] = palette[ i ];
		   }
		}
	}


	/// make the bitmap
	bitmapHandle = ::CreateDIBSection( 0, pBitmapInfo, DIB_RGB_COLORS, &pPixels, 0, 0 );
	if( bitmapHandle == 0 )
	{
		const DWORD errorCode = ::GetLastError();
		delete[] reinterpret_cast<byte*>( pBitmapInfo );
		throw hxa7241::win32::Win32Exception( EXCEPTION_MESSAGE3, errorCode, CLASS_NAME, METHOD_NAME );
	}
}




/// commands ---------------------------------------------------------------------------------------
void Win32Dib::setFrom
(
	const dword,  //width,
	const dword,  //height,
	const dword[] //image[],
)
{
	/*if( (width >= 0) && (height >= 0) )
	{
		/// easy 32 bit
		if( pBitmapInfo_m->bmiHeader.biBitCount == 32 )
		{
			const dword* pImage = image.getMemory();
			const dword  length = image.getLength();

			const dword* pImagePtr = pImage + length;
			      dword* pDibPtr   = static_cast<dword*>( pPixels_m ) + length;

			for( ;  pImagePtr > pImage;  )
			{
				*(--pDibPtr) = *(--pImagePtr);
			}
		}
	}*/
}




/// queries ----------------------------------------------------------------------------------------
dword Win32Dib::getWidth() const
{
	return dword( pBitmapInfo_m->bmiHeader.biWidth );
}


dword Win32Dib::getHeight() const
{
	return dword( -(pBitmapInfo_m->bmiHeader.biHeight) );
}


void* Win32Dib::getPixels() const
{
	return pPixels_m;
}


HBITMAP Win32Dib::getBitmapHandle() const
{
	return bitmapHandle_m;
}


dword Win32Dib::getBitsPerPixel() const
{
	return dword( pBitmapInfo_m->bmiHeader.biBitCount );
}


dword Win32Dib::getWidthInBytes() const
{
	return widthInBytes_m;
}
